//  ----------------------------------
//  Darryl Brown
//  University of Oregon pC++/Sage++
//
//  TIbasics.C  -  module for inspecting a pC++ program
//       at breakpoints.
//
//  
//  ----------------------------------


//  ----------------------------------
#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>

#include "compatible.h"
#ifdef SYS5
#include <string.h>
#else
#include <strings.h>
#endif
 
#include "tag"
#include "baseClasses.h"	
#include "TIbasics.h"
//  ----------------------------------

#define DEBUG 0
#define DEBUGX 0

#define TIFFILE "brk_typeInfoFile"

brk_compositeList *brk_classLIST;
brk_compositeList *brk_collectionLIST;
brk_compositeList *brk_collectionElemLIST;
brk_compositeList *brk_elementTypeLIST;
brk_compositeList *brk_structLIST;
brk_compositeList *brk_unionLIST;

void brk_basicInit() {
  brk_classLIST = new brk_compositeList("class");
  brk_collectionLIST = new brk_compositeList("Collection");
  brk_collectionElemLIST = new brk_compositeList("Collection");
  brk_elementTypeLIST = new brk_compositeList("elementType");
  brk_structLIST = new brk_compositeList("struct");
  brk_unionLIST = new brk_compositeList("union");
}

  
// ----------------------------------

void brk_printTIFFile() {

  brk_aComposite *comp, *pcomp;
  brk_compositeList *lst;
  FILE *fptr;
  char buf[40];

  strcpy(buf, TIFFILE);
  strcat(buf, ".TIF");
  fptr = fopen(buf, "w");
  if (!fptr) {
    printf("file oepn of %s failed\n", buf);
    exit(1);
  }

  lst = new brk_compositeList("temporary");

  fprintf(fptr, "/*\n\n All classes & structs which are elements of \n");
  fprintf(fptr, " a distributed collection \n\n*/\n");

  fprintf(fptr, "All Collections\n\n");
  
  comp = (brk_aComposite *) brk_collectionElemLIST->searchList();
  while (comp) {
    pcomp = brk_collectionLIST->findMember(comp);
    if (pcomp && !lst->findMember(pcomp)) {
      pcomp->print(1, fptr);
      lst->push(pcomp);
    } else {
      printf("Didnt find collection %s for TIF file.\n",
	     comp->name);
    }
    comp = (brk_aComposite *) brk_collectionElemLIST->nextItem();
  }
  
  fprintf(fptr, "All Elements\n\n");
  comp = (brk_aComposite *) brk_elementTypeLIST->searchList();
  while (comp) {
    pcomp = brk_classLIST->findMember(comp);
    if (!pcomp) {
      pcomp = brk_structLIST->findMember(comp);
    }
    if (pcomp && !lst->findMember(pcomp)) {
      pcomp->print(1, fptr);
      lst->push(pcomp);
    } else {
      printf("Not a composite structure to include in TIF file %s(?).\n",
	     comp->name);
    }
    comp = (brk_aComposite *) brk_elementTypeLIST->nextItem();
  }  
  delete lst;
  fclose(fptr);
}

// ----------------------------------
int brk_TIstrToType(char *s) {
  
  if (brk_strsame(s, "class")) {
    return T_CLASS;
  }  else if (brk_strsame(s, "Collection")) {
    return T_COLLECTION;
  } else if (brk_strsame(s, "struct")) {
    return T_STRUCT;
  } else if (brk_strsame(s, "derived class")) {
    return T_DERIVED_CLASS;
  }  else if (brk_strsame(s, "derived Collection")) {
    return T_DERIVED_COLLECTION;
  } else if (brk_strsame(s, "derived type")) {
    return T_DERIVED_TYPE;
  } else if (brk_strsame(s, "union")) {
    return T_UNION;
  } else if (brk_strsame(s, "int")) {
    return T_INT;
  } else if (brk_strsame(s, "long")) {
    return T_INT;
  } else if (brk_strsame(s, "short")) {
    return T_INT;
  } else if (brk_strsame(s, "char")) {
    return T_CHAR;
  } else if (brk_strsame(s, "float")) {
    return T_FLOAT;
  } else if (brk_strsame(s, "double")) {
    return T_DOUBLE;
  } else if (brk_strsame(s, "pointer")) {
    return T_POINTER;
  } else if (brk_strsame(s, "default")) {
    return DEFAULT;
  } else if (brk_strsame(s, "array")) {
    return T_ARRAY;
  } else if (brk_strsame(s, "void")) {
    return T_VOID;
  } else if (brk_strsame(s, "reference")) {
    return T_REFERENCE;
  } else if (brk_strsame(s, "unknown")) {
    return T_UNKNOWN;
  } else if (brk_strsame(s, "descript")) {
    printf("----descriptor---\n");
    return T_DESCRIPT;
  } else if (brk_strsame(s, "function")) {
    return T_FUNCTION;
  } else if (brk_strsame(s, "string")) {
    return T_STRING;
  } else {
    return T_UNKNOWN;
  }
}

// ----------------------------------------------------------
// returns a string representation of type ty;

char * brk_TItypeToStr (int ty) {
  
  switch (ty) {
  case T_ENUM: {
    return "enum";
    /*break;*/
  }
  case T_CLASS: {
    return "class";
    /*break;*/
  }
  case T_COLLECTION: {
    return "Collection";
    /*break;*/
  }
  case T_ARRAY: {
    return "array";
    /*break;*/
  }
  case T_STRUCT: {
    return "struct";
    /*break;*/
  }
  case T_INT: {
    return "int";
    /*break;*/
  }
  case T_FLOAT: {
    return "float";
    /*break;*/
  }
  case T_DOUBLE: {
    return "double";
    /*break;*/
  }
  case T_CHAR: {
    return "char";
    /*break;*/
  }
  case T_STRING: {
    return "string";
    /*break;*/
  }
  case T_BOOL: {
    return "bool";
    /*break;*/
  }
  case T_COMPLEX: {
    return "complex";
    /*break;*/
  }
  case T_VOID: {
    return "void";
    /*break;*/
  }
  case T_POINTER: {
    return "pointer";
    /*break;*/
  }
  case T_UNION: {
    return "union";
    /*break;*/
  }
  case T_FUNCTION: {
    return "function";
    /*break;*/
  }
  case T_DERIVED_CLASS: {
    return "derived class";
    /*break;*/
  }
  case T_DERIVED_TYPE: {
    return "derived type";
    /*break;*/
  }
  case T_DERIVED_COLLECTION: {
    return "derived Collection";
    /*break;*/
  }
  case T_REFERENCE: {
    return "reference";
    /*break;*/
  }
  case T_MEMBER_POINTER: {
    return "member pointer";
    /*break;*/
  }
  case T_DESCRIPT: {
    return "descript";
    /*break;*/
  }
  case T_UNKNOWN: {
    return "unknown";
    /*break;*/
  }
  case DEFAULT: {
    return "default";
    /*break;*/
  }
  }
  char tmp[20];
  sprintf(tmp, "\n WIERD variant Type %d\n", ty);
  printf("returning %s\n", tmp);
  return tmp;
}

// ----------------------------------------------------------
  // returns a format for a scan statement, given the type ty;

char * brk_typeToFormat(brk_basicAType *ty) {
  brk_basicPType *pt;

  switch (ty->thisType) {
  case T_CLASS: {
    return "follow";
    /*break;*/
  }
  case T_COLLECTION: {
    return "follow";
    /*break;*/
  }
  case T_ENUM:
  case T_INT: 
  case T_ARRAY: 
  case T_BOOL: 
  case T_POINTER: 
  case T_FUNCTION: {
    return "%10d";
    /*break;*/
  }
  case T_STRUCT: {
    return "follow";
    /*break;*/
  }
  case T_DOUBLE:
  case T_FLOAT: {
    return "%9.3g";
    /*break;*/
  }
  case T_CHAR: {
    return "%c";
    /*break;*/
  }
  case T_STRING: {
    return "%s";
    /*break;*/
  }
  case T_COMPLEX: {
    return "%d";
    /*break;*/
  }
  case T_VOID: {
    return "";
    /*break;*/
  }
  case T_UNION: {
    return "follow";
    /*break;*/
  }
  case T_DERIVED_CLASS: {
    return "follow";
    /*break;*/
  }
  case T_DERIVED_TYPE: {
    pt = (brk_basicPType *) ty;
    return brk_typeToFormat(pt->baseType);
    /*break;*/
  }
  case T_DERIVED_COLLECTION: {
    return "follow";
    /*break;*/
  }
  case T_REFERENCE: {
    return "follow";
    /*break;*/
  }
  case T_MEMBER_POINTER: {
    return "follow";
    /*break;*/
  }
  case T_DESCRIPT: {
    return "follow";
    /*break;*/
  }
  case T_UNKNOWN: {
    return "unknown";
    /*break;*/
  }
  case DEFAULT: {
    return "default";
    /*break;*/
  }
  }
  return "???";
}


int brk_nonComposite(char *str) {
  if (brk_TIstrToType(str) != T_UNKNOWN) {
    return 1;
  } else {
    return 0;
  }
}


// ----------------------------------
void brk_newComposite(char *theName, int ctype) {
  
  switch (ctype) {

  case T_CLASS : {
    brk_classLIST->push(new brk_myStr(theName));
    break;
  }
  case T_COLLECTION :  {
    brk_collectionLIST->push(new brk_myStr(theName));
    break;
  }

  case T_STRUCT :  {
    brk_structLIST->push(new brk_myStr(theName));
    break;
  }
    
  case T_UNION :  {
    brk_unionLIST->push(new brk_myStr(theName));
    break;
  }
  }
}


// ----------------------------------
brk_aComposite *brk_findInCompositeList(char *compName, char *elemName, 
				brk_compositeList *cList) {
  brk_aComposite *c;

  c = (brk_aComposite *) cList->searchList();
  
  while (c) {
    if (brk_strsame(compName, c->name)) {
      if ((elemName) &&
	  brk_strsame(elemName, c->elementType)) {
	return c;
      } else {
	return c;
      }
    }
    c = (brk_aComposite *) cList->nextItem();
  }
  return NULL;
}

// ----------------------------------
// this should create a new composite if it doesnt find what it;
// is looking for.;
brk_aComposite *brk_lookupComposite(brk_basicAType *cType) {

  brk_compositeList *cList;
  brk_aComposite *c;
  brk_basicAType *t;
  
  switch (cType->thisType) {

  case T_CLASS : {
    cList = brk_classLIST;
    break;
  }
  case T_COLLECTION :  {
    cList = brk_collectionLIST;
    break;
  }
  case T_STRUCT :  {
    cList = brk_structLIST;
    break;
  }
  case T_UNION :  {
    cList = brk_unionLIST;
    break;
  }
  case T_UNKNOWN: {
    // look through collections, classes and structs
    t = new brk_basicAType(cType->description, T_COLLECTION);
    if (c = brk_lookupComposite(t)) {
      return c;
    } else {
      t = new brk_basicAType(cType->description, T_CLASS);
      if (c = brk_lookupComposite(t)) {
	return c;
      }	else {
	t = new brk_basicAType(cType->description, T_STRUCT);
	if (c = brk_lookupComposite(t)) {
	  return c;
	} else {
	  return NULL;
	} 
      }
    }
  }
  }
  if (!(cType->thisType == T_UNKNOWN)) {
    return brk_findInCompositeList(cType->description, NULL, cList);
  }
  return NULL;
}


//  ##########################################################
//  brk_myStr class methods

// constructor for string class;
 brk_myStr::brk_myStr(char *st) {
   ing = brk_stringSave(st);
 } 

void brk_myStr::printToBuf(int tabs, char *st) {
  char tabstr[2];

  *st = '\0';
  // add tabs
  sprintf(tabstr, "\t");
  for (int j = 0; j < tabs; j++) {
    strcat(st, tabstr);
  }
  strcat(st, ing);
}

void brk_myStr::printToBuf(char *st) {
  *st = '\0';
  strcpy(st, ing);
}

void brk_myStr::print() {
  printf("%s", ing);
}

void brk_myStr::print(int tabs) {
  brk_printtabs(tabs);
  printf("%s", ing);
}

void brk_myStr::print(int tabs, FILE *fptr) {
  brk_printtabs(tabs, fptr);
  fprintf(fptr, "%s", ing);
}

void brk_myStr::print(FILE *fptr) {
  fprintf(fptr, "%s", ing);
}

//  ##########################################################
  //  brk_basicAType class methods
  
  
 brk_basicAType::brk_basicAType() {
 }


 brk_basicAType::brk_basicAType(char *desc, int ty) {
   description = brk_stringSave(desc);
   thisType = ty;
 }


  
void brk_basicAType::print() {
  print(0, stdout);
}

void brk_basicAType::print(int tabs) {
  print(tabs, stdout);
}

void brk_basicAType::printToBuf(int tabs, char *buf) {
  char tabstr[2], tbuf[100];
  brk_aComposite *p;
  char *tmp;

  *buf = '\0';

  // add tabs
  sprintf(tabstr, "\t");
  for (int j = 0; j < tabs; j++) {
    strcat(buf, tabstr);
  }

  switch (thisType) {

  case T_CLASS : 
  case T_COLLECTION :  
  case T_STRUCT :  
  case T_UNION :  {
    // lookup the correct composite.;
    p = (brk_aComposite *) brk_lookupComposite(this);
    if (!p) {
      // should we create a new one here??;
      sprintf(tbuf, "%s %s", brk_TItypeToStr(thisType), 
	      this->description);
      strcat(buf, tbuf);
    } else {
      sprintf(tbuf, "%s %s", brk_TItypeToStr(thisType), 
	      this->description);
      strcat(buf, tbuf);
    }
    break;
  }
  default: {	
    tmp = brk_TItypeToStr(thisType);
    sprintf(tbuf, "%s", tmp);
    strcat(buf, tbuf);
    break;
  }
  }
}

void brk_basicAType::print(int tabs, FILE *fptr) {
  
  brk_aComposite *p;
  char *tmp;

  fflush(stdout);
  switch (thisType) {

  case T_CLASS : 
  case T_COLLECTION :  
  case T_STRUCT :  
  case T_UNION :  {
    // lookup the correct composite.;
    p = (brk_aComposite *) brk_lookupComposite(this);
    if (!p) {
      // should we create a new one here??;
      fprintf(fptr, "%s %s", brk_TItypeToStr(thisType), 
	      this->description);
    } else {
      fprintf(fptr, "%s %s", brk_TItypeToStr(thisType), 
	      p->name);
    }
    break;
  } 
  default: {	
    brk_printtabs(tabs, fptr);
    tmp = brk_TItypeToStr(thisType);
    fprintf(fptr, "%s", tmp);
    break;
  }
  }
  fflush(stdout);
}


//  ##########################################################
  //  brk_basicPType class methods
  
  
 brk_basicPType::brk_basicPType() {
   dimensions = 0;
   baseType = NULL;
   description = NULL;
   thisType = 0;
 }


 brk_basicPType::brk_basicPType(char *desc, int ty) {
   description = brk_stringSave(desc);
   thisType = ty;
   baseType = NULL;
   dimensions = 0;
 }

 brk_basicPType::brk_basicPType(brk_basicAType *bType, int ty) {
   description = NULL;
   baseType = bType;
   thisType = ty;
   dimensions = 0;
 }

void brk_basicPType::print() {
  print(0, stdout);
}
void brk_basicPType::print(int tabs) {
  print(tabs, stdout);
}

void brk_basicPType::print(int tabs, FILE *fptr) {
 //print this brk_basic type;
 int i;
 char buf[50];

 if (thisType == T_DERIVED_TYPE) {
   if (baseType) {
#if DEBUG     
     printf("printing out derived type %s\n", description);
#endif
     baseType->printToBuf(0, buf);
     fprintf(fptr, "%s (%s)", description, buf);
   } else {
     fprintf(fptr, "%s", description);
     printf("no derived base Type for %s", description);
     // exit(1);
   }
 } else {
   // print out the baseType;
   if (baseType) {
     baseType->print(tabs, fptr);
   } else {
     fprintf(fptr, "no basetype for pointer or arrayType");
     exit(1);
   }
   
   fprintf(fptr, " ");
   
#if DEBUG
   printf("printing out %d dimensions \n", dimensions);
   fflush(stdout);
#endif
   // now print the dimensions, if there are any...;
   for (i = 0; i < dimensions; i++) {
     fprintf(fptr, "[%d]", sizes[i]);
   }
   
 }
 fflush(stdout);
}

void brk_basicPType::printToBuf(int tabs, char *buf) {
  char tbuf[100];

  *buf = '\0';

 if (thisType == T_DERIVED_TYPE) {
   if (baseType) {
     baseType->printToBuf(0, tbuf);
     sprintf(buf, "%s (%s)", description, tbuf);
#if DEBUG     
     printf("printing out derived type to buffer \"%s\"\n", buf);
#endif
   } else {
     printf("no derived base Type for %s", description);
     exit(1);
   }
 } else {
   // print out the baseType;
   if (baseType) {
     baseType->printToBuf(tabs, tbuf);
     strcat(buf, tbuf);
     
#if DEBUG 
     printf("data type of field %s is %s\n", brk_TItypeToStr(thisType), buf);
     fflush(stdout);
#endif
     
   } else {
     strcat(buf, "noType");
   }
 }
  
  // add a space;
  strcat(buf, " ");
  
  // now print the dimensions...;
  for (int i = 0; i < dimensions; i++) {
    sprintf(tbuf, "[%d]", sizes[i]);
    strcat(buf, tbuf);
  }
  
}

//  ##########################################################
//  brk_aField class methods

// constructor for brk_aField class;
 brk_aField::brk_aField(brk_basicAType *t, char *n) {
   fieldType = t;
   name = brk_stringSave(n);
 }

// constructor for brk_aField class;
 brk_aField::brk_aField() {
 }

void brk_aField::print() {
  print(0);
}

void brk_aField::print(int tabs) {
   
   brk_printtabs(tabs);
   
   printf ("#type# ");
   fieldType->print(0);
   printf(" #field# %s\n", name);
 }


void brk_aField::print(int tabs, FILE *fptr) {
  
  brk_printtabs(tabs, fptr);
  
  fprintf(fptr, "#type# ");
  if (fieldType) {
    fieldType->print(0, fptr);
  } else {
    printf("brk: Warning, no type exists for field %s.\n", name);
  }
  fprintf(fptr, " #field# %s\n", name);
}

void brk_aField::printToBuf(int tabs, char *buf) {
  char tabstr[3], tbuf[100];

  *buf = '\0';
  // add tabs
  sprintf(tabstr, "\t");
  for (int j = 0; j < tabs; j++) {
    strcat(buf, tabstr);
  }
  strcat(buf, "#type# ");
  fieldType->printToBuf(0, tbuf);
  strcat (buf, tbuf);
  sprintf(tbuf, " #field# %s\n", name);
  strcat(buf, tbuf);
}

//  ##########################################################
//  brk_aComposite class methods


void brk_aComposite::push (brk_basePtr *h) {
  
  brk_aField *af;
  
  af = (brk_aField *) h;
  // printf("pushing %s\n", af->name);
  if ((af->fieldType->thisType == T_POINTER) ||
      (af->fieldType->thisType == T_ARRAY)) {
    hasPointer++;
  }
 brk_linkedList::push(h);
}

 brk_aComposite::brk_aComposite(char * aName) {
   name = brk_stringSave(aName);
   superClasses = new brk_linkedList();
   hasPointer = 0;
   elementType = "";
 }

 brk_aComposite::brk_aComposite(char * aName, int t) {
   thisType = t;
   name = brk_stringSave(aName);
   superClasses = new brk_linkedList();
   hasPointer = 0;
   elementType = "";
 }

 brk_aComposite::brk_aComposite() {
   superClasses = new brk_linkedList();
   hasPointer = 0;
   elementType = "";
 }

int brk_aComposite::compare(brk_basePtr *) {
  return -1; /*BM: is that right???*/
}

int brk_aComposite::compare(brk_ptrNode *comp) {
  return strcmp(name, ((brk_aComposite *) comp)->name);
}

void brk_aComposite::print() {
  print(0, stdout);
}

void brk_aComposite::print(int tabs) {
  print(tabs, stdout);
}

void brk_aComposite::print(int tabs, FILE *fptr) {
  brk_myStr *cl;
  brk_aField *f;
  
  brk_printtabs(tabs, fptr);
  
  switch (thisType) {
    
  case T_CLASS : {
    fprintf(fptr, "class");
    break;
  }
  case T_COLLECTION :  {
    fprintf(fptr, "Collection");
    break;
  }

  case T_STRUCT :  {
    fprintf(fptr, "struct");
    break;
  }

  case T_UNION :  {
    fprintf(fptr, "union");
    break;
  }

  }

  if (thisType == T_COLLECTION) {
    fprintf(fptr, " %s < %s > ( ", name, elementType);
  } else {
    fprintf(fptr, " %s ( ", name);
  }

  cl = (brk_myStr *) superClasses->searchList();
  while (cl) {
    cl->print(fptr);
    fprintf(fptr, " ");
    cl = (brk_myStr *) superClasses->nextItem();
  }
  fprintf(fptr, ")\n");

  f = (brk_aField *) searchList();

  while (f) {
    f->print((tabs+1), fptr);
    f = (brk_aField *) nextItem();
  }
  
  searchList();
  
  fprintf(fptr, "\n");
}

void brk_aComposite::printToBuf(int tabs, char *buf) {
  brk_myStr *cl;
  brk_aField *f;
  char tabstr[2], tbuf[1000];
  
  *buf = '\0';

  sprintf(tabstr, "\t");
  for (int j = 0; j < tabs; j++) {
    strcat(buf, tabstr);
  }
  
  switch (thisType) {
    
  case T_CLASS : {
    strcat(buf, "class");
    break;
  }
  case T_COLLECTION :  {
    strcat(buf, "Collection");
    break;
  }

  case T_STRUCT :  {
    strcat(buf, "struct");
    break;
  }

  case T_UNION :  {
    strcat(buf, "union");
    break;
  }

  }

  if (!name) { 
    printf("problem with name type\n");
    fflush(stdout);
    return;
  }


  if (thisType == T_COLLECTION) {
    if (!elementType) { 
      printf("problem with element type: %s\n", name);
      fflush(stdout);
      return;
    }
    sprintf(tbuf, " %s < %s > ( ", name, elementType);
  } else {
    sprintf(tbuf, " %s ( ", name);
  }
  strcat(buf, tbuf);

  cl = (brk_myStr *) superClasses->searchList();
  while (cl) {
    cl->printToBuf(tbuf);
    strcat(buf, tbuf);
    strcat(buf, " ");
    cl = (brk_myStr *) superClasses->nextItem();
  }

  sprintf(tbuf, ")\n");


  strcat(buf, tbuf);
  f = (brk_aField *) searchList();
  while (f) {
    f->printToBuf((tabs+1), tbuf);
    strcat(buf, tbuf);
    f = (brk_aField *) nextItem();
  }
  sprintf(tbuf, "\n");
  strcat(buf, tbuf);

}


//  ##########################################################
//  brk_compositeList functions;

void brk_compositeList::print() {
  print(0);
}

void brk_compositeList::print(int tabs) {
  
  brk_aComposite *f;
  
  f = (brk_aComposite *) searchList();
  while (f) {
    f->print(tabs);
    f = (brk_aComposite *) nextItem();
  }
  
}


void brk_compositeList::print(int tabs, FILE *fptr) {
  
  brk_aComposite *f;
  
  f = (brk_aComposite *) searchList();
  while (f) {
    f->print(tabs, fptr);
    f = (brk_aComposite *) nextItem();
  }
  
}

void brk_compositeList::printToBuf(int tabs, char *buf) {
  char tabstr[2], tbuf[1000];
  brk_aComposite *f;
  
  *buf = '\0';

  // add tabs
  sprintf(tabstr, "\t");
  for (int j = 0; j < tabs; j++) {
    strcat(buf, tabstr);
  }

  f = (brk_aComposite *) searchList();
  while (f) {
    f->printToBuf(tabs, tbuf);
    strcat(buf, tbuf);
    f = (brk_aComposite *) nextItem();
  }
}


brk_aComposite * brk_compositeList::findMember(brk_aComposite *n) {
  brk_aComposite *h;
  
  h = (brk_aComposite *) searchList();
  while (h) {
    if (h->compare((brk_ptrNode *)n) == 0) {
      return h;
    }
    h = (brk_aComposite *) nextItem();
  }
  return NULL;
}

// ----------------------------------;
void brk_addCollectionWithElem(char *collName, char *elemName) {
  
  brk_compositeList *cList = NULL;
  brk_aComposite *comp;
  cList = brk_collectionElemLIST;
  
  if (!cList) {
    printf("No collection Elem list??? Initializing\n");
    brk_basicInit();
    cList = brk_collectionElemLIST;
  }

  // search for the composite this new field belongs to...;
  comp = (brk_aComposite *) cList->searchList();
  while (comp) {
    if (brk_strsame(comp->name, collName) &&
	brk_strsame(comp->elementType, elemName)) {
	break;
      }
    comp =  (brk_aComposite *) cList->nextItem();
  }
  
  // if not found, create the new one...;
  if (!comp) {
    cList->push(comp = new brk_aComposite(collName, T_COLLECTION));
    comp->elementType = brk_stringSave(elemName);
    
    printf("adding element type %s for colleciton %s to collElemList\n",
	   elemName, collName);
    comp->print(0);
    fflush(stdout);
  }
}

// --------------------------------
brk_compositeList::brk_compositeList() {
}


// --------------------------------
brk_compositeList::brk_compositeList(char *ety) {
  elementType = brk_stringSave(ety);
}


// --------------------------------;
void brk_parseDimensions(brk_basicPType *pt, char *dimstr) {

  int dim[10], *dptr, numdim = 0;
  char *p;
  
  p = strtok(dimstr, "[]");
  dptr = dim;

  while (p) {
    numdim++;
    sscanf(p, "%d", dptr);
    p = strtok(NULL, "[]");
    dptr++;
  }
  pt->dimensions = numdim;
  
#if DEBUGX
  printf("getting dimenstions (%d)...",
	 numdim);
#endif

  pt->sizes = new int[numdim];
  for (int i=0; i < numdim; i++) {
    pt->sizes[i] = dim[i];

#if DEBUGX
    printf("[%d]\n",
	   dim[i]);
#endif
  }

}

// --------------------------------;
brk_basicAType *brk_strToBasicType(char *buf1) {

  brk_basicAType *at;
  brk_basicPType *pt;
  char *p, ty[50], *dims, *idx, *idx2;
  char tyType;

  dims = NULL;
  idx = NULL;
  if (idx = strchr(buf1, '[')) {
    // contains array type;
    if (idx2 = strchr(buf1, ')')) {
      if (idx2 < idx) {
	tyType = 'a';
	strncpy(ty, buf1, (idx-buf1));
	*(ty+(idx-buf1)) = '\0';
	dims = idx;
      } else {
	idx = strchr(buf1, '(');
	// array type is inside derived type;
	// is derived type (typedefd);
#if DEBUG 
	printf("\n\n begin derived type: %s \n\n", idx);
#endif
	tyType = 'd';
	strncpy(ty, buf1, (idx-buf1-1));
	*(ty+(idx-buf1-1)) = '\0';
#if DEBUG 
	printf("parsed %s for a derived type name\n", ty);
#endif
	dims = (idx+1);
      }
    } else {
      tyType = 'a';
      strncpy(ty, buf1, (idx-buf1));
      *(ty+(idx-buf1)) = '\0';
      dims = idx;
    }
    
  } else if (idx = strchr(buf1, '*')) {
    // is pointer type;
    tyType = 'p';
    strncpy(ty, buf1, (idx-buf1));
    *(ty+(idx-buf1)) = '\0';
    dims = idx;
  } else if (idx = strchr(buf1, '(')) {
    // is derived type (typedefd);
#if DEBUG 
    printf("\n\n begin derived type: %s\n\n", idx);
#endif
    tyType = 'd';
    strncpy(ty, buf1, (idx-buf1-1));
    *(ty+(idx-buf1-1)) = '\0';
#if DEBUG
    printf("parsed \"%s\" for a derived type name\n", ty);
#endif
    dims = (idx+1);
  } else {
    // is brk_basic primitive type;
    tyType = 'b';
    p = NULL;
    p = strtok(buf1, " ");
    at = new brk_basicAType("", brk_TIstrToType(p));
    // check for structs or class;
    if ((at->thisType == T_STRUCT) ||
	(at->thisType == T_CLASS)) {
      p = strtok(buf1, " ");
      if (p) {
	at->description = brk_stringSave(p);
      } else {
	printf("\n\n\t\t\t\t\tno composite name type??\n\n\n");
	exit(1);
      }
    }
  }
  
#if DEBUGX
  printf("tyType : %c\n", tyType);
#endif
  
  if (!p) {
    printf("problem with turning string \"%s\" into brk_basic type\n", buf1);
  }

  // deal with array or pointer or derived types first;
  if ((tyType == 'a') || (tyType == 'p') || (tyType == 'd')) { 

    if (tyType == 'a') {
      at = brk_strToBasicType(ty);
      // type is array;
      pt = new brk_basicPType("", T_ARRAY);

      pt->baseType = at;
      at = (brk_basicAType *) pt;
      p = strtok(dims, " ");

#if DEBUGX 
      printf("grabbed rest of array or pointer type: \"%s\".\n", 
	     p);
      fflush(stdout);
#endif      

      brk_parseDimensions(pt, p);
    } else if (tyType == 'd') {
      // type must be derived;
#if DEBUG 
      printf("before strtok dims %s\n", dims);
#endif
      p = strtok(dims, ")");
#if DEBUG 
      printf("grabbed rest of derived type: \"%s\", for %s:\n ", 
	     p, ty);
#endif
      
      at = brk_strToBasicType(p);

#if DEBUG 
      printf("brk_basic type of derived type::\n");
      at->print(0);
      printf("\n");
#endif
      
      pt = new brk_basicPType(ty, T_DERIVED_TYPE);
      pt->baseType = at;
      at = (brk_basicAType *) pt;
#if DEBUG 
      printf("whole derived type: \n ");

      pt->print(0);
      printf("\n\n end Derived type \n\n");
      fflush(stdout);
#endif      
    } else {
      at = brk_strToBasicType(ty);
      // type must be pointer;
      pt = new brk_basicPType("", T_POINTER);
      pt->baseType = at;
      at = (brk_basicAType *) pt;
    }
  } 
#if DEBUG 
  printf("\n\n printing type::\n\n");
  at->print(0);
  printf("\n\n completed reading in type\n\n");
  fflush(stdout);
#endif      
  return at;
}
  

// --------------------------------;
void brk_readInFields(FILE *fptr, brk_aComposite *comp) {
	  
  char *format;
  brk_aField *af;
  char line[100], buf1[100], buf2[100];
  brk_basicAType *bt;

  format = "\t\t#type# %[^#] #field# %[^ \n]";

  // now read in all fields...;
  fgets(line, 100, fptr);
  *buf1 = '\0';
  *buf2 = '\0';
  while (sscanf(line, format, buf1, buf2) >= 1) {
#if DEBUG
    printf("READING IN FIELD TYPE %s field %s\n\t field:",
	   buf1, buf2);
#endif
    bt = brk_strToBasicType(buf1);
    af = new brk_aField(bt, buf2);

#if DEBUG
    printf("DONE READING IN:\n");
    af->print(0);
    printf("\n\n\n");

#endif

    comp->push(af);
    fgets(line, 100, fptr);
    *buf1 = '\0';
    *buf2 = '\0';
  }
}


// --------------------------------;
void brk_readIn() {

  FILE *TIfptr;
  char buf1[100], buf2[100], line[100], buf3[100], *p, fbuf[40];
  brk_aComposite * comp;
  brk_myStr *s;
  int num;
  char * formats[4];

  // make certain lists are initialized.;
  if (!brk_collectionLIST) {
    brk_basicInit();
  }

  formats[0] = "\tCollection  %[^<] < %[^>] > ( %[^)] )";
  formats[3] = "\tCollection  %[^<] <  > ( %[^)] )";
  formats[1] = "\tclass %[^(] ( %[^)] )";
  formats[2] = "\tstruct %[^(] ";

  strcpy(fbuf, TIFFILE);
  strcat(fbuf, ".TIF");
  TIfptr = fopen(fbuf, "r");
  if (!TIfptr) {
    printf("file oepn of %s failed\n", fbuf);
    exit(1);
  }
  
  
  fgets(line, 100, TIfptr);
  
  while (!feof(TIfptr)) {

    *buf1 = '\0';
    *buf2 = '\0';
    *buf3 = '\0';
    p = buf2;

    // test to see if it matches collection..;
    if (((num = sscanf(line, formats[0], buf1, buf2, buf3)) > 1) ||
	((num = sscanf(line, formats[3], buf1, buf3)) >= 1)) {
      comp = new brk_aComposite(buf1, T_COLLECTION);
      comp->name = brk_stringSave(buf1);
      comp->elementType = brk_stringSave(buf2);

      p = strtok(buf3, " ");
      while (p) {
	s = new brk_myStr(p);
	comp->superClasses->push(s);
	p = strtok(NULL, " ");
      }
      brk_collectionLIST->push(comp);
      brk_readInFields(TIfptr, comp);
    } else 
      // test for a class def since collection didnt work;
    if ((num = sscanf(line, formats[1], buf1, buf2)) >= 1) {
      comp = new brk_aComposite(buf1, T_CLASS);
      comp->name = brk_stringSave(buf1);
      
      p = strtok(buf2, " ");
      while (p) {
	fflush(stdout);
	s = new brk_myStr(p);
	comp->superClasses->push(s);
	p = strtok(NULL, " ");
      }

      brk_classLIST->push(comp);
      brk_readInFields(TIfptr, comp);
    } else 
      // test for a struct def since collection && class didnt work;
    if ((num = sscanf(line, formats[2], buf1)) >= 1) {
      comp = new brk_aComposite(buf1, T_STRUCT);
      comp->name = brk_stringSave(buf1);
      brk_structLIST->push(comp);
      brk_readInFields(TIfptr, comp);
    } else {
    }
    fgets(line, 100, TIfptr);
  }
}
