//  ----------------------------------
//  Darryl Brown
//  University of Oregon pC++/Sage++
//
//  TItypeDescription.C  -  module for inspecting a pC++ program
//       at breakpoints.
//
//  
//  ----------------------------------
//  ----------------------------------
#include <stdio.h>
#include <stdlib.h>

#include "compatible.h"
#ifdef SYS5
#include <string.h>
#else
#include <strings.h>
#endif
 
#include "sage++user.h"
#include "baseClasses.h"	
#include "sage++classhierarchy.h"
#include "TIbasics.h"
#include "sageUtils.h"
#include "TItypeDesc.h"
//  ----------------------------------

#define DEBUG 0



void brk_mySageInit() {
  brk_classLIST = (brk_sCompositeList *) new brk_sCompositeList("class");
  brk_collectionLIST = (brk_sCompositeList *) new brk_sCompositeList("Collection");
  brk_collectionElemLIST = (brk_sCompositeList *) new brk_sCompositeList("Collection");
  brk_elementTypeLIST = (brk_sCompositeList *) new brk_sCompositeList("elementType");
  brk_structLIST = (brk_sCompositeList *) new brk_sCompositeList("struct");
  brk_unionLIST = (brk_sCompositeList *) new brk_sCompositeList("union");
}

// ----------------------------------;
void brk_addElementType(char *elemName, SgStatement *st) {
  
  brk_compositeList * cList;
  brk_aComposite *comp;
  cList = brk_elementTypeLIST;
  int ty;

  // check to see if this a standard (non-composite) type;
  if (brk_nonComposite(elemName)) {
    printf("collection of non composite: %s\n", elemName);
    return;
  }

  if (cList) {	
    // search for the composite this new field belongs to...;
    comp = (brk_aComposite *) cList->searchList();
    while (comp) {
      if (brk_strsame(comp->name, elemName)) {
	break;
      }
      comp =  (brk_aComposite *) cList->nextItem();
    }
    
    // if not found, create new one...
    if (!comp) {
      if (brk_lookupComposite(new brk_basicAType(elemName, T_CLASS))) {
	ty = T_CLASS;
      } else if (brk_lookupComposite(new brk_basicAType(elemName, T_STRUCT))) {
	ty = T_STRUCT;
      } else {
	ty = NULL;
      }
      if (ty) {
	if (st) {
	  cList->push(comp = new brk_sComposite(elemName, ty, st));
	} else {
	  cList->push(comp = new brk_sComposite(elemName, ty));
	}
	fflush(stdout);
      } else {
	printf("no composite of type %s exists\n", elemName);
	printf("current struct list:\n");
	brk_structLIST->print(0);
	printf("current class list:\n");
	brk_classLIST->print(0);
	perror("cant go on");
	exit(0);
      }
    }
  } else {
    perror("No collection Elem list???\n");
  }
}

// ----------------------------------;
void brk_addSCollectionWithElem(char *collName, char *elemName) {
  
  brk_compositeList *cList = NULL;
  brk_aComposite *comp;
  cList = brk_collectionElemLIST;
  
  if (cList) {
    // search for the composite this new field belongs to...;
    comp = (brk_aComposite *) cList->searchList();
    while (comp) {
      if (brk_strsame(comp->name, collName) &&
	  brk_strsame(comp->elementType, elemName)) {
	break;
      }
      comp =  (brk_aComposite *) cList->nextItem();
    }
    
    // if not found, create the new one...;
    if (!comp) {
      cList->push(comp = new brk_aComposite(collName, T_COLLECTION));
      comp->elementType = brk_stringSave(elemName);
#     if DEBUG
      printf("adding element type %s for colleciton %s to collElemList\n",
	     elemName, collName);
      comp->print(0);
#    endif
      
      fflush(stdout);
    }
  } else {
    perror("No collection Elem list???\n");
  }
}


// ----------------------------------;



void brk_build_coll_types (SgFile *aFile) {
    SgSymbol *classsymb;
    char *classname;
    SgStatement *s, *st;
    SgType *elemType;
    SgDerivedCollectionType *d;
    char *elemname;
    SgCExpStmt *ass;
    SgVarDeclStmt *var;
    SgSymbol *varsymb;
    SgExprListExp *elist;
    SgFile *f;

    f = aFile;

    s = aFile->firstStatement();
    while ( s )
    {
      if ( var = isSgVarDeclStmt (s) )
      {
        for (int i=0; i<var->numberOfSymbols(); i++) {
	  varsymb = var->symbol(i);
          if ( d = isSgDerivedCollectionType (varsymb->type()) ) {
            /* -- Symbol is of type CollectionType<ElementType>    -- */
            /* -- get type names                                   -- */
            /* ------------------------------------------------------ */
            classsymb = d->collectionName();
            classname = classsymb->identifier();
            elemType  = d->elementClass();
            elemname  = (elemType->symbol())->identifier();

	    // add to colleciton type list;
	    brk_addSCollectionWithElem(classname, elemname);
	    st = findCompDeclStmt(elemname, f);
	    brk_addElementType(elemname, st);
          }
        }
      }
      else if ( (ass=isSgCExpStmt (s)) && (elist=isSgExprListExp(ass->expr())))
      {
        SgExpression *e;
        SgNewExp *n;
        SgPointerType *t;
        char *vname;

        for (int i=0; i<elist->length(); i++)
        {
          e = elist->elem(i);
          if ( (e->variant() == ASSGN_OP) &&
               (n = isSgNewExp(e->rhs())) &&
               (t = isSgPointerType(n->type())) &&
               (d = isSgDerivedCollectionType(t->baseType()))
             )
          {
            /* -- dynamic collection allocation Statement:         -- */
            /* -- var = new CollectionType<ElementType>()          -- */
            /* ------------------------------------------------------ */
            classsymb = d->collectionName();
            classname = classsymb->identifier();
            elemType  = d->elementClass();
            elemname  = (elemType->symbol())->identifier();
            vname     = new char[strlen(e->lhs()->symbol()->identifier())+2];
	            sprintf (vname, "*%s", e->lhs()->symbol()->identifier());

	    
	    brk_addSCollectionWithElem(classname, elemname);
	    st = findCompDeclStmt(elemname, f);
	    brk_addElementType(elemname, st);
          }
        }
      }
      s = s->lexNext();
    }
    
    
}


// ########################################################

// ----------------------------------;
void brk_addAccessFuncsToList(brk_sCompositeList *slist, SgFile *f, char *fname) {

// 64 = public, 256 = collection (dont have class??);
#define MY_PUBLIC_FUNC (64+256)
  
  brk_sComposite *scomp;
  SgClassStmt *st;
  SgType *ty;

  ty = new SgType(T_VOID);

  scomp = (brk_sComposite *) slist->searchList();
  while (scomp) {
#if DEBUG
    printf("instrumenting %s class\n", scomp->name);
#endif
    
    st = (SgClassStmt *) findCompDeclStmt(scomp->name, f);
    addFunctionToClass(st, ty, fname, PUBLIC_FIELD);
    scomp = (brk_sComposite *) slist->nextItem();
  }
}


// ----------------------------------;
brk_basicAType *brk_createType(SgType *ty) {

  brk_aType *at;
  brk_pType *pt;
  brk_basicAType *bt;

  if (!ty) return NULL;

  switch(ty->variant()) {
  case T_ARRAY:
  case T_DERIVED_TYPE:
  case T_POINTER: {
    // grab baseType;
    pt = new brk_pType(ty);
    bt = (brk_basicAType *) pt;
    break;
  }
  case T_DESCRIPT: {
    // ignore descriptors;
    bt = brk_createType(ty->baseType());
    break;
  }
  default: {
    at = new brk_aType(ty);
    bt = (brk_basicAType *) at;
    break;
  }
  }
  return bt;
}


// ----------------------------------;
void brk_buildTypesFromFile(SgFile *aFile) {
  int ty;
  SgSymbol  *curSym;
  SgStatement *st;
  SgVarDeclStmt *vrSt;
  SgFieldSymb *flSym;
  int flag;
  brk_basicAType *ftype;
  brk_basicAType *ctype;

  // loop over Symbol table of file.;
  
  curSym  = aFile->firstSymbol(); 
  while (curSym) {
    
    if (isSgMemberFuncSymb(curSym)) {
      
      st = NULL;
      
      // statement of declaration..;
      vrSt = (SgVarDeclStmt *) curSym->declaredInStmt();
      
      // now, control parent of the declarative statement;
      st = vrSt->controlParent();

      // parent controler is not a struct, it is a class or collection...;
      
      // ---------------------------------------------;
      switch  (st->variant()) {
      case CLASS_DECL: {
	ty = T_CLASS;
	break;
      }
      case COLLECTION_DECL: {
	ty = T_COLLECTION;
	break;
      }
      case STRUCT_DECL: {
	ty = T_STRUCT;
	break;
      }
      case UNION_DECL: {
	ty = T_UNION;
	break;
      }
      }
      ctype = new brk_aType(st->symbol()->identifier(), ty); 
      
#if DEBUG
      printf("defining new memb func %s for comp %s \n", curSym->identifier(),
	     st->symbol()->identifier());
#endif

      brk_getOrMakeComposite(ctype);
      
      // ---------------------------------------------;
    }
  
    if (flSym = isSgFieldSymb(curSym)) {
      
      st = NULL;
      
      // statement of declaration..;
      if ((vrSt = (SgVarDeclStmt *) curSym->declaredInStmt()) && 
	   // now, control parent of the declarative statement;
	   (st = vrSt->controlParent())
	  ) {
	
	
	// parent controler is not a struct, it is a class or collection...;
	
	// ---------------------------------------------;
	switch  (st->variant()) {
	case CLASS_DECL: {
	  ty = T_CLASS;
	  break;
	}
	case COLLECTION_DECL: {
	  ty = T_COLLECTION;
	  break;
	}
	case STRUCT_DECL: {
	  ty = T_STRUCT;
	  break;
	}
	case UNION_DECL: {
	  ty = T_UNION;
	  break;
	}
	}
	ctype = new brk_aType(st->symbol()->identifier(), ty); 
	
#if DEBUG
	printf("defining new field %s for comp %s \n", curSym->identifier(),
	       st->symbol()->identifier());
#endif
	
	ftype = brk_createType(curSym->type());
	
	// determine if it is private/public/protected...;
	flag = curSym->attributes();
	
	if (flag & PRIVATE_FIELD) {
	  ftype->mt = PRIVATE_VAR;
	}
	if (flag & PUBLIC_FIELD) {
	  ftype->mt = PUBLIC_VAR;
	}
	if (flag & PROTECTED_FIELD) {
	  ftype->mt = PROTECTED_VAR;
	}
	brk_addFieldToComposite(ctype, curSym->identifier(), ftype);
	
	// ---------------------------------------------;
      }
    }
    // get next symbol in table.;
    curSym = curSym->next();
  }
  
  brk_build_coll_types(aFile);
  
  // now, add access func to element classes...;
  brk_addAccessFuncsToList((brk_sCompositeList *) brk_elementTypeLIST, (aFile), 
		       "brk_accessFunc");
  
#if DEBUG 
  printf("Built types from file.\n");
#endif
  
}


brk_aComposite *brk_getOrMakeComposite(brk_basicAType *ctype) {

  brk_compositeList *cList = NULL;
  brk_sComposite *comp;

  switch (ctype->thisType) {

  case T_CLASS : {
    cList = brk_classLIST;
    break;
  }
  case T_COLLECTION :  {
    cList = brk_collectionLIST;
    break;
  }

  case T_STRUCT :  {
    cList = brk_structLIST;
    break;
  }

  case T_UNION :  {
    cList = brk_unionLIST;
    break;
  }
  default: {
    printf("trying to add field to non composite: %d \n", ctype->thisType);
    break;
  }
  }
  
  if (cList) {
    // search for the composite this new field belongs to...;
    comp = (brk_sComposite *) cList->searchList();
    while (comp) {
      if (brk_strsame(comp->name, ctype->description)) {
	break;
      }
      comp =  (brk_sComposite *) cList->nextItem();
    }
    
    // if not found, create the new one...;
    if (!comp) {
      cList->push(comp = new brk_sComposite(ctype->description, 
	ctype->thisType));
    }
  }
  return comp;
}

// ----------------------------------;
void brk_addFieldToComposite(brk_basicAType *ctype,
			 char *fieldName, brk_basicAType *fieldType) {
  brk_aComposite *comp ;

  comp = brk_getOrMakeComposite(ctype);

  if (comp) {
    // add field to comp;
    switch (fieldType->thisType) {
    case T_ARRAY:
    case T_POINTER: {
      comp->push(new brk_pField((brk_aType *) fieldType, fieldName));
      break;
    }
    default: {
      comp->push(new brk_sField(fieldType, fieldName));
      break;
    }
    }
  } else {
    printf ("No composite type for variant %d\n", ctype->thisType);
  }
}


// ---------------------------------------------------
void brk_determineSuperClasses(classHierarchy *ch) {

  classHierarchy *cl;
  brk_aComposite *ac;
  brk_basicAType *at;
  int t1;

  if (!ch) return;

  cl = (classHierarchy *) ch->children->searchList();
  
  fflush(stdout);

  while (cl) {
    // depth first;

    brk_determineSuperClasses(cl);

    if (cl->isCollection()) {
      t1 = T_COLLECTION;
    } else {
      t1 = T_CLASS;
    }

    at = new brk_basicAType(cl->className, t1);

    ac = brk_lookupComposite(at);
    
    if (!ac) {
      printf("didnt find composite %s\n", cl->className);
      // if not found, create the new one...;
      if (t1 == T_CLASS) {
	brk_classLIST->push(ac = new brk_sComposite(cl->className, t1));
      } else {
	brk_collectionLIST->push(ac = new brk_sComposite(cl->className, t1));
      }
    }
    

    if (ac->superClasses->length == 0) {
      // might not have searched this class yet.;
      
      cl = (classHierarchy *) cl->parents->searchList();

      while (cl) {
	ac->superClasses->push(new brk_myStr(cl->className));
	cl = (classHierarchy *) cl->parents->nextItem();
      }

    }

    cl = (classHierarchy *) ch->children->nextItem();
      
  }
}
    
//  ##########################################################
//  brk_aType class methods

brk_aType::brk_aType(SgType *sgty) {
  SgSymbol *s;
  sageType = sgty;
  if ((s = sageType->symbol()) && (s->identifier())) {
    description = brk_stringSave(s->identifier());
  } else {
    description = "";
  }
  thisType = sageType->variant();
}

brk_aType::brk_aType() {
}


 brk_aType::brk_aType(char *desc, int ty) {
   description = brk_stringSave(desc);
   thisType = ty;
 }



//  ##########################################################
//  brk_pType class methods

brk_pType::brk_pType(SgType *sgty) {

  SgExpression *se;
  char *buf;
  sageType = sgty;
  SgArrayType *at;
  SgSymbol *s;
  
  if((s = sageType->symbol()) && (buf = s->identifier())) {
    description = brk_stringSave(buf);
  } else {
    description = NULL;
  }
  
  thisType = sageType->variant();
  
  switch (thisType) {
  case T_ARRAY: {
    
    // determine base type;
    baseType = brk_createType(sgty->baseType());
    
    // set dimension parameters...;
    at = (SgArrayType *) sgty;
    dimensions = at->dimension();
    sizes = new int[at->dimension()];
    totalSize = 1;
    for (int i = 0; i < dimensions; i++) {
      se = at->sizeInDim(i);
      buf =  unparseExp(se);
      sizes[i] = evalStr(buf);
      totalSize *= sizes[i];
      free(buf);
    }
    break;
  }
  case T_POINTER: {
    // determine base type;
    baseType = brk_createType(sgty->baseType());
    break;
  }
  case T_DERIVED_TYPE: {
    // determine base type;
    if (sgty->symbol()->type()) {
      baseType = brk_createType(sgty->symbol()->type());
    } else {
      baseType = NULL;
    }
    description = brk_stringSave(sgty->symbol()->identifier());
    break;
  }
  }
}


// ########################################################
// brk_pField methods...;
// for fields that are pointers to other structures


// constructor for brk_pField class;
 brk_pField::brk_pField ( brk_aType *t, char * n ) {
   fieldType = t;
   name = brk_stringSave(n);
 }


// this funciton generates the sage code to access a pointer field.
// 
void brk_pField::writeFieldAccess(SgFile *f, brk_aComposite *parent) {

  brk_sComposite *comp;
  char *buf1, *tmpBuf1;
  SgStatement *st, *tmps, *lst, *st2;
  SgSymbol *bufsym, *fs, *ptrsym, *loopvarsym;
  SgPointerType *ptrType;
  SgExpression *fexp, *fexp2, *nullexp;
  SgType *ty, *charPtr, *voidTy, *charTy;
  SgValueExp *ve;
  brk_pType *at;
  SgArrayType *sat, *arrType;
  SgFuncHedrStmt *fhs;
  brk_aType *atype;
  brk_pType *typeptr;
  SgArrayRefExp *arrRef;
  // SgVariableSymb *arrVar;
  SgSymbol *arrVar;
  int i = 0;


  buf1 = new char[100];
  typeptr = (brk_pType *) fieldType;

  at = (brk_pType *) fieldType;
  charPtr = new SgPointerType(*(new SgType(T_CHAR)));
  charTy = new SgType(T_CHAR);
  voidTy = new SgType(T_VOID);

  sprintf(buf1, "brk_accessFunc_%s", name);

  // add func to class
  tmps = findCompDeclStmt(parent->name, f);
  addFunctionToClass(tmps, voidTy, buf1, PUBLIC_FIELD);

  // define new funct
  fs = findVarDeclSymbFromStmt(buf1, tmps);
  fhs = (SgFuncHedrStmt *) buildFuncDefFromSymb(fs);

  st = lastStatement(f);
  st->insertStmtAfter(*fhs, *(f->firstStatement()));

  // add array type to func;
  int arrayDims[1];
  arrayDims[0] = 100;
  sat = (SgArrayType *) buildArrayType(1, arrayDims, charTy);
  
  // add array type to func;
  bufsym = declareVarInFunc(fhs, sat, "buf");

  st = findVarDeclStmt("buf", fhs);
  fexp = new SgExprListExp(*(new 
			     SgArrayRefExp(*bufsym,
					       *(new SgValueExp(100)))));
  st->setExpression(0, *fexp);

  // ---------------------------------;
  
  st = fhs->lastDeclaration();
  

  // this is to take care of typedefs..

  SgDerivedType *dt;
  if (dt = isSgDerivedType(at->sageType)) {
    // look at the basetype;
    //     at->sageType = 
    at = new brk_pType(&(dt->typeName()->type()->copy()));
  }

  // more stuff to take care of typedefs..
  if (at->sageType->baseType()) {
    SgDescriptType *dty;
    if ((dty = isSgDescriptType(at->sageType->baseType())) &&
	(dty->modifierFlag() == BIT_TYPEDEF))  {
      if (arrType = isSgArrayType(at->sageType)) {
	arrType->setBaseType(*(dty->baseType()));
      } else if (ptrType = isSgPointerType(at->sageType)) {
	ptrType->setBaseType(*(dty->baseType()));
      }
      
    }
  }
  
  if (T_ARRAY == at->thisType) {
    
    // ---------------------------------;
    
    // assign null to strings..;
    nullexp = new SgValueExp((int) NULL);
    
    if (parent->thisType == T_STRUCT) {
      // "  %s * ptr;\n", typeToStr(at->baseType->thisType);
      ty = new SgPointerType(*(new SgType(at->baseType->thisType)));
      ptrsym = declareVarInFunc(fhs, ty, "ptr");
      // must be a struct ;
      // "  ptr = castedPtr->%s;\n", name;
    }    
    
    // "  for (int i = 0; i < %d; i++) { \n", totalSize;
    // "    sprintf(buf, \"%s#\", *ptr);",  brk_typeToFormat(at->baseType) ;
    // "    BPstrcat(returnStr, buf);";
    // "    ptr++;\n";
    // "  }\n";
    
    char *loopstr;
    st2 = NULL;

    arrType = (SgArrayType *) &(at->sageType->copy());
    arrVar = declareVarInFunc(fhs, arrType, "arrVar");
    arrRef = new SgArrayRefExp(*arrVar);
    SgForStmt *fs, *fs2;

    tmps = NULL;
    for (i =  0; i < at->dimensions; i++) {
      loopstr = new char[3];
      sprintf(loopstr, "i%d", i);
      loopvarsym = declareVarInFunc(fhs, new SgType(T_INT), loopstr);
      fs = (SgForStmt *) buildForStmt(0, at->sizes[i], loopvarsym, NULL);
      if (tmps) {
	fs2 = (SgForStmt *) tmps;
	fs2->set_body(*fs);
	tmps = (SgStatement *) fs;
      } else {
	tmps = (SgStatement *) fs;
	st->insertStmtAfter(*tmps);
      }
      st = tmps;
      arrRef->addSubscript(*(new SgVarRefExp(*loopvarsym)));
    }

    // add a function call ;
    lst = buildFuncCallStmt("sprintf", f);
    // add first parameter;
    addParamToFuncCall("buf", charTy, lst);
    // add ac constant value parameter;
    sprintf(buf1, "%s#", brk_typeToFormat(typeptr->baseType));
    ve = new SgValueExp( brk_stringSave(buf1));
    fexp = extractFuncExpFromHedr(lst);  
    addParamExpToFuncExp(fexp, ve);
    // add last parameter;
    addParamExpToFuncExp(fexp, arrRef);
    // replace body..;
    fs->set_body(*lst);
    st = lst;


    // add another function call ;
    lst = buildFuncCallStmt("brk_writeSocket", f);
    // add first parameter;
    addParamToFuncCall("buf", charTy, lst);
    st2 = st->controlParent(); /*BM ???*/
    st->insertStmtAfter(*lst, *st2);
    tmps = lst;
    st = lst;
    
  } 
  
  atype = (brk_aType *) at->baseType;
  if (T_POINTER == at->thisType) {
    if ((atype->thisType == T_STRUCT) ||
	(atype->thisType == T_CLASS)) {

      comp = (brk_sComposite *) brk_findInCompositeList(atype->description, 
						NULL, brk_elementTypeLIST);
      comp->writeAccessFields(f, fhs);

    } else {

      // ---------------------------------;
      // need space for name and value;

      tmpBuf1 = brk_typeToFormat(atype);
      
      if (brk_strsame(tmpBuf1, "follow")) {
	// need to follow this, because it is a composite itself;
      } else if ((brk_strsame(tmpBuf1, "unknown")) || 
		 (brk_strsame(tmpBuf1, "follow"))) {
	// this is not a member you can look at???, might be a problem...;
	printf("this member:\n");
	printf("of type: \n");
	printf("cannot be accesses\n----------------------\n\n");
      } else {


	// add a function call ;
	lst = buildFuncCallStmt("brk_writeSocket", f);
	// add first parameter;
	fexp = extractFuncExpFromHedr(lst);  
	sprintf(buf1, "#%s#", name);
	ve = new SgValueExp(brk_stringSave(buf1));
	addParamExpToFuncExp(fexp, ve);
	// add last parameter;
	st->insertStmtAfter(*lst);
	st = lst;
	
	// now do different things whether it is a struct or a class;
	if (parent->thisType == T_CLASS) {
	  
	  // add a function call ;
	  lst = buildFuncCallStmt("sprintf", f);
	  // add first parameter;
	  addParamToFuncCall("buf", charTy, lst);
	  // add ac constant value parameter;
	  sprintf(buf1, "%s", tmpBuf1);
	  ve = new SgValueExp(brk_stringSave(buf1));
	  fexp = extractFuncExpFromHedr(lst);  
	  addParamExpToFuncExp(fexp, ve);
	  // add last parameter;
	  ty = new SgType(atype->thisType);
	  tmps = findCompDeclStmt(parent->name, f);
	  fs = findVarDeclSymbFromStmt(name, tmps);
	  fexp2 = &(SgDerefOp(*(new SgVarRefExp(*fs))));
	  addParamExpToFuncExp(fexp, fexp2);
	  st->insertStmtAfter(*lst);
	  st = lst;

	} 
	//  check to see if parent is a struct..;
	if (parent->thisType == T_STRUCT) {
	  // must be a struct ;
	  printf("SHOULDNT BE HERE");
	  // add a function call ;
	  lst = buildFuncCallStmt("sprintf", f);
	  // add first parameter;
	  addParamToFuncCall("buf", charTy, lst);
	  // add a constant value parameter;
	  sprintf(buf1, "%s", brk_typeToFormat(atype));
	  ve = new SgValueExp( brk_stringSave(buf1));
	  fexp = extractFuncExpFromHedr(lst);  
	  addParamExpToFuncExp(fexp, ve);
	  // add last parameter;
	  fexp2 = buildMembFuncCallExp(parent->name,
				       name,
				       f,
				       searchForParamInFuncStmt(fhs, "parent")
				       );
	  addParamExpToFuncExp(fexp, fexp2);
	  st->insertStmtAfter(*lst);
	  st = lst;
	}
	
	// put buf in return string;
	// add another function call ;
	lst = buildFuncCallStmt("brk_writeSocket", f);
	// add first parameter;
	addParamToFuncCall("buf", charTy, lst);
	st->insertStmtAfter(*lst);
	st = lst;
	fflush(stdout);
      }
    }
  }

#if DEBUG
  printf("written field access func...\n");
  fhs->unparsestdout();
#endif
}

SgStatement *brk_sField::addPtrAccessCond(SgFile *f, brk_sComposite *parent, 
				      SgStatement *fhs) {

  SgStatement *st, *lst, *tmps;
  SgExpression *fexp, *fexp2;
  SgType *charPtr;
  char buf[100];

  charPtr = new SgPointerType(*(new SgType(T_CHAR)));
  fexp2 = new SgValueExp(name);
  fexp = buildFuncCallExp("brk_strsame", f);
  addParamExpToFuncExp(fexp, fexp2);
  
  addParamToFuncExp(fexp, new SgVariableSymb("fieldName", *charPtr));

  *buf = NULL;
  strcpy(buf, "brk_accessFunc_");
  strcat(buf, name);

  tmps = buildThisMembCallStmt(parent->name,
			       buf,
			       f);

  lst = new SgIfStmt(*fexp, *tmps);
  st = fhs->lastDeclaration();
  st->insertStmtAfter(*lst);
  return lst;
}


SgStatement *brk_sField::writePtrAccessFunc(SgFile *f, brk_sComposite *parent) {

  char *buf1;
  SgStatement *st, *tmps;
  SgSymbol *fs;
  SgExpression *fexp, *fexp2;
  SgType *charPtr, *voidTy;
  brk_pType *at;
  SgFuncHedrStmt *fhs;
  SgVariableSymb *vs;

  buf1 = new char[100];

  at = (brk_pType *) fieldType;
  voidTy = new SgType(T_VOID);
  charPtr = new SgPointerType(*(new SgType(T_CHAR)));

  sprintf(buf1, "brk_accessFunc", name);
			      
  tmps = findCompDeclStmt(parent->name, f);
  fexp = addFunctionToClass(tmps, voidTy, buf1, PUBLIC_FIELD);
  vs = new SgVariableSymb("fieldName", *charPtr);
  addParamToFuncExp(fexp, vs);
  
  fs = findVarDeclSymbFromStmt(buf1, tmps);
  
  fhs = (SgFuncHedrStmt *) buildFuncDefFromSymb(fs);

  st = lastStatement(f);
  st->insertStmtAfter(*fhs, *(f->firstStatement()));

  // add parameters to func hedr;
  fexp2 = addParamToFuncStmt("fieldName", charPtr, fhs);

#if DEBUG
  printf("writing access func\n");
  fhs->unparsestdout();
#endif

  return fhs;
}


// ########################################################
// brk_sComposite methods...;



 brk_sComposite::brk_sComposite(char * aName, SgStatement *ds) {
  name = brk_stringSave(aName);
  superClasses = new brk_linkedList();
  declStmt = ds;
}
 brk_sComposite::brk_sComposite(char * aName, int t, SgStatement *ds) {
   thisType = t;
   name = brk_stringSave(aName);
   superClasses = new brk_linkedList();
   declStmt = ds;
 }


 brk_sComposite::brk_sComposite(char * aName) {
   name = brk_stringSave(aName);
   superClasses = new brk_linkedList();
 }

 brk_sComposite::brk_sComposite(char * aName, int t) {
   thisType = t;
   name = brk_stringSave(aName);
   superClasses = new brk_linkedList();
 }

int isPtrType(brk_basicAType *ty) {

  if ((ty->thisType == T_POINTER) ||
	(ty->thisType == T_ARRAY)) {
    return 1;
  }
  if (ty->thisType == T_DERIVED_TYPE) {
    brk_basicPType *pt;
    pt = (brk_basicPType *) ty;
    return isPtrType(pt->baseType);
  }
  return 0;
}

// --------------------------------------------;
// this function is only called for structs and classes...;
SgStatement * brk_sComposite::writeAccessFields(SgFile *f, SgStatement *fhdr) {
  char *tmpBuf1, *buf1;
  brk_sField *fld;
  brk_basicAType *ty2;
  SgStatement *st, *lst;
  SgSymbol *bufsym;
  SgPointerType *pt;
  SgExpression *fexp, *fexp2, *nullexp;
  SgType *ty, *charPtr, *charTy;
  SgValueExp *ve;
  SgSymbol *brk_basicsym, *castsym, *objsym;

  st = fhdr;

  buf1 = new char[100];

  if (thisType == T_STRUCT) {
    
    // pointer declarations necessary for structs ;

    // find the correct class symbol;
    pt = new 
      SgPointerType(*(buildDerivedType("brk_basicAType", f)));
    brk_basicsym = declareVarInFunc(fhdr, pt, "ty");
    
    // and build a pointer to it;
    pt = new 
      SgPointerType(*(buildDerivedType(name, f)));
    
    // and add a variable to the function;
    castsym = declareVarInFunc(fhdr, pt, "castedPtr");

    // -------------------------------------;
    
    // cast objPtr to the casted ptr;
    
    st = fhdr->lastDeclaration();
    objsym = searchForParamInFuncStmt(fhdr, "objPtr");
    lst  = composeCastAssignStatement(castsym, objsym);
    st->insertStmtAfter(*lst);
    st = lst;
    
  }
  
  charPtr = new SgPointerType(* new SgType(T_CHAR));
  charTy = new SgType(T_CHAR);

  // other declarations, necessary for classes and structs...;
  // ----------------------------------------------;
  // other declarations...;
  
  // add array type to func;
  int arrayDims[1];
  arrayDims[0] = 100;
  ty = buildArrayType(1, arrayDims, charTy);
  
  // add array type to func;
  bufsym = declareVarInFunc(fhdr, ty, "buf");
  st = findVarDeclStmt("buf", fhdr);
#if DEBUG
  printf("before my modifications:\n");
  st->unparsestdout();
#endif
  fexp = new SgExprListExp(*(new 
			     SgArrayRefExp(*bufsym,
					   *(new SgValueExp(100)))));
  st->setExpression(0, *fexp);
#if DEBUG
  printf("after my modifications:\n");
  st->unparsestdout();
#endif

  // -------------------------------------;
  // Executable statements;

  st = fhdr->lastDeclaration();


  // assign null to strings..;
  nullexp = new SgValueExp((int) NULL);
  lst = buildAssignStmt(&(SgDerefOp(*(new SgVarRefExp(*bufsym)))), nullexp);
  st->insertStmtAfter(*lst);
  st = lst;

  
  // add another function call ;
  lst = buildFuncCallStmt("brk_writeSocket", f);
  // add first parameter;
  fexp = extractFuncExpFromHedr(lst);  
  ve = new SgValueExp("$");
  addParamExpToFuncExp(fexp, ve);
  st->insertStmtAfter(*lst);
  st = lst;


  // now, loop through the fields;
  fld = (brk_sField *) searchList();
  while (fld) {
    ty2 = fld->fieldType;
    
    // will ned to change for specific field types, thi;
    // one only works for char *;
    tmpBuf1 = brk_typeToFormat(ty2);
    if (brk_strsame(tmpBuf1, "follow")) {
      // need to follow this, because it is a composite itself;
    } else if ((brk_strsame(tmpBuf1, "unknown")) || (brk_strsame(tmpBuf1, "follow"))) {
      // this is not a member you can look at???, might be a problem...;
      printf("this member:\n");
      fld->print(1);
      printf("of type: \n");
      ty2->print(1);
      printf("cannot be accesses\n----------------------\n\n");
    } else {
      // add a function call ;
      lst = buildFuncCallStmt("brk_writeSocket", f);
      // add first parameter;
      fexp = extractFuncExpFromHedr(lst);  
      sprintf(buf1, "#%s#", fld->name);
      ve = new SgValueExp(brk_stringSave(buf1));
      addParamExpToFuncExp(fexp, ve);
      // add last parameter;
      st->insertStmtAfter(*lst);
      st = lst;
      
      // now do different things whether it is a struct or a class;
      if (thisType == T_CLASS) {
	
	// add a function call ;
	lst = buildFuncCallStmt("sprintf", f);
	// add first parameter;
	addParamToFuncCall("buf", charPtr, lst);
	// add ac constant value parameter;
	sprintf(buf1, "%s", brk_typeToFormat(ty2));
	
	ve = new SgValueExp( brk_stringSave(buf1));
	fexp = extractFuncExpFromHedr(lst);  
	addParamExpToFuncExp(fexp, ve);
	
	// add last parameter;
	ty = new SgType(ty2->thisType);
	addParamToFuncCall(fld->name, ty, lst);
	st->insertStmtAfter(*lst);
	st = lst;
	
      } else {
	// must be a struct ;
	
	// add a function call ;
	lst = buildFuncCallStmt("sprintf", f);
	// add first parameter;
	addParamToFuncCall("buf", charPtr, lst);
	// add a constant value parameter;
	sprintf(buf1, "%s", brk_typeToFormat(ty2));
	ve = new SgValueExp( brk_stringSave(buf1));
	fexp = extractFuncExpFromHedr(lst);  
	addParamExpToFuncExp(fexp, ve);
	// add last parameter;
	fexp2 = buildMembFuncCallExp(name,
				     fld->name,
				     f,
				     searchForParamInFuncStmt(fhdr, "parent")
				     );
	addParamExpToFuncExp(fexp, fexp2);
	st->insertStmtAfter(*lst);
	st = lst;
	
      }
      
      // put buf in return string;
      
      // add another function call ;
      lst = buildFuncCallStmt("brk_writeSocket", f);
      // add first parameter;
      addParamToFuncCall("buf", charPtr, lst);
      st->insertStmtAfter(*lst);
      st = lst;
      
    }
    fld = (brk_sField *) nextItem();
  }

  // add another function call ;
  lst = buildFuncCallStmt("brk_writeSocket", f);
  // add first parameter;
  fexp = extractFuncExpFromHedr(lst);  
  ve = new SgValueExp("$");
  addParamExpToFuncExp(fexp, ve);
  // add statement;
  st->insertStmtAfter(*lst);
  st = lst;

  // ------------------------------------------------------------
  // Now, make access functions for each pointer in this composite;
  SgFuncHedrStmt *tfhs;
  brk_pField *af;
  int done = 0;
  af = (brk_pField *) searchList();

  while (af) {
    if (isPtrType(af->fieldType)) {
      // write another acess function for this pointer;

      if (!done) {
	// lets create the function for parsing these pointer,
	// fields.;
	tfhs = (SgFuncHedrStmt *) af->writePtrAccessFunc(f, this);
	done++;
      }

#if DEBUG      
      printf("writing access function for %s\n", af->name);
#endif
      
      af->writeFieldAccess(f, this);
      af->addPtrAccessCond(f, this, tfhs);
    }
    af = (brk_pField *) nextItem();
  }

#if DEBUG      
      printf("done writing field access functions for %s\n", name);
#endif

  return st;
}

// --------------------------------------------;
SgStatement * brk_sComposite::writeAccessFunction(SgFile *f) {

  char *buf1;
  brk_sComposite *comp = NULL;
  SgStatement *st, *tmps, *lst, *tmps2;
  SgSymbol *vs, *bufsym, *castsym, *elemsym, *objsym, *fs, 
        *elemnumsym, *fsym;
  SgPointerType *pt;
  SgExpression *fexp, *fexp2, *nullexp;
  SgFuncHedrStmt *fhs;
  SgType *charPtr, *voidTy, *elemType, *intty;
  SgValueExp *ve;
  SgArrayType *at;
  brk_basicPType *bpt;
  brk_aField *af;
  int arrayDims[1];

  // object pointer declaration and cast;
  
  buf1 = new char[100];

  switch (thisType) {
  case T_COLLECTION: {
    //  Must be collection ;
    if (elementType) {
      //  defined element type (necessary)...;
 
      // ----------------------------------------------
      // function declaration with argument;
      sprintf(buf1,"brk_AccessFunc%s__%s", name, elementType);
      charPtr = new SgPointerType(*(new SgType(T_CHAR)));
      voidTy = new SgType(T_VOID);
      
      // st = lastStatement(f);

      fhs = buildFuncDefStmt(buf1, voidTy);

      st = lastStatement(f);
      st->insertStmtAfter(*fhs, *(f->firstStatement()));
       
      // add parameters to func hedr;
      pt = new SgPointerType(*(new SgType(T_VOID)));
      fexp2 = addParamToFuncStmt("objPtr", pt, fhs);
      intty = new SgType(T_INT);
      // element number parameter;
      fexp2 = addParamToFuncStmt("elemNum", intty, fhs);
      // the field name (could be NULL);
      fexp2 = addParamToFuncStmt("fname", charPtr, fhs);
      

      // ----------------------------------------------
      // declarations...;

      comp = (brk_sComposite *) 
	brk_lookupComposite(new 
			brk_basicAType(elementType, T_UNKNOWN));
      int maxspace = 1;
      af = (brk_aField *) comp->searchList();
      while (af) {
	if (af->fieldType->thisType == T_ARRAY) {
	  bpt = (brk_basicPType *) af->fieldType;
	  if (bpt->totalSize > maxspace) {
	    maxspace = bpt->totalSize;
	  }
	}
	af = (brk_aField *) comp->nextItem();
      }
    
      arrayDims[0] = 100;
      at = (SgArrayType *) buildArrayType(1, arrayDims, charPtr);
      
      // add array type to func;
      bufsym = declareVarInFunc(fhs, at, "buf");

      st = findVarDeclStmt("buf", fhs);
      fexp = new SgExprListExp(*(new 
				 SgArrayRefExp(*bufsym,
					       *(new SgValueExp(100)))));
      st->setExpression(0, *fexp);
    
      // must be a collection, so declare the pointer specially;

      // find the correct collection symbol;
      vs = findCollectionSymbol(name, f);
      
      // and build a collection pointer;
      elemType = new SgType(brk_TIstrToType(elementType));
      if (elemType->variant() == T_UNKNOWN) {
	elemType = buildDerivedType(elementType, f);
      }
      pt = new 
	SgPointerType(*(buildDerivedCollectionType(elemType,
						   vs->identifier(),
						   vs->type())));
      
      // and add a variable to the function;
      castsym = declareVarInFunc(fhs, pt, "castedPtr");
      
      elemsym = declareVarInFunc(fhs, new SgPointerType(*elemType), 
				 "elementPtr");



      // -------------------------------------;

      st = fhs->lastDeclaration();

      // ---------------------------------;

      // cast objPtr to the casted ptr;

      objsym = searchForParamInFuncStmt(fhs, "objPtr");
      lst  = composeCastAssignStatement(castsym, objsym);
      st->insertStmtAfter(*lst);
      st = lst;

      // assign null to strings..;
      nullexp = new SgValueExp((int) NULL);
      lst = buildAssignStmt(&(SgDerefOp(*(new SgVarRefExp(*bufsym)))), 
					  nullexp);
      st->insertStmtAfter(*lst);
      st = lst;
      

      // retrieve element;
      elemnumsym = searchForParamInFuncStmt(fhs, "elemNum");
      fexp = buildCollGetElemExp(&(SgDerefOp(*(new SgVarRefExp(*castsym)))), 
				 new SgVarRefExp(*elemnumsym));
      fexp2 = new SgVarRefExp(*elemsym);
      lst = buildAssignStmt(fexp2, fexp);
      st->insertStmtAfter(*lst);
      st = lst;

      // add a function call ;
      lst = buildFuncCallStmt("sprintf", f);
      // add first parameter;
      addParamToFuncCall("buf", charPtr, lst);
      // add a constant value parameter;
      fexp = extractFuncExpFromHedr(lst);  
      sprintf(buf1, "@element(%s)", "%d");
      ve = new SgValueExp(brk_stringSave(buf1));
      addParamExpToFuncExp(fexp, ve);
      // add last parameter;
      addParamToFuncCall("elemNum", intty, lst);
      st->insertStmtAfter(*lst);
      st = lst;

      
      // add another function call ;
      lst = buildFuncCallStmt("brk_writeSocket", f);
      // add first parameter;
      addParamToFuncCall("buf", charPtr, lst);
      st->insertStmtAfter(*lst);
      st = lst;

      // write each member access ;
      // check to see if type is a composite...;
      if (!comp) {
      
	// add a function call ;
	lst = buildFuncCallStmt("sprintf", f);
	// add first parameter;
	addParamToFuncCall("buf", charPtr, lst);
	// add a constant value parameter;
	fexp = extractFuncExpFromHedr(lst);  
	sprintf(buf1, "#%s#", 
		brk_typeToFormat(new brk_basicAType("", brk_TIstrToType(elementType))));
	ve = new SgValueExp(brk_stringSave(buf1));
	addParamExpToFuncExp(fexp, ve);
	// add last parameter;
	addParamToFuncCall("elementPtr", 
			   new SgPointerType(*elemType), lst);
	st->insertStmtAfter(*lst);
	st = lst;

      } else {
	// it is a composite, so call the appropriate;
	// access function;
	
	// only of pointer field exists in element composite 
	// do we use an if statement;

	if (comp->hasPointer) {
	  fsym = searchForParamInFuncStmt(fhs, "fname");
	  
	  // compose true branch of if statement;
	  // add last parameter, an expression;
	  tmps = buildMembFuncCallStmt(elementType,
				       "brk_accessFunc",
				       f,
				       elemsym);
	  fexp2 = extractFuncExpFromHedr(tmps);  
	  // get func_call out of pointst exp;
	  fexp = fexp2->rhs();
	  addParamExpToFuncExp(fexp, new SgVarRefExp(*(fsym)));
	  // build assing statement, 
	  
	  // compose false branch of if statement;
	  // add last parameter, an expression;
	  tmps2  = buildMembFuncCallStmt(elementType,
				       "brk_accessFunc",
				       f,
					elemsym);
	  
	  fexp = &(SgDerefOp(*(new SgVarRefExp(*fsym))));
	  lst = new SgIfStmt(*fexp,
			     *tmps,
			     *tmps2);
	  st->insertStmtAfter(*lst);
	  st = lst;
	} else {
	  lst = buildMembFuncCallStmt(elementType,
				       "brk_accessFunc",
				       f,
				       elemsym);
	  st->insertStmtAfter(*lst);
	  st = lst;
	}
      }
    }

    // add another function call ;
    lst = buildFuncCallStmt("brk_writeSocket", f);
    // add first parameter;
    fexp = extractFuncExpFromHedr(lst);  
    sprintf(buf1, "@");
    ve = new SgValueExp(brk_stringSave(buf1));
    addParamExpToFuncExp(fexp, ve);
    // add statement
    st->insertStmtAfter(*lst);
    st = lst;
    // should be last bif node in file...;
    break;
  }
  case T_CLASS: {
    //  A class...;
    // function declaration with argument;
    
    // declare a member function;
    // st = f->firstStatement();

    // find the correct collection symbol;
    tmps = findCompDeclStmt(name, f);
    fs = findVarDeclSymbFromStmt("brk_accessFunc", tmps);

    // void Vector:: accessFunc() {
    fhs = (SgFuncHedrStmt *) buildFuncDefFromSymb(fs);
    
    st = lastStatement(f);
    st->insertStmtAfter(*fhs, *(f->firstStatement()));

    //
    comp = (brk_sComposite *) brk_lookupComposite(new brk_basicAType(name, thisType));
    st = (SgStatement *) comp->writeAccessFields(f, fhs);
    break;
  } 
  case T_STRUCT: {
    // ----------------------------------------------;
    // function declaration with argument;
    sprintf(buf1,"brk_AccessFunc%s", name);
    charPtr = new SgPointerType(*(new SgType(T_CHAR)));
    
    fhs = buildFuncDefStmt(buf1, new SgType(T_VOID));

    st = lastStatement(f);
    st->insertStmtAfter(*fhs, *(f->firstStatement()));
    
    // add a parameter;
    pt = new SgPointerType(*(new SgType(T_VOID)));
    fexp2 = addParamToFuncStmt("objPtr", pt, fhs);
    
    
    comp =  (brk_sComposite *) brk_lookupComposite(new brk_basicAType(name, thisType));
    st = (SgStatement *) comp->writeAccessFields(f, fhs);
    break;
  } 
  } // end of switch (thisType);

#if DEBUG
  printf("written access function\n");
  fhs->unparsestdout();
#endif
  return fhs;
}


// --------------------------------
brk_sCompositeList::brk_sCompositeList() {
}


// --------------------------------
brk_sCompositeList::brk_sCompositeList(char *ety) {
  elementType = brk_stringSave(ety);
}


// --------------------------------
void brk_sCompositeList::writeAccessEntries(SgFile *f) {

  brk_aComposite *c;

  char *buf1;
  SgStatement *st, *lst;
  SgSymbol *vs, *compsym, *bsym, *fs;
  SgExpression *fexp, *fexp2, *fexp3;
  SgFuncHedrStmt *fhs;
  SgType *bType, *compType;
  SgValueExp *ve;
  char complistName[30];

#if DEBUG      
  printf("started writing field access entries\n");
  fflush(stdout);
#endif

 // object pointer declaration and cast;
  
  buf1 = new char[100];

  // build name of data structure.;
  strcpy(complistName, elementType);
  strcat(complistName, "LIST");


  // put table entries in init function;
  
  // ----------------------------------------------;
  // function declaration with argument;
  sprintf(buf1, "brk_%sAccessFunctionEntryInit",
	  elementType);
  
  fhs = buildFuncDefStmt(buf1, new SgType(T_VOID));
  
  st = lastStatement(f);
  st->insertStmtAfter(*fhs, *(f->firstStatement()));
  
  
  compType = buildDerivedType("brk_aComposite", f);
  // and add a variable to the function;
  compsym = declareVarInFunc(fhs, new SgPointerType(*compType), 
			     "comp");

  bType = buildDerivedType("brk_basicAType", f);
  // and add a variable to the function;
  bsym = declareVarInFunc(fhs, new SgPointerType(*bType), 
			  "ty");

  st = fhs->lastDeclaration();
  
  c = (brk_aComposite *) searchList();
  while (c) {
 
#if DEBUG      
    printf("started writing field access entries specifically for %s\n", 
	   c->name);
    fflush(stdout);
#endif
    // make certain the name is not null;
    if (brk_strsame(c->name, "")) {
      // probably a struct with no name, print it;
      printf(" this composite did not have a name (table init).\n");
      c->print(0);
      perror ("should this happen???\n");
    } else if (c->thisType == T_COLLECTION && c->elementType) {
 
#if DEBUG      
      printf("\n\tlooking at a collection with and element type %s\n",
	     c->elementType);
      fflush(stdout);
#endif
      
      // this must be a collection and it has an element type...;
      
      // add a function call ;
      lst = buildFuncCallStmt("brk_addCollectionWithElem", f);
      // add first parameter;
      // add a constant value parameter;
      fexp = extractFuncExpFromHedr(lst);  
      ve = new SgValueExp(brk_stringSave(c->name));
      addParamExpToFuncExp(fexp, ve);
      // add last parameter;
      ve = new SgValueExp(brk_stringSave(c->elementType));
      addParamExpToFuncExp(fexp, ve);
      st->insertStmtAfter(*lst);
      st = lst;

      fexp = new SgVarRefExp(*compsym);
      fexp2 = buildFuncCallExp("brk_findInCompositeList", f);
      // add first parameter;
      // add a constant value parameter;
      ve = new SgValueExp(brk_stringSave(c->name));
      addParamExpToFuncExp(fexp2, ve);
      // add parameter;
      ve = new SgValueExp(brk_stringSave(c->elementType));
      addParamExpToFuncExp(fexp2, ve);
  
#if DEBUG      
      printf("\n\tsection 1\n");
      fflush(stdout);
#endif
      // add parameter;
      vs = findVarDeclSymb("brk_collectionElemLIST", f);

#if DEBUG      
      printf("\n\tsection 2\n");
      fflush(stdout);
#endif

      fexp3 = new SgVarRefExp(*vs);
      addParamExpToFuncExp(fexp2, fexp3);


#if DEBUG      
      printf("\n\tsection 3\n");
      fflush(stdout);
#endif

 
      lst = buildAssignStmt(fexp, fexp2);
      st->insertStmtAfter(*lst);
      st = lst;


      // comp->accessFunc = AccessFuncDistVector__Vector ;
      lst = findCompDeclStmt("brk_aComposite", f);
      fs = findVarDeclSymbFromStmt("accessFunc", lst);
		
      fexp = &(SgPointStOp(*(new SgVarRefExp(*compsym)),
			 *(new SgVarRefExp(*fs))));
      sprintf(buf1, "brk_AccessFunc%s__%s",
	      c->name, c->elementType);
      fexp2 = new SgVarRefExp(*findFunctionSymb(buf1, f));
      lst = buildAssignStmt(fexp, fexp2);
      st->insertStmtAfter(*lst);
      st = lst;


#if DEBUG      
  printf("\n\tdone with this section\n", c->name);
  fflush(stdout);
#endif
    

    } else if (c->thisType == T_STRUCT ) {

      fexp = new SgNewExp(*buildDerivedType("brk_basicAType", f));
      // add a constant value parameter;
      ve = new SgValueExp(brk_stringSave(c->name));
      addParamExpToFuncExp(fexp2, ve);
      // add a constant value parameter;
      ve = new SgValueExp(T_UNKNOWN);
      addParamExpToFuncExp(fexp2, ve);
      
      lst = buildAssignStmt(new SgVarRefExp(*bsym),
			    fexp);
      st->insertStmtAfter(*lst);
      st = lst;

      fexp2 = buildFuncCallExp("brk_lookupComposite", f);
      // add first parameter;
      addParamToFuncExp(fexp2, bsym);
      lst = buildAssignStmt(new SgVarRefExp(*compsym),
			    fexp2);
      st->insertStmtAfter(*lst);
      st = lst;

      fs = findVarDeclSymbFromStmt("brk_accessFunc", 
				   findCompDeclStmt("brk_aComposite", f));
      fexp = &(SgPointStOp(*(new SgVarRefExp(*compsym)),
			 *(new SgVarRefExp(*fs))));
      
      sprintf(buf1, "brk_AccessFunc%s", c->name);
      lst = findFuncHedrStmt(buf1, f);
      fexp2 = extractFuncExpFromHedr(lst);
      lst = buildAssignStmt(fexp, fexp2);
      st->insertStmtAfter(*lst);
      st = lst;


    }
    c = (brk_aComposite *) nextItem();
    
#if DEBUG      
    printf("moving on to the next item %s\n", c->name);
    fflush(stdout);
#endif
    
 }

#if DEBUG      
  printf("DONE writing field access entries for composite list\n");
  fflush(stdout);
#endif

}


// --------------------------------
SgStatement *brk_sCompositeList::writeAccessFunctions(SgFile *fl) {
  brk_sComposite *f;
  SgStatement *st;

  // write each access function to file;
  f = (brk_sComposite *) searchList();

  while (f) {
#if DEBUG
    printf("searching composites for %s, %s\n", elementType, f->name);
    fflush(stdout);
#endif
    // make certain the name is not null;
    if (brk_strsame(elementType, "struct") && 
	brk_strsame(f->name, "")) {
      // a struct with no name, print it;
#     if DEBUG
      printf(" this composite did not have a name (access function).\n");
      f->print(0);
#     endif
     } else {
       st = f->writeAccessFunction(fl);
    }
    f = (brk_sComposite *) nextItem();
  }

 
#if DEBUG      
  printf("finished writing access functions, now doing access entries\n");
  fflush(stdout);
#endif
  
  // write table of functions to same file..;
  writeAccessEntries(fl);
  return st;
}

// **************************************************
// brk_sField functions...

// constructor for brk_sField class;
 brk_sField::brk_sField(brk_basicAType *t, char *n) {
   fieldType = t;
   name = brk_stringSave(n);
 }

// constructor for brk_sField class;
 brk_sField::brk_sField() {
 }

void brk_sField::writeFieldAccess(brk_sComposite *comp) {
  printf("shouldnt be called writeFieldAcess %s for %s\n",
	 name, comp->name);
  exit(1);
 }
  
