/*
  Darryl Brown
  April 94
  An implementation of sockets for tk/tcl.
*/

#include <stdio.h>
#include <sys/file.h>

#include "compatible.h"
#ifdef SYS5
#include <string.h>
#else
#include <strings.h>
#endif
 
#include "tk.h"
#include "tcl.h"
#include "tclInt.h"
#include "brkSockets.h"
#define DEBUG 0

/*  str_save  -- accepts a string, and mallocs space for it, or;
 *    returns an error if malloc fails...;
 */
char * brk_stringSave(char * str)
{
  char * tmp;
  if ((tmp = (char *) malloc (sizeof ( strlen(str)+1 ))) == NULL) 
    {
      printf("malloc failed\n");
      exit(1);
    }
  strcpy (tmp, str);
  return tmp;
}

int Tk_server_connection(ClientData clientData, Tcl_Interp *interp,
		      int argc, char **argv) {

  int socketfd, socketNum;
  char buf[100];

  if (argc != 2) {
    interp->result = "wrong # args";
    return TCL_ERROR;
  }
  /* print socket number into an int; */
  sscanf(argv[1], "%d", &socketNum);
  socketfd = brk_server_connection(socketNum);
  sprintf(buf, "%d", socketfd);
  interp->result = brk_stringSave(buf);
  return TCL_OK;
}


int Tk_brk_client_connection(ClientData clientData, Tcl_Interp *interp,
		      int argc, char **argv) {

  int socketfd, socketNum;
  char buf[100];
  printf("trying connection to port %d with host %s\n", socketNum, argv[2]);
  fflush(stdout);

  if (argc != 3) {
    interp->result = "wrong # args";
    return TCL_ERROR;
  }
  sscanf(argv[1], "%d", &socketNum);
  printf("still trying to port %d with host %s\n", socketNum, argv[2]);
  fflush(stdout);
  /* call client connection with host addr (argv[2]) and socket num (argv[1]);*/
  socketfd = brk_client_connection(argv[2], socketNum);
  sprintf(buf, "%d", socketfd);
  interp->result = brk_stringSave(buf);
  return TCL_OK;
}


int Tk_close_connection(ClientData clientData, Tcl_Interp *interp,
		      int argc, char **argv) {

  int socketfd;
  char buf[100];

  if (argc != 2) {
    interp->result = "wrong # args for proc";
    return TCL_ERROR;
  }

  if (sscanf(argv[1], "%d", &socketfd) != 1) {
    interp->result = "not a socket number";
    Tcl_AppendResult(interp, "\nUsage:\n\treadSocket <sockedNum> <maxlen>");
    return TCL_ERROR;
  }

  sprintf(buf, "%d", close(socketfd));
  interp->result = brk_stringSave(buf);
  return TCL_OK;
}


int Tk_writeStrSocket(ClientData clientData, Tcl_Interp *interp,
		      int argc, char **argv) {
  int socketfd;
  char buf[100];
  if (argc != 3) {
    interp->result = "wrong # args:";
    Tcl_AppendResult(interp, "\nUsage:\n\twriteSocket <sockedNum> <str>");
    return TCL_ERROR;
  }
  
  if (sscanf(argv[1], "%d", &socketfd) != 1) {
    interp->result = "not a socket number";
    Tcl_AppendResult(interp, "\nUsage:\n\twriteSocket <sockedNum> <str>");
    return TCL_ERROR;
  }

  sprintf(buf, "%d", brk_sockWritestr(socketfd, argv[2]));
  interp->result = brk_stringSave(buf);
  return TCL_OK;
}


int Tk_readLineSocket(ClientData clientData, Tcl_Interp *interp,
		      int argc, char **argv) {
  int socketfd, maxlen;
  char *buf;

  if (argc != 3) {
    interp->result = "wrong # args:";
    Tcl_AppendResult(interp, "\nUsage:\n\twriteSocket <sockedNum> <maxlen>");
    return TCL_ERROR;
  }
  
  if (sscanf(argv[1], "%d", &socketfd) != 1) {
    interp->result = "not a socket number";
    Tcl_AppendResult(interp, "\nUsage:\n\twriteSocket <sockedNum> <maxlen>");
    return TCL_ERROR;
  }

  if (sscanf(argv[2], "%d", &maxlen) != 1) {
    interp->result = "not a valid maximun length";
    Tcl_AppendResult(interp, "\nUsage:\n\twriteSocket <sockedNum> <maxlen>");
    return TCL_ERROR;
  }

  if ((buf = (char *) malloc (sizeof ( maxlen+1))) == NULL)  {
    perror("malloc failed\n");
  }
  brk_sockReadline(socketfd, buf, maxlen);
  interp->result = buf;
  return TCL_OK;
}

int Tk_readStrSocket(ClientData clientData, Tcl_Interp *interp,
		      int argc, char **argv) {
  int socketfd, maxlen, result;
  char *buf;

  if (argc != 3) {
    interp->result = "wrong # args:";
    Tcl_AppendResult(interp, "\nUsage:\n\treadSocket <sockedNum> <maxlen>");
    return TCL_ERROR;
  }
  
  if (sscanf(argv[1], "%d", &socketfd) != 1) {
    interp->result = "not a socket number";
    Tcl_AppendResult(interp, "\nUsage:\n\treadSocket <sockedNum> <maxlen>");
    return TCL_ERROR;
  }

  if (sscanf(argv[2], "%d", &maxlen) != 1) {
    interp->result = "not a valid maximun length";
    Tcl_AppendResult(interp, "\nUsage:\n\treadSocket <sockedNum> <maxlen>");
    return TCL_ERROR;
  }

  if (maxlen > SOCKET_BUF_SIZE) {
#if DEBUG    
    printf("maximum length (%d) too large for buffer (%d)\n",
	   maxlen, SOCKET_BUF_SIZE);
#endif
    maxlen =  SOCKET_BUF_SIZE;
  }

  if ((buf = (char *) malloc ((sizeof (char))*(maxlen+2))) == NULL)  {
   perror("malloc failed\n");
  }

  if ((result = brk_sockReadStr(socketfd, buf, maxlen)) >= 0) {
#if DEBUG    
    printf("returned %d from readSocketStr\n", result);
#endif
    interp->result = buf;
    return TCL_OK; 
  } else {
    if (maxlen < 50) {
      /* must allocate enough space for error message */
      free (buf);
      if ((buf = (char *) malloc (sizeof (50))) == NULL)  {
	perror("malloc failed\n");
      }
    }
    /* return error message */
#if DEBUG    
    printf("returned %d from readSocketStr\n", result);
#endif
    sprintf(buf, "ERROR: socket read result %d", result);
    interp->result = buf;
    return TCL_ERROR; 
  }
}

int Tk_readAllSocket(ClientData clientData, Tcl_Interp *interp,
		      int argc, char **argv) {
  int socketfd, maxlen, result, i;
  char *buf, *tbuf, *rbuf;

  maxlen = SOCKET_BUF_SIZE;

  if (argc != 2) {
    interp->result = "wrong # args:";
    Tcl_AppendResult(interp, "\nUsage:\n\treadSocket <sockedNum> <maxlen>");
    return TCL_ERROR;
  }
  
  if (sscanf(argv[1], "%d", &socketfd) != 1) {
    interp->result = "not a socket number";
    Tcl_AppendResult(interp, "\nUsage:\n\treadSocket <sockedNum> <maxlen>");
    return TCL_ERROR;
  }

  if ((buf = (char *) malloc ((sizeof (char))*(maxlen+2))) == NULL)  {
    perror("malloc failed\n");
  }
  if ((rbuf = (char *) malloc ((sizeof (char))*(maxlen+2))) == NULL)  {
    perror("malloc failed\n");
  }

  result = maxlen;
  i = 1;
  *rbuf = (char) NULL;
  while (result == maxlen) {
    if ((result = brk_sockReadStr(socketfd, buf, maxlen)) < 0) {
      /* error condition */
      if (maxlen < 50) {
	/* must allocate enough space for error message */
	free (buf);
	if ((buf = (char *) malloc (sizeof (50))) == NULL)  {
	  perror("malloc failed\n");
	}
      }
      /* return error message */
      sprintf(buf, "ERROR: socket read result %d", result);
      interp->result = buf;
      return TCL_ERROR; 
    } else {
      /* copy into result buffer */
      strcat(rbuf, buf);
#if DEBUG    
      printf("..just returned %s from read..", buf);
#endif
      i++;
      /* if result filled the buffer, then allocate more memory */
      if (result == maxlen) {
	if ((tbuf = (char *) malloc ((sizeof (char))*(maxlen+2)*i)) == NULL) {
	  perror("malloc failed\n");
	}
	/* move contents of rbuf into new buffer */
	strcpy(tbuf, rbuf);
	/* free old buffer */
	free(rbuf);
	/* point result buffer to newly allocated buffer */
	rbuf = tbuf;
      }
    }
  }

#if DEBUG    
  printf("returned %d from readSocketStr..", strlen(rbuf));
#endif
  interp->result = rbuf;
  return TCL_OK; 
}


char *appendAndAllocate(char *oldstr, char *newstr) {
  char *tbuf;

  if ((tbuf = (char *) malloc 
       (sizeof (char)*((strlen(oldstr)
			+strlen(newstr)+1)))) == NULL) {
    perror("malloc failed\n");
  }

  /* move contents of rbuf into new buffer */
  strcpy(tbuf, oldstr);
  /* free old buffer */
  free(oldstr);
  strcat(tbuf, newstr);
  free(newstr);
  /* point result buffer to newly allocated buffer */
  return tbuf;
}

int Tk_readSocketMessage(ClientData clientData, Tcl_Interp *interp,
		      int argc, char **argv) {
  int socketfd, maxlen, result, i;
  char *buf, *tbuf, *rbuf, endFormat[6];
  int notDone = 1;
  int cmdArgc;
  char *cmdArgv[3];

  maxlen = SOCKET_BUF_SIZE;
  
  if (argc != 2) {
    interp->result = "wrong # args:";
    Tcl_AppendResult(interp, "\nUsage:\n\treadSocket <sockedNum> <maxlen>");
    return TCL_ERROR;
  }
  
  if (sscanf(argv[1], "%d", &socketfd) != 1) {
    interp->result = "not a socket number";
    Tcl_AppendResult(interp, "\nUsage:\n\treadSocket <sockedNum> <maxlen>");
    return TCL_ERROR;
  }
  
  if ((buf = (char *) malloc ((sizeof (char))*maxlen)) == NULL)  {
    perror("malloc failed\n");
  }
  
  if ((rbuf = (char *) malloc ((sizeof (char)))) == NULL)  {
    perror("malloc failed\n");
  }
  
  *rbuf = (char) NULL;
  sprintf(endFormat, "\n!!\n");
  result = maxlen;


  cmdArgv[0] = "readAllSocket";
  /* store socket number */
  cmdArgv[1] = argv[1];
  /* store number of args */
  cmdArgc = 2;
  
  do {
    Tk_readAllSocket((ClientData) NULL, interp, 
		     cmdArgc, cmdArgv);
    buf = interp->result;

    if (strcmp(buf, endFormat)) {
      /* copy into result buffer */

#if DEBUG    
      printf(".. %d bytes from readmessage (%s)..", result, buf);
      fflush(stdout);
#endif
      rbuf = appendAndAllocate(rbuf, buf);
    } else {
      notDone = 0;
    }
  }  while(notDone);

#if DEBUG    
  printf("total returned %d from readmessage\n", strlen(rbuf));
  fflush(stdout);
#endif

  interp->result = rbuf;

  return TCL_OK; 
}

int brkSendDataToViz(ClientData clientData, Tcl_Interp *interp,
		     int argc, char **argv) {
  int cmdArgc;
  char *cmdArgv[3], *fileId, *curPtr, line[30], fname[40];
  FILE *f;
  int i, newline, elemNum;
  float tmpf;
  int numcount = 0;

#if DEBUG    
  int tsti;
#endif

  if ((argc < 4) && (argc > 5)) {
    sprintf(line, " argc(%d) ", argc);
    for (i = 1; i < argc; i++) {
      strcat(line, argv[i]);
      strcat(line, " ");
    }
    Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
		     " socketNum pipeNum dataRequest <?extract?>\"", line,
		     (char *) NULL);
    return TCL_ERROR;
  }

  /* find field name */
  curPtr = strchr(argv[3], '>');
  curPtr++;
  strcpy(fname, curPtr);

#if DEBUG    
  printf("fname: %s\n", fname);
  for (tsti = 0; tsti < argc; tsti++) {
    printf("argv[%d]: %s\n", tsti, argv[tsti]);
  }
#endif

  cmdArgv[0] = brk_stringSave("writeSocket");
  /* store socket number */
  cmdArgv[1] = argv[1];
  /* store data request */
  cmdArgv[2] = argv[3];
  /* store number of args */
  cmdArgc = 3;


#if DEBUG    
  for (tsti = 0; tsti < cmdArgc; tsti++) {
    printf("cmdArgv[%d]: %s\n", tsti, cmdArgv[tsti]);
  }
#endif

  Tk_writeStrSocket((ClientData) NULL, interp, 
		    cmdArgc, cmdArgv);

#if DEBUG    
  printf("done writeing data request \"%s\"\n",   cmdArgv[2]); 
  fflush(stdout);
#endif

  /* free allocated space */

  cmdArgv[0] = "readSocketMessage";
  /* store socket number */
  cmdArgv[1] = argv[1];
  /* store number of args */
  cmdArgc = 2;
  Tk_readSocketMessage((ClientData) NULL, interp, 
		   cmdArgc, cmdArgv);
  
#if DEBUG    
  printf("done reading\n"); fflush(stdout);
#endif

  /* now pipe interp result to viz process */
  
  /* store fileId argument */
  fileId = argv[2];
  
  /* get the open file */
  if (Tcl_GetOpenFile(interp, fileId, 1, 1, &f) != TCL_OK) {
    return TCL_ERROR;
  }

#if DEBUG    
  printf("retrieved open file descriptor\n"); fflush(stdout);
#endif
  
  clearerr(f);
  
  /* write retrieved data to pipe */
  curPtr = interp->result;
#if DEBUG    
  if (!curPtr) {
    printf("interpreter returned null pointer\n");
  } else {
    printf("interpreter returned \"%s\"\n", curPtr);
  }
  fflush(stdout);
#endif

  if (sscanf(curPtr, "@element(%d)", &elemNum) != 1) {

    Tcl_AppendResult(interp, "error in retrieved data \"", fileId,
		     "\" data:", 
		     curPtr, Tcl_PosixError(interp), (char *) NULL);
    return TCL_ERROR;
  }

#if DEBUG    
    printf("scanned elem number %d\n", elemNum);
    fflush(stdout);
#endif
    

  curPtr = strchr(curPtr, ')');
  curPtr++;
  strchr(curPtr, ')');
  curPtr = strtok(curPtr, " #@");
  if (argc == 4) {
    while (curPtr) {
      if (sscanf(curPtr, "%f", &tmpf) == 1) {
	if (0 && (tmpf > 1000000000)) {
	  printf("big number: %s\n", curPtr);
	  fprintf(f, " 0.00000");
	} else {
	  numcount++;
	  fprintf(f, " %s", curPtr);
	}
      } else {
	printf( "%s was not a number, substituting 0", curPtr);
	fprintf(f, " 0");
#if DEBUG    
#endif
      }
      
      if (ferror(f)) {
	Tcl_AppendResult(interp, "error writing \"", fileId,
			 "\": ", Tcl_PosixError(interp), (char *) NULL);
	return TCL_ERROR;
      }
      curPtr = strtok(NULL, " #");
    }
  } else {
    /* argc must be five, so we must pick out all non numbers... */
    while (curPtr) {
      /* find field name */
      while (strcmp(curPtr, fname) != 0) {
#if DEBUG    
	printf("looking at string %s\n", curPtr);
	fflush(stdout);
#endif
	curPtr = strtok(NULL, " #");
	if (curPtr == NULL) {
	  break;
	}
      }
      if (curPtr) {
#if DEBUG    
	printf("found field name %s\n", curPtr);
	fflush(stdout);
#endif
	/* now get next item, should be the data */
	curPtr = strtok(NULL, " #");
	
	if (sscanf(curPtr, "%f", &tmpf) == 1) {
	  if (0 && (tmpf > 1000000000)) {
	    printf("big number: %s\n", curPtr);
	    fprintf(f, " 0.00000");
	  } else {
	    numcount++;
#if DEBUG    
	    printf("(n:%d) ", curPtr);
	    fflush(stdout);
#endif
	    fprintf(f, " %s", curPtr);
	  }
	} else {
	  printf( "%s was not a number, substituting 0", curPtr);
	  fprintf(f, " 0");
	}
	
	if (ferror(f)) {
	  Tcl_AppendResult(interp, "error writing \"", fileId,
			   "\": ", Tcl_PosixError(interp), (char *) NULL);
	  return TCL_ERROR;
	}
	curPtr = strtok(NULL, ")");
	curPtr = strtok(NULL, " #");
#if DEBUG    
	if (curPtr) {
	  printf("\n\tnext token: \"%s\"\n", curPtr);
	  fflush(stdout);
	}
#endif
      }
    }
  }
    
  fprintf(f, "\n");

#if DEBUG    
  printf("done parsing interp result (returning from vizDataFromSocket)\n");
  fflush(stdout);
#endif
  return TCL_OK;
}



int brk_Init(Tcl_Interp *interp) {
  Tcl_CreateCommand(interp, "serverConn", 
		    Tk_server_connection,
		    (ClientData) NULL,
		    (Tcl_CmdDeleteProc *) NULL);
  Tcl_CreateCommand(interp, "clientConn", 
		    Tk_brk_client_connection,
		    (ClientData) NULL,
		    (Tcl_CmdDeleteProc *) NULL);
  Tcl_CreateCommand(interp, "closeSocket", 
		    Tk_close_connection,
		    (ClientData) NULL,
		    (Tcl_CmdDeleteProc *) NULL);
  Tcl_CreateCommand(interp, "writeSocket", 
		    Tk_writeStrSocket,
		    (ClientData) NULL,
		    (Tcl_CmdDeleteProc *) NULL);
  Tcl_CreateCommand(interp, "readSocket", 
		    Tk_readStrSocket,
		    (ClientData) NULL,
		    (Tcl_CmdDeleteProc *) NULL);
  Tcl_CreateCommand(interp, "readLineSocket", 
		    Tk_readLineSocket,
		    (ClientData) NULL,
		    (Tcl_CmdDeleteProc *) NULL);
  Tcl_CreateCommand(interp, "readAllSocket", 
		    Tk_readAllSocket,
		    (ClientData) NULL,
		    (Tcl_CmdDeleteProc *) NULL);
  Tcl_CreateCommand(interp, "vizDataFromSocket", 
		    brkSendDataToViz,
		    (ClientData) NULL,
		    (Tcl_CmdDeleteProc *) NULL);
  return TCL_OK;
}
