// This may look like C code, but it is really -*- C++ -*-

#ifndef ITERATOR_H
#define ITERATOR_H 1

#include "sage++user.h"

class Iterator {
  public:
     Iterator (SgFile *file);
     virtual ~Iterator();

     /* This begins the iteration */
     void begin (void);

     /* This member function is called for every statement in the
	scope of the iteration. Subclasses must override this!

	The value `1'  should be returned if no special
	behaviour is desired.
	     o If a 0 is returned, the current statement's children
	     will not be descended. [This should be returned if the
	     current statement is deleted.]
	     
	     o If a -1 is returned, the iterator backs up to the
	     previous statement and re-processes it's "next". This can
	     be useful when destructively replacing statements when
	     the replacement may need to be processed.

	     Note that `before_descent' will NOT be called again
	     when rescanning the first member of a child list.
	*/
     virtual int process (SgStatement *current_stmt) = 0;

     /* descent/ascent control:
	[[before_descent]] is called before scanning a child list.
	The [[current]] position is the parent of the list.

	[[after_ascent]] is called just after the scanning of a child
	list is finished. The [[current]] position is *the last item
	in the child list*.
	*/
     
     virtual void before_descent (SgStatement *current_stmt);
     virtual void after_ascent (SgStatement *current_stmt);

     /* Variables that indicate where we are in the traversal */
     SgStatement *prev;
     SgStatement *current;
     SgStatement *next;

  private:
     int iterate (SgStatement *current_stmt);

     SgFile *the_file;
};

#include "iterator.icc"

#endif /* ndef ITERATOR_H */
