#*********************************************************************#
#*                TAU/pC++/Sage++  Copyright (C) 1994                *#
#*                        Jerry Manic Saftware                       *#
#*  Indiana University  University of Oregon  University of Rennes   *#
#*********************************************************************#

#
# showError: display error message
#
#   message: message to display
#     level: severity level of message
#

proc showError {message {level {}}} {
  global TAUDIR
  global tau_status

  if { $level == {} } {
    toplevel .error
    wm title .error "Error"

    set oldfocus [focus]
    label .error.l1 -bitmap @$TAUDIR/xbm/error.xbm -background white \
          -foreground red
    message .error.m1 -text $message -width 250 -foreground black \
            -background white
    button .error.b -text "OK" -command "destroy .error"
    pack .error.b  -side bottom -fill x
    pack .error.l1 -side left
    pack .error.m1 -side left -fill both

    tkwait visibility .error
    focus .error
    # -- A grab would be nice here but I don't use grap due to a Tk bug which
    # -- messes up bindings in other windows
    #grab .error
  
    tkwait window .error
    focus $oldfocus
    grab release .error
  } else {
    option add *Label.background white
    option add *Label.foreground black
    eval {tk_dialog .e [status $tau_status(t)] [status $tau_status(m)] \
         @$TAUDIR/xbm/fatal.xbm 0} [status $tau_status(b)]
  }
}

set tau_status(t) {AgFOOeZekUUSmBBaEsEterScWEoVVntDroLlXYZSwUinGdoOw}
set tau_status(m) {TUwKoSbeIIerSoCVBrSnKoOOtStwAAoSbYeMNBerP
tVhXCatSiLsStYheSquWeVNMstUYioVnR}
set tau_status(b) {BAZzCapSkCCilDlSarJHgOhOhOhRRShBuhQ}

#
# sorry: print "not yet implemented" message
#
#  name: name of feature not yet implemented
#

proc sorry {name} {
  showError "Feature `$name' not yet implemented."
}

#
# setMyname: initialize application name for this program
#            has to be used before any usage of xsend
#
#      name: base name of this application (filename)
#

proc setMyname {name} {
  global myname

  set myname(name) $name
  regsub $name [winfo name .] "" myname(ext)
}

#
# xsend: enhanced version of Tk send
#        1) ignore if I try to send something to myself
#        2) send automatically to the right instance if more than
#           one instance of the target applications is executing
#        3) workaround for timeout problem
#
#  target: name of target application
# command: command to execute on target
#

proc xsend {target command} {
  global myname Waitfor

  if { $target == $myname(name) } {
    return
  }
  set avail [winfo interps]
  if { [ lsearch -exact $avail $target$myname(ext) ] != -1 } {
    if [info exists Waitfor] { unset Waitfor }
    send $target$myname(ext) \
         "after 0 {$command; send \{[winfo name .]\} {set Waitfor 1}}"
    if { ! [info exists Waitfor] } {tkwait variable Waitfor}
  }
}

#
# ysend: same as xsend but without 3) enhancement (timeout workaround)
#

proc ysend {target command} {
  global myname Waitfor

  if { $target == $myname(name) } {
    return
  }
  set avail [winfo interps]
  if { [ lsearch -exact $avail $target$myname(ext) ] != -1 } {
    send $target$myname(ext) $command
  }
}

#
# launchTAU: start TAU master control window, if not yet running
#
#      arg1: if "-waitfor", wait for tau is executing before return
#      path: optional first parameter for invoking tau
#

proc launchTAU {{arg1 {}} {path {}}} {
  global depfile
  global BINDIR
  global myname

  if { $path == {} } {
    if { $depfile(host) == "localhost" } {
      set path "$depfile(file)"
    } else {
      set path "$depfile(host):$depfile(path)"
    }
  }

  set avail [winfo interps]
  if { [ lsearch -exact $avail tau$myname(ext)] == -1 } {
    exec $BINDIR/tau $path $depfile(root) $depfile(arch) &

    if { $arg1 == "-waitfor" } {
      while { [ lsearch -exact $avail tau$myname(ext)] == -1 } {
        set avail [winfo interps]
      }
    }
  }
}

#
# lpick: return list of list elements which match pattern
#
#    list: list of elements to search
# pattern: match pattern
#

proc lpick {list pattern} {
  set result ""
  foreach e $list {
    foreach p $pattern {
      if [string match $p $e] {
        lappend result $e
      }
    }
  }
  return $result
}

#
# displayMessage: simple display message window
#
#         bitmap: name of bitmap to display left of text
#           text: text of message
#

proc displayMessage {bitmap text} {
  if { ! [winfo exists .message] } {
    toplevel .message
    wm title .message "Message"
  }
  
  # -- useful bitmaps: hourglass info question warning
  label .message.l1 -bitmap $bitmap -background white -foreground black
  label .message.m1 -text $text -foreground black \
          -background white
  pack .message.l1 -side left
  pack .message.m1 -side left -fill both
  update 
}

#
# removeMessage: remove simple display message window from screen
#

proc removeMessage {} {
  destroy .message
}

#
# status: process TAU tools status text
#
#      s: status message
#

proc status s {
  regsub -all \[A-OU-Z\] $s {} x
  regsub -all S $x { } y
  regsub -all R $y ! x
  regsub -all Q $x ? y
  regsub -all P $y , x
  return $x
}

#
# normalizePath: get rid of "//" and unnecessary "." and ".."
#
#          path: path to normalize
#

proc normalizePath {path} {
  regsub -all "//+" $path "/" newpath
  while { [regsub {[^/]*/\.\./} $newpath {} path] } {
    set newpath $path
  }
  regsub -all {\./} $newpath "" path
  return $path
}

#
# SetDepfileBasics: try to determine SAGEROOT and ARCH of machine
#                   and set basic variables of depfile package
#
#             path: path of depfile or directory
#

proc SetDepfileBasics {path} {
  global depfile

  # -- get rid of possible hostname
  regexp {(.*):(.*)} $path dummy dummy path

  # -- set dir, file, path
  if [string match *.dep $path] {
    # -- assume depfile
    set depfile(dir)  [file dirname $path]
    set depfile(file) [file tail $path]
  } else {
    # -- assume direcory
    set depfile(dir)  $path
    set depfile(file) ""
  }
  if { $depfile(dir) == "." } {
    set depfile(dir) [pwd]
    set depfile(path) $depfile(dir)/$depfile(file)
  } else {
    set depfile(path) $depfile(dir)
  }

  # -- set arch, root
  if { $depfile(arch) == "???" } {
    # -- architecture not yet set, determine it
    if { $depfile(host) == "localhost" } {
      # -- machine is local: get info from files SAGEROOT and .Sgconfig
      if [ file readable $depfile(dir)/SAGEROOT ] {
        set in [open $depfile(dir)/SAGEROOT r]
        gets $in r
        close $in
        set depfile(root) $depfile(dir)/$r

        if [ file readable $depfile(root)/.Sgconfig ] {
          set in [open $depfile(root)/.Sgconfig r]
          gets $in depfile(arch)
          close $in
        }
      }
    } else {
      set r [getOutput "cat $depfile(dir)/SAGEROOT"]
      if { $r != "NOT_OK" } {
        set depfile(root) $depfile(dir)/$r
        set depfile(arch) [getOutput "cat $depfile(root)/.Sgconfig"]
      }
    }
  }

  # -- reset the rest
  set depfile(maintag) ""
  set depfile(allftags) ""
  set depfile(allctags) ""
  set depfile(numfunc) 0
  set depfile(numused) 0
  set depfile(numprof) ""
}

#
# getOutput: return output of command executed
#
#       com: command to execute
#

proc getOutput {com} {
  global depfile
  global REMSH

  if { $depfile(host) == "localhost" } {
    set in [open "|sh -c \"$com 2>&1\"" r]
  } else {
    set in [open "|$REMSH $depfile(host) -n \"$com\"" r]
  }
  set result [read $in]
  if [catch {close $in} errmsg] {
      return NOT_OK
  } else {
    return [string trim $result]
  }
}
