#*********************************************************************#
#*                TAU/pC++/Sage++  Copyright (C) 1994                *#
#*                        Jerry Manic Saftware                       *#
#*  Indiana University  University of Oregon  University of Rennes   *#
#*********************************************************************#

source "$TAUDIR/inc/depfile.tcl"

set cg_colors {black #0000A0000000}; # basic display colors
set cg_showlib 0;                # display library (non-body) functions?
set cg_showclass 1;              # display class name of member functions too?
set cg_showtype 1;               # display type of functions?
set cg_sel_tag NONE;             # current selected function
set cg_sel_cs_obj -1;            # currently selected callsite
set cg_sel_cs_info "";           # info about currently selected callsite
set cg_drawmode "EXPANDED";      # drawing mode (EXPANDED | COMPACT)

#
# selectLine: implementation of global feature "selectLine" for fancy
#             highlight specified line in file
#
#       line: line number to select
#       file: file name to select
#        tag: tag of call site to select or -1
#

proc selectLine {line file {tag -1}} {
  global myself
  global funcs
  global cg_sel_cs_obj cg_sel_cs_info cg_drawmode

  # -- if there is already a function selected, un-select it
  # -- (display in black)
  if { $cg_sel_cs_obj != -1 } {
    .$myself.graph.can itemconfigure $cg_sel_cs_obj -fill black
  }

  # -- if selected call site is part of the callgraph
  # -- select it (display it in blue)
  set obj -1
  if { $cg_drawmode == "EXPANDED" } {
    if [info exists funcs($tag,css)] {
      set idx [lsearch -exact $funcs($tag,css) [list $line $file]]
      if { $idx != -1 } {
        set obj [lindex $funcs($tag,objs) $idx]
        .$myself.graph.can itemconfigure $obj -fill blue
        set cg_sel_cs_info [list $line $file $tag]
        displayObj $obj
      }
    }
  } else {
    if [info exists funcs($tag,objs)] {
      set obj [lindex $funcs($tag,objs) 0]
      .$myself.graph.can itemconfigure $obj -fill blue
      set cg_sel_cs_info [list $line $file $tag]
      displayObj $obj
    }
  }
  set cg_sel_cs_obj $obj
}

#
# selectFuncTag: implementation of global feature "selectFunction" for cagey
#                display all instances of selected function in red
#
#           tag: id of function to select
#

proc selectFuncTag {tag} {
  global myself
  global funcs
  global cg_sel_tag

  # -- if there is already a function selected, un-select it
  # -- (display in black)
  if { $cg_sel_tag != "NONE" } {
    if [info exists funcs($cg_sel_tag,objs)] {
      foreach o $funcs($cg_sel_tag,objs) {
        .$myself.graph.can itemconfigure $o -fill black
      }
    }
  }

  # -- if selected function is part of the callgraph
  # -- select it (display it in red)
  if [info exists funcs($tag,objs)] {
    foreach o $funcs($tag,objs) {
      .$myself.graph.can itemconfigure $o -fill red
    }
    displayObj [lindex $funcs($tag,objs) 0]
  }
  set cg_sel_tag $tag
}

#
# createWindow: create and display main window of cagey
#

proc createWindow {} {
  global TAUDIR
  global myself

  toplevel .$myself
  wm title .$myself "CAGEY"
  wm iconbitmap .$myself @$TAUDIR/xbm/$myself.xbm
  wm minsize .$myself 400 600

  frame .$myself.graph

  # -- configure graph area
  canvas .$myself.graph.can -width 400 -height 600 -background white
  bind .$myself.graph.can <2> ".$myself.graph.can scan mark %x %y"
  bind .$myself.graph.can <B2-Motion> ".$myself.graph.can scan dragto %x %y"


  frame .$myself.graph.f1

  scrollbar .$myself.graph.hscroll -orient horiz -relief sunken \
            -command ".$myself.graph.can xview"
  .$myself.graph.can configure -xscroll ".$myself.graph.hscroll set"

  frame .$myself.graph.f2 -width 20 -height 20

  pack .$myself.graph.hscroll -in .$myself.graph.f1 -side left \
                              -expand yes -fill x
  pack .$myself.graph.f2      -in .$myself.graph.f1 -side right

  scrollbar .$myself.graph.vscroll -orient vert -relief sunken \
            -command ".$myself.graph.can yview"
  .$myself.graph.can configure -yscroll ".$myself.graph.vscroll set"

  pack .$myself.graph.f1      -side bottom -fill x
  pack .$myself.graph.vscroll -side right  -fill y
  pack .$myself.graph.can     -side top -padx 15 -pady 15 -fill both -expand yes

  # -- configure menu bar
  frame .$myself.bar -relief raised -borderwidth 2

  menubutton .$myself.bar.b1 -text File -menu .$myself.bar.b1.m1 -underline 0
  menu .$myself.bar.b1.m1
  .$myself.bar.b1.m1 add command -label "Generate instrfile" -underline 0 \
                   -command "sorry Generate"
  .$myself.bar.b1.m1 add command -label "Print graph" -underline 0 \
                   -command "printGraph"
  .$myself.bar.b1.m1 add separator
  .$myself.bar.b1.m1 add command -label "Exit"  -underline 0 -command "exit"

  menubutton .$myself.bar.b2 -text View -menu .$myself.bar.b2.m1 -underline 0
  menu .$myself.bar.b2.m1
  .$myself.bar.b2.m1 add radiobutton -label "expanded callgraph" \
                   -underline 1 -variable cg_drawmode -value EXPANDED \
                   -command "redrawGraph"
  .$myself.bar.b2.m1 add radiobutton -label "compact callgraph" \
                   -underline 1 -variable cg_drawmode -value COMPACT \
                   -command "redrawGraph"
  .$myself.bar.b2.m1 add separator
  .$myself.bar.b2.m1 add checkbutton -label "Show library functions" \
                   -underline 5 -variable cg_showlib -onvalue 1 -offvalue 0 \
                   -command "redrawGraph"
  .$myself.bar.b2.m1 add checkbutton -label "Show class names" \
                   -underline 5 -variable cg_showclass -onvalue 1 -offvalue 0 \
                   -command "redrawGraph"
  .$myself.bar.b2.m1 add checkbutton -label "Show type of function" \
                   -underline 5 -variable cg_showtype -onvalue 1 -offvalue 0 \
                   -command "redrawGraph"
  .$myself.bar.b2.m1 add separator
  .$myself.bar.b2.m1 add cascade -label "Expand graph" -underline 0 \
                                 -menu .$myself.bar.b2.m1.1
  menu .$myself.bar.b2.m1.1

  menubutton .$myself.bar.b3 -text Help -menu .$myself.bar.b3.m1 -underline 0
  menu .$myself.bar.b3.m1
  .$myself.bar.b3.m1 add command -label "on $myself" -underline 3 \
                   -command "xsend tau \[list showHelp $myself 1-$myself 1\]"
  .$myself.bar.b3.m1 add separator
  .$myself.bar.b3.m1 add command -label "on menubar" -underline 3 \
                   -command "xsend tau \[list showHelp $myself 1.1-menu 1\]"
  .$myself.bar.b3.m1 add command -label "on display area" -underline 3 \
                   -command "xsend tau \[list showHelp $myself 1.2-display 1\]"
  .$myself.bar.b3.m1 add separator
  .$myself.bar.b3.m1 add command -label "on using help" -underline 3 \
                   -command "xsend tau {showHelp general 1-help 1}"

  pack .$myself.bar.b1 .$myself.bar.b2 -side left -padx 5
  pack .$myself.bar.b3 -side right -padx 5
  tk_menuBar .$myself.bar .$myself.bar.b1 .$myself.bar.b2 .$myself.bar.b3

  pack .$myself.bar   -side top -fill x
  pack .$myself.graph -side left -padx 15 -pady 15 -fill both -expand yes
  focus .$myself.bar
}

#
# loadDep: switch to another application
#          this is invoked from the TAU master control window
#
#    path: UNIX pathname of depfile to read
#    root: patchname of SAGEROOT
#    arch: architecture of machine where depfile resides
#

proc loadDep {path root arch} {
  global myself
  global funcs
  global depfile
  global cg_sel_tag
  global cg_sel_cs_obj

  if { ! [regexp {(.*):(.*)} $path dummy depfile(host) path] } {
    set depfile(host) "localhost"
  }
  set depfile(root) $root
  set depfile(arch) $arch

  # -- delete old database of functions
  unset funcs

  # -- read new depfile and rebuild function database
  if { [readDep -dumpcg $path] == "NOT_OK" } {
    exit
  }

  # -- draw new callgraph
  set funcs($depfile(maintag),fold) 0
  redrawGraph
  set cg_sel_tag NONE
  set cg_sel_cs_obj -1

  # -- update "expand to level" menu
  .$myself.bar.b2.m1.1 delete 0 last
  expandMenu
}

#
# displayObj: scroll window so that a specific object is on the same
#             position on the screen after display operations
#
#        obj: object of interest
#

proc displayObj {obj} {
  global myself

  set winpos [.$myself.graph.can canvasy 0]
  set objpos [lindex [.$myself.graph.can coords $obj] 1]

  if { [expr $objpos-$winpos] < 0 || [expr $objpos-$winpos] > 600 } {
    # -- object not on screen
    set target [expr $objpos-300]
    set scry [lindex [.$myself.graph.can configure -scrollincrement] 4]
    .$myself.graph.can yview [expr $target/$scry]
  }
}

#
# toggleFold: toggle object between "expanded" and "folded" mode
#
#        obj: object to toggle mode
#        tag: Sage++ id of the corresponding function
#

proc toggleFold {obj tag} {
  global myself
  global funcs

  set f $funcs($tag,fold)
  set funcs($tag,fold) [expr 1-$f]
  set hasChildren [llength $funcs($tag,calls)]

  if { $hasChildren } {
    # -- function has children, redisplay new graph layout
    # -- also, scroll window so that the toggled object is on the same
    # -- position on the screen after redisplay
    set oldmyself [lsearch $funcs($tag,objs) $obj]
    set oldobj [lindex [.$myself.graph.can coords $obj] 1]
    set oldwin [.$myself.graph.can canvasy 0]
    redrawGraph
    set newmyself [lindex $funcs($tag,objs) $oldmyself]
    set newobj [lindex [.$myself.graph.can coords $newmyself] 1]
    set scry [lindex [.$myself.graph.can configure -scrollincrement] 4]
    .$myself.graph.can yview [expr ($newobj-$oldobj+$oldwin)/$scry]
  }
}

#
# toggleObj: toggle object between two display modes (indicated by colors)
#            not yet used for something useful
#
#       tag: Sage++ id of the corresponding function
#

proc toggleObj {tag} {
  global myself
  global funcs
  global cg_colors

  set s $funcs($tag,state)
  set s [set funcs($tag,state) [expr 1-$s]]
  foreach o $funcs($tag,objs) {
    .$myself.graph.can itemconfigure $o -fill [lindex $cg_colors $s]
  }
}

#
# transGraph: traverse callgraph to expand the graph up to a specific
#             level given in cg_exp_level
#             does not work as expected, as the same function can be
#             called on different levels
#             *called recursively *
#
#        tag: function id of root of subtree to traverse
#      level: it's level
#     parent: function id of parent function
#

proc transGraph {tag level parent} {
  global funcs
  global cg_exp_level cg_marked

  set calls $funcs($tag,calls)

  if { $tag == $parent } { return }

  if { $level < $cg_exp_level } {
    set funcs($tag,fold) 0
    lappend cg_marked $tag
  } elseif { $level == $cg_exp_level } {
    if { [lindex $cg_marked $tag] == -1 } {
      set funcs($tag,fold) 1
    }
    return
  }

  foreach t $calls {
    transGraph $t [expr $level+1] $tag
  }
}

#
# getMaxLavel: compute depth of call graph by traversing it recursively
#
#         tag: function id of root of subtree to traverse
#       level: it's level
#      parent: function id of parent function
#

proc getMaxLavel {tag level parent} {
  global funcs
  global cg_max_level

  set calls $funcs($tag,calls)

  if { $tag == $parent } { return }

  if { $level > $cg_max_level } { set cg_max_level $level }

  foreach t $calls {
    getMaxLavel $t [expr $level+1] $tag
  }
}

#
# expandMenu: update "expand" menu, so that it has an entry for each
#             level of the callgraph
#

proc expandMenu {} {
  global depfile
  global myself cg_exp_level cg_max_level

  set cg_max_level 0
  getMaxLavel $depfile(maintag) 0 -1

  for {set i 0} {$i<=$cg_max_level} {incr i} {
    .$myself.bar.b2.m1.1 add radiobutton -label "upto level $i" -underline 11 \
                  -variable cg_exp_level -value $i \
                  -command "expandHierarchy"
  }
  set cg_exp_level 1
}

#
# expandHierarchy: expand callgraph after user selected level through
#                  the "expand" menu
#

proc expandHierarchy {} {
  global depfile
  global cg_marked

  set cgmarked ""
  transGraph $depfile(maintag) 0 -1
  redrawGraph
}

set cg_cx 50;   # start x coordinate
set cg_cy 50;   # start y coordinate
set cg_dx 30;   # delta x to next node
set cg_dy 30;   # delta y to next node

#
# drawGraphExpd: draw "expanded" callgraph by traversing it recursively
#
#           tag: function id of root of subtree to traverse
#         level: it's level
#        parent: function id of parent function
#          file: filename where this instance of the function is called
#          line: linenumber where this instance of the function is called
#

proc drawGraphExpd {tag level parent {file -} {line -1}} {
  global myself
  global funcs
  global cg_cx cg_cy cg_dx cg_dy
  global cg_colors
  global cg_showlib cg_showclass cg_showtype

  # -- initialize parameters for current node
  set name  $funcs($tag,name)
  set calls $funcs($tag,calls)
  set class [lindex $funcs($tag,class) 0]
  if [llength $calls] {
    set cfile [lindex $funcs($tag,file) 1]
    set cline $funcs($tag,childline)
  }
  set ptype [lindex $funcs($tag,type) 0]
  set s     $funcs($tag,state)
  set fold  $funcs($tag,fold)
  set x [expr $cg_cx+$level*$cg_dx]
  set y [expr $cg_cy+10]

  # -- prepend classname if necessary
  if { $cg_showclass && $class != "-" } {
    set name "$class::$name"
  }

  # -- show type if necessary
  if { $cg_showtype && ! [lindex $funcs($tag,file) 0] } {
    set name "<$name>"
  }
  if { $cg_showtype && $ptype == "par" } {
    set name "|| $name"
  }

  # -- mark recursive procedures
  if { $tag == $parent } {
    set name "$name **"
  }

  # -- mark nodes which are folded and have children
  if { $fold && [llength $calls] } {
    set setFold 1
    if { ! $cg_showlib } {
      set setFold 0
      foreach t $calls {
        set body [lindex $funcs($t,file) 0]
        if { $body } {
          set setFold 1
          break
        }
      }
    }
    if { $setFold } {
      set name "$name  . . ."
    }
  }

  # -- finally draw node and add bindings
  set obj [.$myself.graph.can create text $x $cg_cy -text $name -anchor w \
                        -fill [lindex $cg_colors $s]]
  .$myself.graph.can bind $obj <Button-1> "toggleFold $obj $tag"
  .$myself.graph.can bind $obj <Shift-Button-1> "toggleObj $tag"
  .$myself.graph.can bind $obj <Button-2> "global_showFuncTag $tag"
  .$myself.graph.can bind $obj <Button-3> "global_selectFuncTag $tag"
  lappend funcs($tag,objs) $obj
  lappend funcs($tag,css) [list $line $file]
  incr cg_cy $cg_dy
  if { $fold || $tag == $parent } {
    return
  }

  # -- draw children of current node
  set c 0
  foreach t $calls {
    set body [lindex $funcs($t,file) 0]

    if { $cg_showlib || $body } {
      .$myself.graph.can create line [expr $x+5] $y \
                                [expr $x+5] $cg_cy \
                                [expr $x+$cg_dx-5] $cg_cy -arrow last
      drawGraphExpd $t [expr $level+1] $tag $cfile [lindex $cline $c]
    }
    incr c
  }
}

#
# redrawGraph: redraw callgraph by resetting variables, deleting old
#              callgraph, then draw new one
#

proc redrawGraph {} {
  global myself
  global funcs
  global depfile
  global cg_sel_tag
  global cg_sel_cs_obj cg_sel_cs_info

  # -- delete obsolete function object information
  foreach t $depfile(allftags) {
    if [ info exists funcs($t,objs) ] { unset funcs($t,objs) }
    if [ info exists funcs($t,x) ] { unset funcs($t,x) }
    if [ info exists funcs($t,y) ] { unset funcs($t,y) }
  }

  # -- delete graphic objects
  .$myself.graph.can delete all

  # -- redraw graph
  drawGraph
  if { $cg_sel_tag != "NONE" } {
    if [info exists funcs($cg_sel_tag,objs)] {
      foreach o $funcs($cg_sel_tag,objs) {
        .$myself.graph.can itemconfigure $o -fill red
      }
    }
  }
  if { $cg_sel_cs_obj != -1 } {
    eval selectLine $cg_sel_cs_info
  }
}

set pr_printVar window;  # print mode: "window" or "all"
set pr_psfile "";        # PostScript filename

#
# printGraph: generate PostScript version of callgraph
#

proc printGraph {} {
  global depfile
  global pr_psfile
  global pr_printVar
  global cg_cx cg_cy cg_dx cg_dy

  toplevel .print
  wm title .print "Print"

  regsub .dep $depfile(file) .ps pr_psfile
  frame .print.top
  pack .print.top -side top -padx 15 -pady 15

  radiobutton .print.r1 -text "Graph" -variable pr_printVar \
                    -value all -width 10 -anchor w
  radiobutton .print.r2 -text "Window" -variable pr_printVar \
                    -value window -width 10 -anchor w
  frame .print.f1 -width 20 -height 20
  label .print.l1 -text "Filename:"
  entry .print.e1 -textvariable pr_psfile -relief sunken
  pack .print.r1 .print.r2 .print.f1 .print.l1 .print.e1 \
                 -in .print.top -side top -anchor w

  frame .print.bottom -relief sunken -bd 1
  pack .print.bottom -side left -padx 15 -pady 10

  button .print.b1 -text "print" -command {
      if { [info exists pr_psfile] } {
        if { $pr_printVar == "all" } {
          .$myself.graph.can postscript -colormode color -file $pr_psfile \
                       -x 0 -y 0 -height [expr $cg_cy-$cg_dy+50] -width 600 \
                       -pageheight 10.5i
        } else {
          .$myself.graph.can postscript -colormode color -file $pr_psfile
        }
      }
      destroy .print
    }
  pack .print.b1 -in .print.bottom -side top -padx 5 -pady 5

  button .print.b2 -text "cancel" -command "destroy .print"
  bind .print.e1 <Return> {.print.b1 invoke}
  bind .print <Return> {.print.b1 invoke}
  pack .print.b2 -side right -padx 15

  tkwait visibility .print
  set oldfocus [focus]
  focus .print.e1

  tkwait window .print
  focus $oldfocus
}

#
# traverseGraphX: traverse callgraph to compute X coordinates for
#                 compact callgraph
#
#            tag: function id of root of subtree to traverse
#              x: it's x coordinate (level)
#         parent: function id of parent function
#

proc traverseGraphX {tag x parent} {
  global funcs
  global myself
  global cg_max_x
  global cg_showlib

  # -- determine maximal level (x coordinate)
  if { $x > $cg_max_x } { set cg_max_x $x}

  # -- set x coordinate for node (always use the deepest level of a function)
  if [info exists funcs($tag,x)] {
    if { $funcs($tag,x) < $x } {
      set funcs($tag,x) $x
    }
  } else {
    set funcs($tag,x) $x
  }
  incr x

  # -- traverse children nodes if necessary
  if { ! $funcs($tag,fold) && $tag != $parent } {
    foreach c $funcs($tag,calls) {
      set body [lindex $funcs($c,file) 0]
      if { $cg_showlib || $body } {
        traverseGraphX $c $x $tag
      }
    }
  }
}

#
# traverseGraphY: traverse callgraph to compute Y coordinates for
#                 compact callgraph
#
#            tag: function id of root of subtree to traverse
#

proc traverseGraphY {tag} {
  global funcs
  global myself
  global cg_max_y
  global cg_showlib

  if { $funcs($tag,fold) } {
    # -- node folded: just assign next available y coordinate
    incr cg_max_y
    set my_y $cg_max_y
  } else {
    # -- otherwise: first determine number of "real" children
    # -- (children which are displayed)
    set realcalls ""
    foreach c $funcs($tag,calls) {
      set body [lindex $funcs($c,file) 0]
      if { ![info exists funcs($c,y)] && ($cg_showlib || $body) } {
        lappend realcalls $c
        set funcs($c,y) "PRESET"
      }
    }

    switch -exact [llength $realcalls] {
    0 {
        # -- no children: assign next available y coordinate
        incr cg_max_y
        set my_y $cg_max_y
      }
    1 {
        # -- one children: assign y coordinate of child
        set c [lindex $realcalls 0]
        set my_y [traverseGraphY $c]
      }
    default {
        # -- more than one children: place parent in the middle of all childs
        set start_y $cg_max_y
        foreach c $realcalls {
          set end_y [traverseGraphY $c]
        }
        set my_y [expr $start_y+($end_y-$start_y+1)/2]
      }
    }
  }
  set funcs($tag,y) $my_y
  return $my_y
}

#
# drawGraphCompact: draw "compact" callgraph by traversing it recursively
#
#              tag: function id of root of subtree to traverse
#           parent: function id of parent function
#

proc drawGraphCompact {tag parent} {
  global funcs
  global myself
  global cg_colors
  global cg_showlib cg_showclass cg_showtype

  # -- initialize parameters for current node
  set x [expr 30+$funcs($tag,x)*300]
  set y [expr 30+$funcs($tag,y)*30]
  set name  $funcs($tag,name)
  set calls $funcs($tag,calls)
  set class [lindex $funcs($tag,class) 0]
  set ptype [lindex $funcs($tag,type) 0]
  set s     $funcs($tag,state)
  set fold  $funcs($tag,fold)

  # -- prepend classname if necessary
  if { $cg_showclass && $class != "-" } {
    set name "$class::$name"
  }

  # -- show type if necessary
  if { $cg_showtype && ! [lindex $funcs($tag,file) 0] } {
    set name "<$name>"
  }
  if { $cg_showtype && $ptype == "par" } {
    set name "|| $name"
  }

  # -- mark recursive procedures
  if { $tag == $parent } {
    set name "$name **"
  }

  # -- mark nodes which are folded and have children
  if { $fold && [llength $calls] } {
    set setFold 1
    if { ! $cg_showlib } {
      set setFold 0
      foreach t $calls {
        set body [lindex $funcs($t,file) 0]
        if { $body } {
          set setFold 1
          break
        }
      }
    }
    if { $setFold } {
      set name "$name  . . ."
    }
  }

  # -- finally draw node and add bindings
  set obj [.$myself.graph.can create text $x $y \
           -text $name -fill [lindex $cg_colors $s] -anchor w]
  .$myself.graph.can bind $obj <Button-1> "toggleFold $obj $tag"
  .$myself.graph.can bind $obj <Shift-Button-1> "toggleObj $tag"
  .$myself.graph.can bind $obj <Button-2> "global_showFuncTag $tag"
  .$myself.graph.can bind $obj <Button-3> "global_selectFuncTag $tag"
  lappend funcs($tag,objs) $obj

  # -- compute size of node name
  set bbox [.$myself.graph.can bbox $obj]
  set ex [expr [lindex $bbox 2] - 10]
  set ey [expr [lindex $bbox 1] + ([lindex $bbox 3] - [lindex $bbox 1]) / 2]

  # -- draw children nodes and connecting arrows
  set drawn ""
  if { ! $fold && $tag != $parent } {
    foreach c $calls {
      set body [lindex $funcs($c,file) 0]

      if { $cg_showlib || $body } {
        # -- determine whether child is called more than once
        if { [set idx [lsearch $drawn "$c *"]] >= 0 } {
          # -- child already called from this function
          # -- (re-)label arrow with number of current call
          set info [lindex $drawn $idx]
          set num [lindex $info 4]
          incr num
          if { $num == 2 } {
            set t [.$myself.graph.can create text \
                                      [lindex $info 2] [lindex $info 3] \
                                      -text $num]
            set info [lreplace $info 1 1 $t]
          } else {
            set t [lindex $info 1]
            .$myself.graph.can itemconfigure $t -text $num
          }
          set info [lreplace $info 4 4 $num]
          set drawn [lreplace $drawn $idx $idx $info]
        } else {
          # -- child not yet called
          # -- draw arrow andf remember arrow position
          set pos [drawGraphCompact $c $tag]
          set lobj [.$myself.graph.can create line $ex $ey \
                                       [lindex $pos 0] [lindex $pos 1] \
                                       -arrow last]
          set tx [expr $ex+([lindex $pos 0]-$ex)/2]
          set ty [expr $ey+([lindex $pos 1]-$ey)/2]
          .$myself.graph.can lower $lobj
          lappend drawn [list $c {} $tx $ty 1]
        }
      }
    }
  }

  # -- draw borderless rectangle the size of the text
  # -- to make text standout from all the connecting arrows
  .$myself.graph.can create rectangle \
             [expr [lindex $bbox 0]-3] [expr [lindex $bbox 1]-3] \
             [expr [lindex $bbox 2]+3] [expr [lindex $bbox 3]+3] \
             -fill white -outline ""

  # -- raise text above last drawn rectangles
  # -- cannot done the other way round, as the size of the rectangle
  # -- depends on the text
  .$myself.graph.can raise $obj

  return [list [expr $x-10] $y]
}

#
# drawGraph: initialize data, display callgraph, and set scrollregion
#

proc drawGraph {} {
  global funcs
  global myself
  global depfile
  global cg_max_x cg_max_y
  global cg_drawmode
  global cg_cx cg_cy cg_dx cg_dy

  set tag $depfile(maintag)

  if { $cg_drawmode == "COMPACT" } {
    set cg_max_x 0
    set cg_max_y 0

    traverseGraphX $tag 0 -1
    traverseGraphY $tag
    drawGraphCompact $tag -1
    .$myself.graph.can configure -scrollregion \
                     [list 0 0 [expr 260+$cg_max_x*300] [expr 60+$cg_max_y*30]]
  } else {
    set cg_cx 50
    set cg_cy 50

    drawGraphExpd $depfile(maintag) 0 -1
    .$myself.graph.can configure -scrollregion \
                     [list 0 0 600 [expr $cg_cy-$cg_dy+50]]
  }
}

# ------------
# -- main code
# ------------
if {$argc == 2 || $argc > 3} {
  puts stderr "usage: $myself \[\[host:\]depFile \| \[host:\]directory\]"
  exit
}

if {$argc == 1} {
  set dfile [findDep [lindex $argv 0]]
} elseif {$argc == 3} {
  set depfile(root) [lindex $argv 1]
  set depfile(arch) [lindex $argv 2]
  set dfile [findDep [lindex $argv 0]]
} else {
  set dfile [findDep .]
}

if { $dfile == "NOT_OK" || [readDep -dumpcg $dfile] == "NOT_OK" } {
  exit
}
launchTAU
createWindow
set funcs($depfile(maintag),fold) 0
drawGraph
expandMenu
