#*********************************************************************#
#*                TAU/pC++/Sage++  Copyright (C) 1994                *#
#*                        Jerry Manic Saftware                       *#
#*  Indiana University  University of Oregon  University of Rennes   *#
#*********************************************************************#

source "$TAUDIR/inc/depfile.tcl"
source "$TAUDIR/inc/printcan.tcl"

set ch_maxlevel 0;     # depth of class hierarchy (maximum x coordinate)
set ch_maxpos 0;       # maximum y coordinate
set ch_showtype 1;     # mark collections with "||" ?
set ch_sel_ctag NONE;  # id of currently selected class

#
# traverseHierarchy: determine graph layout by recursively traversing it
#                    position determined by "level" (x) and "(max)pos" (y)
#
#               sub: Sage++ id of root of subgraph
#             level: current level in the graph (passed down)
#

proc traverseHierarchy {sub level} {
  global myself
  global classes ch_maxlevel ch_maxpos

  set classes($sub,level) $level
  if { $level > $ch_maxlevel } {set ch_maxlevel $level}

  if { $classes($sub,fold) } {
    # -- current node is "folded"
    incr ch_maxpos
    set mypos $ch_maxpos
  } else {
    incr level
    switch -exact [llength $classes($sub,subs)] {
    0 {
        # -- no subclasses
        # -- take next free position
        incr ch_maxpos
        set mypos $ch_maxpos
      }
    1 {
        # -- exactly one subclass
        set c [lindex $classes($sub,subs) 0]
        if [ info exists classes($c,pos) ] {
          # -- subclass has already y position elsewhere
          # -- handle like no subclass case
          incr ch_maxpos
          set mypos $ch_maxpos
        } else {
          # -- subclass has NOT yet y position
          # -- use the y coordinate of subclass for myself too
          set mypos [traverseHierarchy $c $level]
        }
      }
    default {
        # -- more than one subclass
        # -- position myself in the middle between first and last subclass
        set startpos $ch_maxpos
        foreach c $classes($sub,subs) {
          if { ! [info exists classes($c,pos)] } {
            set endpos [traverseHierarchy $c $level]
          }
        }
        set mypos [expr $startpos+($endpos-$startpos+1)/2]
      }
    }
  }
  set classes($sub,pos) $mypos

  if { $level == 0 } {
    .$myself.graph.can configure -scrollregion \
               [list 0 0 [expr 260+$ch_maxlevel*200] [expr 60+$ch_maxpos*30]]
  }
  return $mypos
}

#
# selectFuncTag: implementation of global feature "selectFunction"
#                this tool doesn't provide one
#

proc selectFuncTag {tag} {}

#
# functions to invoke in class member table
#

set selectclass(show) global_showFuncTag
set selectclass(select) global_selectFuncTag

#
# proc selectClassTag: implementation of global feature "selectClass"
#                      classy displays the selected class in red
#                      and pop up a class member table
#
#                   c: Sage++ id of class to select
#

proc selectClassTag {c} {
  global classes members selectclass
  global ch_sel_ctag myself

  # -- de-select previously selected class if necessary
  if { $ch_sel_ctag != "NONE" } {
    if [info exists classes($ch_sel_ctag,obj)] {
      .$myself.graph.can itemconfigure $classes($ch_sel_ctag,obj) -fill black
    }
  }

  # -- display selected class in red, if it is displayed
  if [info exists classes($c,obj)] {
    .$myself.graph.can itemconfigure $classes($c,obj) -fill red
  }
  set ch_sel_ctag $c
  
  # -- pop up class member table, if class has members
  if [ info exists classes($c,mem) ] {
    # -- build sorted list of member functions
    foreach m $classes($c,mem) {
      lappend mems $members($m)
    }
    set mems [lsort $mems]
    set selectclass(mem) $mems
  
    if [winfo exists .member] {
      # -- there is already a class member window; reset it
      .member.l1 delete 0 end
    } else {
      # -- create new class member window
      toplevel .member
      wm minsize .member 50 50
      listbox .member.l1 -relief sunken -geometry 60x15
      tk_listboxSingleSelect .member.l1
      scrollbar .member.s1 -orient vert -relief sunken \
                      -command ".member.l1 yview"
      .member.l1 configure -yscroll ".member.s1 set"

      frame .member.bottom
      frame .member.f1 -relief sunken -bd 1
      button .member.b1 -text "show" -command {
            set s [.member.l1 curselection]
            if { $s != "" } {
              set type [lindex [lindex $selectclass(mem) $s] 2]
              if { [string first f $type] == 1 } {
                set tag [lindex [lindex $selectclass(mem) $s] 1]
                $selectclass(show) $tag
              }
            }
          }
      pack .member.b1 -in .member.f1 -side left -padx 5 -pady 5
      pack .member.f1 -in .member.bottom -side left -padx 10 -pady 10
      button .member.b2 -text "select" -command {
            set s [.member.l1 curselection]
            if { $s != "" } {
              set type [lindex [lindex $selectclass(mem) $s] 2]
              if { [string first f $type] == 1 } {
                set tag [lindex [lindex $selectclass(mem) $s] 1]
                $selectclass(select) $tag
              }
            }
          }
      pack .member.b2 -in .member.bottom -side left -padx 15 -pady 15
      button .member.b3 -text "cancel" -command "destroy .member"
      pack .member.b3 -in .member.bottom -side right -padx 15 -pady 15

      pack .member.bottom -side bottom -fill x
      pack .member.s1 -side right -fill y
      pack .member.l1 -side left  -expand yes -fill both

      bind .member.l1 <Double-1> ".member.b1 invoke"
    }

    # -- display title
    if { $classes($c,coll) == "-" } {
      wm title .member "class  $classes($c,name)"
    } else {
      wm title .member "collection  $classes($c,name)"
    }

    # -- display formatted class member list
    foreach m $mems {
      .member.l1 insert end \
                 "[format "%-25s" [lindex $m 0]] [typeDescr [lindex $m 2]]"
    }
  }
}

#
# typeDescr: transform short typedescription into readable string
#
#      type: short form
#

proc typeDescr {type} {
  set t [split $type ""]

  switch -exact [lindex $t 0] {
    o  {set r "protected"}
    u  {set r "public   "}
    i  {set r "private  "}
    e  {set r "element  "}
  }

  switch -exact [lindex $t 4] {
    " "  { switch -exact [lindex $t 1] {
             v  {append r " variable   "}
             f  {append r " function   "}
             c  {append r " constructor"}
           }
         }
    c    {append r " constructor"}
    d    {append r " destructor "}
    o    {append r " operator   "}
  }

  if { [lindex $t 2] == "v" } {append r " virtual"}
  if { [lindex $t 3] == "i" } {append r " inline"}

  return $r
}

#
# displayHierarchy: display class hierarchy by recursively traversing graph
#                   traverseHierarchy has to be called first to compute
#                   positions!
#
#              sub: Sage++ id of root of graph to display
#

proc displayHierarchy {sub} {
  global myself
  global classes
  global ch_sel_ctag ch_showtype

  # -- compute screen coordinates from level/pos computed by traverseHierarchy
  set x [expr 30+$classes($sub,level)*200]
  set y [expr 30+$classes($sub,pos)*30]
  set name $classes($sub,name)

  # -- mark collections with "||" if necessary
  if { $ch_showtype && $classes($sub,coll) != "-" } {
    set name "|| $name"
  }

  # -- if class selected, display name in red
  if { $sub == $ch_sel_ctag } {
    set obj [.$myself.graph.can create text $x $y \
         -text $name -fill red -anchor w]
  } else {
    set obj [.$myself.graph.can create text $x $y \
         -text $name -fill black -anchor w]
  }
  .$myself.graph.can bind $obj <Button-1> "toggleClass $sub"
  .$myself.graph.can bind $obj <Button-2> "selectClassTag $sub"
  .$myself.graph.can bind $obj <Button-3> "global_selectClassTag $sub"
  set classes($sub,obj) $obj

  # -- compute center of node
  set bbox [.$myself.graph.can bbox $obj]
  set ex [expr [lindex $bbox 0] + ([lindex $bbox 2] - [lindex $bbox 0]) / 2]
  set ey [expr [lindex $bbox 1] + ([lindex $bbox 3] - [lindex $bbox 1]) / 2]

  # -- if class has subclasses, draw them and connect with lines
  if { ! $classes($sub,fold) } {
    foreach c $classes($sub,subs) {
      set pos [displayHierarchy $c]
      .$myself.graph.can create line $ex $ey [lindex $pos 0] [lindex $pos 1] \
                         -arrow last
    }
  }

  # -- draw surrounding rectangle
  .$myself.graph.can create rectangle \
             [expr [lindex $bbox 0]-3] [expr [lindex $bbox 1]-3] \
             [expr [lindex $bbox 2]+3] [expr [lindex $bbox 3]+3] \
             -fill white

  # -- if node folded, draw double-border rectangle
  if { $classes($sub,fold) && [llength $classes($sub,subs)] } {
    .$myself.graph.can create rectangle \
               [expr [lindex $bbox 0]-1] [expr [lindex $bbox 1]-1] \
               [expr [lindex $bbox 2]+1] [expr [lindex $bbox 3]+1] \
               -fill white
  }

  # -- raise text above last drawn rectangles
  # -- cannot done the other way round, as the size of the rectangle
  # -- depends on the text
  .$myself.graph.can raise $obj

  return [list [expr $x-10] $y]
}

#
# toggleClass: toggle class node between "expanded" and "folded"
#
#           c: id of class to toggle
#

proc toggleClass {c} {
  global classes myself

  set f $classes($c,fold)
  set classes($c,fold) [expr 1-$f]
  set hasSubs [llength $classes($c,subs)]

  if { $hasSubs } {
    set oldobj [expr 30+$classes($c,pos)*30]
    set oldwin [.$myself.graph.can canvasy 0]

    redrawHierarchy

    set newobj [expr 30+$classes($c,pos)*30]
    set scry [lindex [.$myself.graph.can configure -scrollincrement] 4]
    .$myself.graph.can yview [expr ($newobj-$oldobj+$oldwin)/$scry]
  }
}

#
# redrawHierarchy: redraw class hierarchy by deleteing old one,
#                  then redrawing it
#

proc redrawHierarchy {} {
  global myself
  global classes ch_maxlevel ch_maxpos
  global depfile

  # -- reset variables
  .$myself.graph.can delete all
  set ch_maxlevel 0
  set ch_maxpos 0
  foreach t $depfile(allctags) {
    if [info exists classes($t,pos)] { unset classes($t,pos) }
  }

  # -- recompute positions and display new graph
  # -- start with "virtual" root which has id -1 and level -1
  traverseHierarchy -1 -1
  foreach c $classes(-1,subs) {displayHierarchy $c}
}

#
# loadDep: switch to another application
#          this is invoked from the TAU master control window
#
#    path: UNIX pathname of depfile to read
#    root: patchname of SAGEROOT
#    arch: architecture of machine where depfile resides
#

proc loadDep {path root arch} {
  global myself
  global classes members
  global ch_sel_ctag
  global depfile

  if { ! [regexp {(.*):(.*)} $path dummy depfile(host) path] } {
    set depfile(host) "localhost"
  }
  set depfile(root) $root
  set depfile(arch) $arch

  # -- delete old class database
  unset classes members

  # -- read new data, and display new graph
  if { [readDep -dumpch $path] == "NOT_OK" } {
    exit
  }
  redrawHierarchy
  set ch_sel_ctag NONE
  
  # -- adjust "expand" menu list
  .$myself.bar.b2.m1.1 delete 0 last
  expandMenu
}

#
# createWindow: create and display main class hierarchy window
#

proc createWindow {} {
  global myself TAUDIR

  toplevel .$myself
  wm title .$myself "CLASSY"
  wm iconbitmap .$myself @$TAUDIR/xbm/$myself.xbm
  wm minsize .$myself 50 50

  frame .$myself.graph

  # -- configure graph area
  canvas .$myself.graph.can -width 600 -height 350 -background white
  bind .$myself.graph.can <2> ".$myself.graph.can scan mark %x %y"
  bind .$myself.graph.can <B2-Motion> ".$myself.graph.can scan dragto %x %y"

  frame .$myself.graph.f1

  scrollbar .$myself.graph.hscroll -orient horiz -relief sunken \
            -command ".$myself.graph.can xview"
  .$myself.graph.can configure -xscroll ".$myself.graph.hscroll set"

  frame .$myself.graph.f2 -width 20 -height 20

  pack .$myself.graph.hscroll -in .$myself.graph.f1 -side left \
                              -expand yes -fill x
  pack .$myself.graph.f2      -in .$myself.graph.f1 -side right

  scrollbar .$myself.graph.vscroll -orient vert -relief sunken \
            -command ".$myself.graph.can yview"
  .$myself.graph.can configure -yscroll ".$myself.graph.vscroll set"

  pack .$myself.graph.f1      -side bottom -fill x
  pack .$myself.graph.vscroll -side right  -fill y
  pack .$myself.graph.can     -side top -padx 15 -pady 15 -fill both -expand yes

  # -- configure menu bar
  frame .$myself.bar -relief raised -borderwidth 2

  menubutton .$myself.bar.b1 -text File -menu .$myself.bar.b1.m1 -underline 0
  menu .$myself.bar.b1.m1
  .$myself.bar.b1.m1 add command -label "Print graph" -underline 0 \
                   -command "printCanvas .$myself.graph.can $myself"
  .$myself.bar.b1.m1 add separator
  .$myself.bar.b1.m1 add command -label "Exit"  -underline 0 -command "exit"

  menubutton .$myself.bar.b2 -text View -menu .$myself.bar.b2.m1 -underline 0
  menu .$myself.bar.b2.m1
  .$myself.bar.b2.m1 add checkbutton -label "Show type of class" \
                   -underline 0 -variable ch_showtype -onvalue 1 -offvalue 0 \
                   -command "redrawHierarchy"
  .$myself.bar.b2.m1 add separator

  .$myself.bar.b2.m1 add cascade -label "Expand" -underline 0 \
                                 -menu .$myself.bar.b2.m1.1
  menu .$myself.bar.b2.m1.1

  menubutton .$myself.bar.b3 -text Help -menu .$myself.bar.b3.m1 -underline 0
  menu .$myself.bar.b3.m1
  .$myself.bar.b3.m1 add command -label "on $myself" -underline 3 \
                   -command "xsend tau \[list showHelp $myself 1-$myself 1\]"
  .$myself.bar.b3.m1 add separator
  .$myself.bar.b3.m1 add command -label "on menubar" -underline 3 \
                   -command "xsend tau \[list showHelp $myself 1.1-menu 1\]"
  .$myself.bar.b3.m1 add command -label "on display area" -underline 3 \
                   -command "xsend tau \[list showHelp $myself 1.2-display 1\]"
  .$myself.bar.b3.m1 add separator
  .$myself.bar.b3.m1 add command -label "on using help" -underline 3 \
                   -command "xsend tau {showHelp general 1-help 1}"

  pack .$myself.bar.b1 .$myself.bar.b2 -side left -padx 5
  pack .$myself.bar.b3 -side right -padx 5
  tk_menuBar .$myself.bar .$myself.bar.b1 .$myself.bar.b2 .$myself.bar.b3

  pack .$myself.bar   -side top -fill x
  pack .$myself.graph -side left -padx 15 -pady 15 -fill both -expand yes
  focus .$myself.bar
}

#
# expandMenu: update "expand" menu, so that it has an entry for each
#             level of the callgraph
#

proc expandMenu {} {
  global myself ch_maxlevel ch_exp_level

  set ch_exp_level $ch_maxlevel
  for {set i 0} {$i<=$ch_maxlevel} {incr i} {
    .$myself.bar.b2.m1.1 add radiobutton -label "upto level $i" -underline 11 \
                  -variable ch_exp_level -value $i \
                  -command "expandHierarchy $i"
  }
}

#
# expandHierarchy: expand class hierarchy after user selected level through
#                  the "expand" menu
#
#           level: level up to which graph has to be expanded
#

proc expandHierarchy {level} {
  global depfile
  global classes

  foreach t $depfile(allctags) {
    if { $classes($t,level) < $level } {
      set classes($t,fold) 0
    } elseif { $classes($t,level) == $level } {
      set classes($t,fold) 1
    }
  }
  redrawHierarchy
}

# ------------
# -- main code
# ------------
if {$argc == 2 || $argc > 3} {
  puts stderr "usage: $myself \[\[host:\]depFile \| \[host:\]directory\]"
  exit
}

if {$argc == 1} {
  set dfile [findDep [lindex $argv 0]]
} elseif {$argc == 3} {
  set depfile(root) [lindex $argv 1]
  set depfile(arch) [lindex $argv 2]
  set dfile [findDep [lindex $argv 0]]
} else {
  set dfile [findDep .]
}

if { $dfile == "NOT_OK" || [readDep -dumpch $dfile] == "NOT_OK" } {
  exit
}
launchTAU
createWindow

# -- compute positions and draw class hierarchy
# -- start with "virtual" root which has id -1 and level -1
traverseHierarchy -1 -1
foreach c $classes(-1,subs) {displayHierarchy $c}
expandMenu
