#*********************************************************************#
#*                TAU/pC++/Sage++  Copyright (C) 1994                *#
#*                        Jerry Manic Saftware                       *#
#*  Indiana University  University of Oregon  University of Rennes   *#
#*********************************************************************#

source "$TAUDIR/inc/depfile.tcl"

set cosy_var(make) "norm";     # program "type": prf-, trc-, or, brk-
set cosy_var(warn) 0;          # show warnings?
set cosy_var(com) "";          # holds current command to execute

#
# machineMenu: generate menu item which represent available executables
#              for the current machine
#
#        menu: pathname of menu widget
#         com: command to execute for each item
#

proc machineMenu {menu com} {
  global cosy_var

  $menu delete 0 last
  foreach m $cosy_var(machines) {
    $menu add command -label $m -underline 0 -command "$com $m"
  }
}

#
# setMachines: determine available executables for current machines
#

proc setMachines {} {
  global depfile
  global cosy_var

  # -- uniprocessor mode always possible
  set cosy_var(machines) "uniproc"

  # -- check if running on multiprocessor
  switch -exact $depfile(arch) {
    ptx           -
    symmetry      { lappend cosy_var(machines) symmetry }
    cm5           { lappend cosy_var(machines) cm5 }
    paragon       { lappend cosy_var(machines) paragon }
    tc2000        { lappend cosy_var(machines) tc2000 }
    ksr1          { lappend cosy_var(machines) ksr }
  }

  # -- check for cross-development environments
  switch -exact $depfile(arch) {
    rs6000    -
    iris4d    -
    solaris2  -
    sun4      {
                set xdev [getOutput "$depfile(root)/utils/archfind -x"]
                switch -exact $xdev {
                  sp1     { lappend cosy_var(machines) sp1euih }
                  cs2     { lappend cosy_var(machines) cs2.nx }
                  sgimp   { lappend cosy_var(machines) sgimp }
                  cm5     { lappend cosy_var(machines) cm5 }
                  paragon { lappend cosy_var(machines) paragon }
                }
              }
  }

  # -- get local hostname if necessary
  set cosy_var(host) ""
  if { $depfile(host) == "localhost" } {
    if [catch {exec uname -n} cosy_var(host)] {
      [catch {exec hostname} cosy_var(host)]
    }
  } else {
    set cosy_var(host) $depfile(host)
  }
}

#
# createWindow: create main application window
#

proc createWindow {} {
  global TAUDIR
  global myself
  global cosy_var

  toplevel .$myself
  wm title .$myself "COSY"
  wm minsize .$myself 50 50
  wm iconbitmap .$myself @$TAUDIR/xbm/$myself.xbm

  # -- menu bar
  frame .$myself.mbar -relief raised -borderwidth 2
  menubutton .$myself.mbar.b1 -text File -menu .$myself.mbar.b1.m1 -underline 0
  menu .$myself.mbar.b1.m1
  .$myself.mbar.b1.m1 add cascade -label "MAKE" -underline 0 \
                          -menu .$myself.mbar.b1.m1.1 \
                          -command "machineMenu .$myself.mbar.b1.m1.1 doMake"
  .$myself.mbar.b1.m1 add cascade -label "make" -underline 0 \
                          -menu .$myself.mbar.b1.m1.2
  .$myself.mbar.b1.m1 add cascade -label "Run" -underline 0 \
                          -menu .$myself.mbar.b1.m1.3 \
                          -command "machineMenu .$myself.mbar.b1.m1.3 doRun"
  .$myself.mbar.b1.m1 add command -label "list directory" -underline 5 \
                          -command "doExec {ls -l}"
  .$myself.mbar.b1.m1 add command -label "list queues" -underline 5 \
                          -command "doQueue"
  .$myself.mbar.b1.m1 add separator
  .$myself.mbar.b1.m1 add command -label "Exit" -underline 0 -command "exit"

  menu .$myself.mbar.b1.m1.1
  machineMenu .$myself.mbar.b1.m1.1 "doMake"
  menu .$myself.mbar.b1.m1.3
  machineMenu .$myself.mbar.b1.m1.3 "doRun"

  menu .$myself.mbar.b1.m1.2
  .$myself.mbar.b1.m1.2 add command -label "clean" -underline 0 \
                                    -command "doExec {make clean}"
  .$myself.mbar.b1.m1.2 add command -label "realclean" -underline 0 \
                                    -command "doExec {make realclean}"

  menubutton .$myself.mbar.b2 -text Options -menu .$myself.mbar.b2.m1 \
                              -underline 0
  menu .$myself.mbar.b2.m1
  .$myself.mbar.b2.m1 add command -label "set parameters ..." \
                 -underline 4 -command setParameter
  .$myself.mbar.b2.m1 add separator
  .$myself.mbar.b2.m1 add radiobutton -label "normal" \
                 -underline 0 -variable cosy_var(make) -value "norm"
  .$myself.mbar.b2.m1 add radiobutton -label "profiling" \
                 -underline 0 -variable cosy_var(make) -value "prf-"
  .$myself.mbar.b2.m1 add radiobutton -label "tracing" \
                 -underline 0 -variable cosy_var(make) -value "trc-"
  .$myself.mbar.b2.m1 add radiobutton -label "breakpointing" \
                 -underline 0 -variable cosy_var(make) -value "brk-"
  .$myself.mbar.b2.m1 add separator
  .$myself.mbar.b2.m1 add checkbutton -label "show warnings" \
                 -underline 0 -variable cosy_var(warn) -onvalue 1 -offvalue 0

  menubutton .$myself.mbar.b3 -text Help -menu .$myself.mbar.b3.m1 -underline 0
  menu .$myself.mbar.b3.m1
  .$myself.mbar.b3.m1 add command -label "on $myself" -underline 3 \
                 -command "xsend tau \[list showHelp $myself 1-$myself 1\]"
  .$myself.mbar.b3.m1 add separator
  .$myself.mbar.b3.m1 add command -label "on File menu" -underline 3 \
                 -command "xsend tau \[list showHelp $myself 1.1-file 1\]"
  .$myself.mbar.b3.m1 add command -label "on Options menu" -underline 3 \
                 -command "xsend tau \[list showHelp $myself 1.2-options 1\]"
  .$myself.mbar.b3.m1 add separator
  .$myself.mbar.b3.m1 add command -label "on using help" -underline 3 \
                 -command "xsend tau {showHelp general 1-help 1}"

  # -- command line display
  frame .$myself.top
  frame .$myself.left
  label .$myself.l1 -text "executing:"
  pack .$myself.l1 -side top -anchor e -in .$myself.left

  frame .$myself.right
  label .$myself.r1 -textvariable cosy_var(com)
  pack .$myself.r1 -side top -anchor w -in .$myself.right
  pack .$myself.left .$myself.right -side left -in .$myself.top

  # -- command output display
  text .$myself.txt -width 80 -height 12 -background white -foreground black
  scrollbar .$myself.s1 -orient vert -relief sunken \
                        -command ".$myself.txt yview"
  .$myself.txt configure -yscroll ".$myself.s1 set"

  pack .$myself.mbar.b1 .$myself.mbar.b2 -side left -padx 5
  pack .$myself.mbar.b3 -side right -padx 5
  pack .$myself.mbar -side top -fill x
  tk_menuBar .$myself.mbar .$myself.mbar.b1 .$myself.mbar.b2 .$myself.mbar.b3

  pack .$myself.top -side top -anchor w -padx 5 -pady 5
  pack .$myself.s1 -side right -fill y
  pack .$myself.txt -side top -padx 10 -pady 10 -expand yes -fill both

  bind .$myself <Enter> "focus .$myself.mbar"
  focus .$myself.mbar
}

#
# doMake: execute pC++ MAKE command
#
#   what: executable to make
#

proc doMake {what} {
  global cosy_var
  global depfile

  # -- execute right MAKE command
  if { $cosy_var(make) == "norm" } {
    doExec "MAKE ${what}"
  } else {
    doExec "MAKE $cosy_var(make)${what}"
  }

  # -- if no error occurred, signal other tools to reload depfile
  if { ! $cosy_var(err) } {
    if { $depfile(host) == "localhost" } {
      xsend tau {loadDepUncond $depfile(path) 1 0}
    } else {
      xsend tau {loadDepUncond $depfile(host):$depfile(path) 0 0}
    }
  }
}

#
# doRun: run compiled executable
#
#  what: binary to run
#

proc doRun {what} {
  global cosy_var cosy_old_val
  global depfile
  global BINDIR

  # -- set program and number of nodes to use
  if { $what == "uniproc" } {
    set numproc 1
    set param ""
  } else {
    set numproc $cosy_var(numproc)
    set param "-pcxx_NUMPROC $cosy_var(numproc)"
  }
  set prog "$cosy_var(make)$what"

  # -- what 

  # -- execute binary
  switch $cosy_var(make) {
    norm   { set prog "$what" }
    prf-   { }

    trc-   { # -- TF is user-specified tracefile, if specified
             set TF ""
             if { [info exists cosy_var(tracefile)] &&
                  $cosy_var(tracefile) != "" } {
               set TF "-pcxx_TRACEFILE '$cosy_var(tracefile)'"
             }

             # -- EC is user-specified event classes
             set EC ""
             foreach n [array names cosy_old_val] {
               if { [regexp ^p_ $n] && $cosy_var($n) != "" } {
                 if { $EC != "" } {
                   append EC "+$cosy_var($n)"
                 } else {
                   set EC "$cosy_var($n)"
                 }
               }
             }
             if { $EC != "" } {set EC "-pcxx_EVENTCLASS $EC"}
             append param "$TF $EC"
           }
    brk-   { # for breakpointing start breezy user interface first
             exec $BINDIR/breezy &

             # then start program, with hostname if necessary
             if { $depfile(host) != "localhost" } {
               append param "-pcxx_BRKHOST $cosy_var(host)"
             }
           }
  }

  # -- does user-defined run command exist?
  set execfunc "exec_$cosy_var(host)"
  if { [ info procs $execfunc ] == "$execfunc" } {
    doExec [$execfunc $prog $numproc $param]
  } elseif { [ info procs exec_default ] == "exec_default" } {
    doExec [exec_default $prog $numproc $param]
  } else {
    doExec "$prog $param"
  }
}

#
# doQueue: list run queues if host has batch queue system
#

proc doQueue {} {
  global cosy_var

  # -- does user-defined queue command exist?
  set execfunc "queue_$cosy_var(host)"
  if { [ info procs $execfunc ] == "$execfunc" } {
    doExec [$execfunc]
  } elseif { [ info procs queue_default ] == "queue_default" } {
    doExec [queue_default]
  } else {
    showError "no queue command defined for this host."
  }
}

#
# doExec: execute (local or remote) command
#
#   what: command to execute
#

proc doExec {what} {
  global myself
  global cosy_var depfile

  # -- init variables and reset command output display
  set cosy_var(err) 0
  set cosy_var(com) "[string range $what 0 62] ..."
  .$myself.txt delete 1.0 end
  update

  # -- start program in (remote) shell
  if { $depfile(host) != "localhost" } {
    set in [open "|remsh $depfile(host) -n \"cd $depfile(dir); $what\"" r]
  } else {
    set in [open "|sh -c \"$what 2>&1\"" r]
  }

  # -- read output of command
  while {[gets $in line] >= 0} {
    # -- tyest for errors in putput
    if { [regexp -nocase error $line] || [regexp "up to date" $line] } {
      set cosy_var(err) 1
    }

    # -- suppress warnings, if specified
    if { $cosy_var(warn) || [regexp -nocase warning $line] == 0 } {
      .$myself.txt insert end "$line\n"
      .$myself.txt yview -pickplace end
      update
    }
  }
  # -- close subshell and catch/display  error message if necessary
  if [catch {close $in} errmsg] {
    .$myself.txt insert end "$errmsg\n"
    .$myself.txt yview -pickplace end
    update
  }
  set cosy_var(com) "[string range $what 0 62] ... done"
  update
}

set cosy_old_val(numproc) 16;  # default parameter for number of nodes
set cosy_var(numproc) 16;      # previous value of this parameter

#
# setParameter: user dialog window to allow compile and run parameter settings
#

proc setParameter {} {
  global cosy_var
  global cosy_old_val

  set cosy_var(oldfocus) [focus]

  if [winfo exists .param] {
    raise .param
  } else {
    toplevel .param
    wm title .param "Set Runtime Parameters"

    frame .param.top
    frame .param.tleft
    label .param.tl1 -text "NUMPROC:"
    label .param.tl2 -text "TRACEFILE:"
    pack .param.tl1 .param.tl2 -side top -anchor e -in .param.tleft -padx 5

    frame .param.tright
    entry .param.tr1 -textvariable cosy_old_val(numproc) -relief sunken
    entry .param.tr2 -textvariable cosy_old_val(tracefile) -relief sunken
    pack .param.tr1 .param.tr2 -side top -anchor w -in .param.tright

    pack .param.tleft .param.tright -side left -in .param.top
    pack .param.top -side top -padx 5 -pady 5

    frame .param.mid
    label .param.midl1 -text "EVENTCLASS:"
    pack .param.midl1 -side left  -in .param.mid -padx 5

    frame .param.midleft
    checkbutton .param.midleft.l1 -text "basic" \
                                  -variable cosy_old_val(p_basic) \
                                  -onvalue "B" -offvalue "" -relief flat
    checkbutton .param.midleft.l2 -text "kernel" \
                                  -variable cosy_old_val(p_kernel) \
                                  -onvalue "K" -offvalue "" -relief flat
    checkbutton .param.midleft.l3 -text "runtime" \
                                  -variable cosy_old_val(p_runtime) \
                                  -onvalue "R" -offvalue "" -relief flat
    checkbutton .param.midleft.l4 -text "timers" \
                                  -variable cosy_old_val(p_timers) \
                                  -onvalue "Ti" -offvalue "" -relief flat
    checkbutton .param.midleft.l5 -text "trace" \
                                  -variable cosy_old_val(p_trace) \
                                  -onvalue "Tr" -offvalue "" -relief flat
    pack .param.midleft.l1 .param.midleft.l2 .param.midleft.l3\
         .param.midleft.l4 .param.midleft.l5 -side top -anchor w
    pack .param.midleft -side left -padx 10 -pady 10 -in .param.mid

    frame .param.midright
    checkbutton .param.midright.r1 -text "user1" \
                                  -variable cosy_old_val(p_user1) \
                                  -onvalue "1" -offvalue "" -relief flat
    checkbutton .param.midright.r2 -text "user2" \
                                  -variable cosy_old_val(p_user2) \
                                  -onvalue "2" -offvalue "" -relief flat
    checkbutton .param.midright.r3 -text "user3" \
                                  -variable cosy_old_val(p_user3) \
                                  -onvalue "3" -offvalue "" -relief flat
    checkbutton .param.midright.r4 -text "user4" \
                                  -variable cosy_old_val(p_user4) \
                                  -onvalue "4" -offvalue "" -relief flat
    checkbutton .param.midright.r5 -text "profile" \
                                  -variable cosy_old_val(p_profile) \
                                  -onvalue "P" -offvalue "" -relief flat
    pack .param.midright.r1 .param.midright.r2 .param.midright.r3\
         .param.midright.r4 .param.midright.r5 -side top -anchor w
    pack .param.midright -side left -padx 10 -pady 10 -in .param.mid
    pack .param.mid -side top

    frame .param.bottom -relief groove -bd 2
    frame .param.frame -relief sunken -bd 1
    button .param.b1 -text "set" -command {
      foreach n [array names cosy_old_val] {
        set cosy_var($n) $cosy_old_val($n)
      }
      focus $cosy_var(oldfocus)
      destroy .param
    }
    button .param.b2 -text "cancel" -command {
      foreach n [array names cosy_old_val] {
        if [info exists cosy_var($n)] {set cosy_old_val($n) $cosy_var($n)} }
      focus $cosy_var(oldfocus)
      destroy .param
    }
    pack .param.b1 -side top -ipadx 5 -padx 5 -pady 5 -in .param.frame
    pack .param.frame -side left -padx 15 -pady 10 -in .param.bottom
    pack .param.b2 -side right -ipadx 5 -padx 15 -in .param.bottom
    pack .param.bottom -side bottom -fill x -expand yes
  }
}

#
# initialize: initialize application
#
#        dir: next working directory
#       root: SAGEROOT of current machine
#       arch: architecture of current machine
#

proc initialize {dir {root "???"} {arch "???"}} {
  global depfile
  global myself
  global cosy_var

  set depfile(root) $root
  set depfile(arch) $arch

  # -- set global info about the current application (represented through
  # -- its depfile
  if { [set dfile [findDep $dir]] == "NOT_OK" } {
    SetDepfileBasics $dir
  } else {
    SetDepfileBasics $dfile
  }

  # -- if not remote, set working directory to new application directory
  if { $depfile(host) == "localhost" } { cd $depfile(dir) }

  # -- reset command output display
  if [winfo exists .$myself.txt] {
    .$myself.txt delete 1.0 end
  }
  set cosy_var(com) ""

  # -- determine available executables of current machine
  setMachines
}

# ------------
# -- main code
# ------------
if {$argc == 2 || $argc > 3} {
  puts stderr "usage: $myself \[\[host:\]directory\]"
  exit
}

if {$argc == 1} {
  if { [initialize [lindex $argv 0]] == "NOT_OK" } { exit }
} elseif {$argc == 3} {
  if { [initialize [lindex $argv 0] [lindex $argv 1] [lindex $argv 2]] \
       == "NOT_OK" } { exit }
} else {
  if { [initialize .] == "NOT_OK" } { exit }
}

# -- read user rc file
if [ file readable $env(HOME)/.cosyrc ] {
  if [catch "source $env(HOME)/.cosyrc" errmsg] {
    tkerror "error reading $env(HOME)/.cosyrc"
    if [winfo exists .tkerrorTrace] {
      tkwait window .tkerrorTrace
    }
  }
}

# -- launch tau if not already running
launchTAU

# -- create new toplevel window
createWindow
