#*********************************************************************#
#*                TAU/pC++/Sage++  Copyright (C) 1994                *#
#*                        Jerry Manic Saftware                       *#
#*  Indiana University  University of Oregon  University of Rennes   *#
#*********************************************************************#

source "$TAUDIR/inc/fileselect.tcl"
source "$TAUDIR/inc/getFile.tcl"
source "$TAUDIR/inc/depfile.tcl"
source "$TAUDIR/inc/selectfunc.tcl"
source "$TAUDIR/inc/stack.tcl"
source "$TAUDIR/inc/help.tcl"

set selectfunc(show) showFuncTag
set selectfunc(select) global_selectFuncTag

set ${myself}winnr 0

#
# showFuncTag: implementation of global feature "showFunction" for this tool
#              show small function information window in form of a table
#
#         tag: id of function to show
#

proc showFuncTag {tag} {
  global myself
  global funcs
  global ${myself}winnr

  if { ! [info exists funcs($tag,name)] } {
    showError "No info about tag $tag."
    return
  } elseif { ! [winfo exists .$myself.t$tag] } {
    set name  $funcs($tag,name)
    set class $funcs($tag,class)
    set file  $funcs($tag,file)
    set ptype [lindex $funcs($tag,type) 0]
    set type  [lindex $funcs($tag,type) 1]
    set num   $funcs($tag,num)
    set calls [llength $funcs($tag,calls)]

    if { [lindex $class 0] == "-" } {
      set class ""
    } elseif { [lindex $class 1] == "-" } {
      set class [lindex $class 0]
    } else {
      set class "Collection [lindex $class 0]"
    }

    if { [lindex $file 0] == "0" } {
      set file ""
    } else {
      set file "[lindex $file 1]:[lindex $file 0]"
    }

    switch $ptype {
      seq { set ptype "sequential" }
      par { set ptype "parallel" }
    }

    switch $type {
      Mem { set type "Member Function" }
      MoE { set type "Method of Element" }
      MoC { set type "Member of Collection" }
      Ord { set type "Ordinary Function" }
    }

    set w [toplevel .$myself.t$tag]
    wm title $w "Function Info"

    button $w.b -text "close" -command "destroy $w"
    frame $w.l
    frame $w.r

    pack $w.b -side bottom -fill x
    pack $w.l $w.r -side left -padx 5 -pady 5

    label $w.l1 -text "name:"
    label $w.l2 -text "class:"
    label $w.l3 -text "file:"
    label $w.l4 -text "type:"
    label $w.l5 -text "call sites:"
    label $w.l6 -text "calls:"
    pack $w.l1 $w.l2 $w.l3 $w.l4 $w.l5 $w.l6 -in $w.l -side top -anchor e

    label $w.r1 -text $name -relief sunken -anchor w
    label $w.r2 -text $class -relief sunken -anchor w
    label $w.r3 -text $file -relief sunken -anchor w
    label $w.r4 -text "$ptype $type" -relief sunken -anchor w
    label $w.r5 -text $num -relief sunken -anchor w
    label $w.r6 -text $calls -relief sunken -anchor w
    pack $w.r1 $w.r2 $w.r3 $w.r4 $w.r5 $w.r6 -in $w.r -side top -fill x
  } else {
    raise .$myself.t$tag
  }
}

#
# selectFuncTag: implementation of global feature "selectFunction"
#                this tool doesn't provide one
#

proc selectFuncTag {tag} {}

#
# getRemoteFile: ask user for new remote file
#

proc getRemoteFile {} {
  global depfile
  global myself
  global ${myself}file

  if [winfo exists .$myself.loc] {
    raise .$myself.loc
  } else {
    set ${myself}file(host) $depfile(host)
    set ${myself}file(path) $depfile(path)

    toplevel .$myself.loc
    wm title .$myself.loc "Set Host and Pathname"
    grab .$myself.loc

    frame .$myself.loc.bottom -relief groove -bd 2
    frame .$myself.loc.frame -relief sunken -bd 1
    button .$myself.loc.b1 -text "set" -command {
      set depfile(host) [expr $${myself}file(host)]
      set depfile(path) [expr $${myself}file(path)]
      destroy .$myself.loc
    }
    button .$myself.loc.b2 -text "cancel" -command {
      set ${myself}file(host) $depfile(host)
      set ${myself}file(path) $depfile(path)
      destroy .$myself.loc
    }
    pack .$myself.loc.b1 -side top -ipadx 5 -padx 5 -pady 5 \
                         -in .$myself.loc.frame
    pack .$myself.loc.frame -side left -padx 15 -pady 10 \
                            -in .$myself.loc.bottom
    pack .$myself.loc.b2 -side right -ipadx 5 -padx 15 -in .$myself.loc.bottom
    pack .$myself.loc.bottom -side bottom -fill x -expand yes

    frame .$myself.loc.left
    label .$myself.loc.l1 -text "host:"
    label .$myself.loc.l2 -text "path:"
    pack .$myself.loc.l1 .$myself.loc.l2 -side top -anchor e \
                                         -in .$myself.loc.left -padx 5

    frame .$myself.loc.right
    entry .$myself.loc.r1 -textvariable ${myself}file(host) -relief sunken \
                          -width 40
    entry .$myself.loc.r2 -textvariable ${myself}file(path) -relief sunken \
                          -width 40
    pack .$myself.loc.r1 .$myself.loc.r2 -side top -anchor w \
                                         -in .$myself.loc.right

    pack .$myself.loc.left .$myself.loc.right -side left -padx 10 -pady 10
    bind .$myself.loc.r1 <Return> {.$myself.loc.b1 invoke}
    bind .$myself.loc.r2 <Return> {.$myself.loc.b1 invoke}
  }
}

#
# loadDep: let the user switch to another application by loading new depfile
#          it is a global feature
#
#   local: load local or remote depfile?
#

proc loadDep {local} {
  global myself
  global depfile

  # -- display file selector box and get user selected file
  set oldfocus [focus]
  if { $local } {
    set path [getFile "Select .dep File" *.dep]
    file stat $path statbuf
    if { $statbuf(ino) == $depfile(ino) } {
      return
    }
    # -- file is now different from current one, load it;
    # -- comparing inode numbers is not save, but better than
    # -- comparing the filenames
  } else {
    set oldpath "$depfile(host):$depfile(path)"
    getRemoteFile
    tkwait window .$myself.loc
    set path "$depfile(host):$depfile(path)"
    # -- don't know yet how to get the inode number for remote files
    # -- so compare pathnames
    if { $path == $oldpath } {
      return
    }
  }
  focus $oldfocus

  loadDepUncond $path $local 1
}

#
# loadDepUncond: load new depfile
#
#          path: depfile to load
#         local: depfile on local host?
#     startcosy: update cosy?
#

proc loadDepUncond {path local startcosy} {
  global myself
  global depfile

  displayMessage hourglass "loading file: ..."

  # -- reset architecture, so it is determined again
  set depfile(arch) "???"

  # -- get rid of old "show Function" windows
  foreach t $depfile(allftags) {
    if [winfo exists .$myself.t$t] { destroy .$myself.t$t }
  }

  set dfile [findDep $path]
  if { $dfile == "NOT_OK" } {
    # -- no depfie available, but pC++ SOURCE directory
    SetDepfileBasics $path
    if { $depfile(host) == "localhost" } { cd $depfile(dir) }
    getLocalHost

    # -- update cosy, if running
    if { $startcosy } {
      if { $local } {
        xsend cosy "initialize $depfile(dir) $depfile(root) $depfile(arch)"
      } else {
        xsend cosy "initialize $depfile(host):$depfile(dir) \
                               $depfile(root) $depfile(arch)"
      }
    }

    # -- close other TAU tools, as no depfile available
    ysend racy "after 1 exit"
    ysend cagey "after 1 exit"
    ysend classy "after 1 exit"
    ysend fancy "after 1 exit"

  } elseif { [readDep -dumpall $dfile] != "NOT_OK" } {
    # -- depfile could be read, update select function listbox if necessary
    if [ winfo exists .select ] {
      selectFunction
    }
    getLocalHost

    # -- load depfile on all other running TAU applications too
    if { $startcosy } {
      if { $local } {
        xsend cosy "initialize $depfile(dir) $depfile(root) $depfile(arch)"
      } else {
        xsend cosy "initialize $depfile(host):$depfile(dir) \
                               $depfile(root) $depfile(arch)"
      }
    }
    xsend cagey "loadDep $path $depfile(root) $depfile(arch)"
    xsend classy "loadDep $path $depfile(root) $depfile(arch)"
    xsend fancy "loadDep $path $depfile(root) $depfile(arch)"
    if { $local } {
      xsend racy "loadProfile $depfile(dir) $depfile(root) $depfile(arch)"
    } else {
      xsend racy "loadProfile $depfile(host):$depfile(dir) \
                              $depfile(root) $depfile(arch)"
    }
  }
  removeMessage
}

#
# bitmapbutton: button widget which has a text AND a bitmap
#
#          win: pathname for new button
#       bitmap: bitmap to display
#         text: text label to display
#      command: command to execute if button is pressed
#

proc bitmapbutton {win bitmap text command} {
  frame $win
  button $win.b -bitmap "$bitmap" -command "$command" -back white -fore black
  button $win.t -text "$text" -command "$command" -back white -fore black
  bind $win.t <ButtonPress-1> "tk_butDown %W"
  bind $win.t <ButtonRelease-1> "tk_butUp %W"
  bind $win.t <Enter> "tk_butEnter %W; $win.b configure -state active"
  bind $win.t <Leave> "tk_butLeave %W; $win.b configure -state normal"
  bind $win.b <ButtonPress-1> "tk_butDown %W"
  bind $win.b <ButtonRelease-1> "tk_butUp %W"
  bind $win.b <Enter> "tk_butEnter %W; $win.t configure -state active"
  bind $win.b <Leave> "tk_butLeave %W; $win.t configure -state normal"
  pack $win.b $win.t -side top -fill x
  return $win
}

#
# createWindow: create and display main window for TAU master control
#

proc createWindow {} {
  global myself
  global TAUDIR
  global depfile

  toplevel .$myself
  wm title .$myself "TAU"
  wm iconbitmap .$myself @$TAUDIR/xbm/$myself.xbm

  # -- menubar
  frame .$myself.bar -relief raised -borderwidth 2
  menubutton .$myself.bar.b1 -text File -menu .$myself.bar.b1.m1 -underline 0
  menu .$myself.bar.b1.m1
  .$myself.bar.b1.m1 add command -label "Load local depfile" -underline 5 \
                   -command "loadDep 1"
  .$myself.bar.b1.m1 add command -label "Load remote depfile" -underline 5 \
                   -command "loadDep 0"
  .$myself.bar.b1.m1 add command -label "Select function" -underline 0 \
                   -command "selectFunction"
  .$myself.bar.b1.m1 add separator
  .$myself.bar.b1.m1 add command -label "Exit"  -underline 0 -command "exit"
  menubutton .$myself.bar.b2 -text Help -menu .$myself.bar.b2.m1 -underline 0
  menu .$myself.bar.b2.m1
  .$myself.bar.b2.m1 add command -label "on $myself" -underline 3 \
                   -command "showHelp $myself 1-$myself 1"
  .$myself.bar.b2.m1 add separator
  .$myself.bar.b2.m1 add command -label "on menubar" -underline 3 \
                   -command "showHelp $myself 1.1-menu 1"
  .$myself.bar.b2.m1 add command -label "on display area" -underline 3 \
                   -command "showHelp $myself 1.2-display 1"
  .$myself.bar.b2.m1 add command -label "on command buttons" -underline 3 \
                   -command "showHelp $myself 1.3-buttons 1"
  .$myself.bar.b2.m1 add separator
  .$myself.bar.b2.m1 add command -label "on using help" -underline 3 \
                   -command "showHelp general 1-help 1"
  tk_menuBar .$myself.bar .$myself.bar.b1 .$myself.bar.b2

  pack .$myself.bar.b1 -side left -padx 5
  pack .$myself.bar.b2 -side right -padx 5
  pack .$myself.bar -side top -fill x

  # -- display area
  frame .$myself.info1
  frame .$myself.info1.left
  label .$myself.info1.l1 -text "host:"
  label .$myself.info1.l2 -text "dir:"
  pack .$myself.info1.l1 .$myself.info1.l2 \
                      -side top -anchor e -in .$myself.info1.left
  pack .$myself.info1.left -side left -padx 5

  frame .$myself.info1.right
  label .$myself.info1.r1 -textvariable depfile(hostarch) -relief sunken \
        -anchor w
  label .$myself.info1.r2 -textvariable depfile(dir) -relief sunken -anchor w
  pack .$myself.info1.r1 .$myself.info1.r2 \
       -side top -anchor w -in .$myself.info1.right -fill x -expand 1 -ipadx 10
  pack .$myself.info1.right -side right -padx 5 -fill x -expand 1

  frame .$myself.info2
  frame .$myself.info2.left
  label .$myself.info2.l1 -text "File:"
  label .$myself.info2.l2 -text "Functions defined:"
  label .$myself.info2.l3 -text "Functions used:"
  label .$myself.info2.l4 -text "Functions profiled:"
  pack .$myself.info2.l1 .$myself.info2.l2 .$myself.info2.l3 .$myself.info2.l4 \
                      -side top -anchor e -in .$myself.info2.left
  pack .$myself.info2.left -side left -padx 5

  frame .$myself.info2.right
  label .$myself.info2.r1 -textvariable depfile(file) -relief sunken -anchor w
  label .$myself.info2.r2 -textvariable depfile(numfunc) -relief sunken \
        -anchor w
  label .$myself.info2.r3 -textvariable depfile(numused) -relief sunken \
        -anchor w
  label .$myself.info2.r4 -textvariable depfile(numprof) -relief sunken \
        -anchor w
  pack .$myself.info2.r1 .$myself.info2.r2 .$myself.info2.r3 .$myself.info2.r4 \
       -side top -anchor w -in .$myself.info2.right -fill x -expand 1 -ipadx 30
  pack .$myself.info2.right -side right -padx 5 -fill x -expand 1

  pack .$myself.info1 .$myself.info2 -side top -padx 10 -pady 10 -fill x

  # -- command button bar
  frame .$myself.coms
  bitmapbutton .$myself.b0 @$TAUDIR/xbm/cosy.xbm cosy "launch cosy"
  bitmapbutton .$myself.b1 @$TAUDIR/xbm/fancy.xbm fancy "launch fancy"
  bitmapbutton .$myself.b2 @$TAUDIR/xbm/cagey.xbm cagey "launch cagey"
  bitmapbutton .$myself.b3 @$TAUDIR/xbm/classy.xbm classy "launch classy"
  bitmapbutton .$myself.b4 @$TAUDIR/xbm/racy.xbm racy "launch racy"
  pack .$myself.b0 .$myself.b1 .$myself.b2 .$myself.b3 .$myself.b4\
                   -side left -padx 10 -pady 5 -in .$myself.coms
  pack .$myself.coms -side bottom -pady 10

  bind .$myself <Destroy> atExit
  bind .$myself <Enter> "focus .$myself.bar"
  focus .$myself.bar
}

#
# launch: invoke another TAU tool
#
#   prog: TAU tool to invoke
#

proc launch {prog} {
  global depfile
  global BINDIR
  global myname

  set avail [winfo interps]
  if { [ lsearch -exact $avail $prog$myname(ext) ] == -1 } {
    if { $depfile(host) == "localhost" } {
      set file "$depfile(file)"
    } else {
      set file "$depfile(host):$depfile(path)"
    }
    # -- not yet running, invoke new tool
    switch -exact $prog {
      fancy {
          if { $depfile(path) == "" } {
            showError "Load depfile first."
          } else {
            exec $BINDIR/fancy $file $depfile(root) $depfile(arch) &
          }
      }

      racy   {
          if { $depfile(host) == "localhost" } {
            if { ! [file readable ./profile.ftab] } {
              showError "No profile data available for $depfile(file)."
            } else {
              exec $BINDIR/racy . $depfile(root) $depfile(arch) &
            }
          } else {
            exec $BINDIR/racy $depfile(host):$depfile(dir) \
                              $depfile(root) $depfile(arch) &
          }
      }

      cagey  {
          exec $BINDIR/cagey $file $depfile(root) $depfile(arch) &
      }

      classy {
          exec $BINDIR/classy $file $depfile(root) $depfile(arch) &
      }

      cosy {
          if { $depfile(host) == "localhost" } {
            exec $BINDIR/cosy . $depfile(root) $depfile(arch) &
          } else {
            exec $BINDIR/cosy $depfile(host):$depfile(dir) \
                              $depfile(root) $depfile(arch) &
          }
      }
    }
  } else {
    # -- already running, bring it in front (raise)
    xsend $prog "raise .$prog"
  }
}

#
# atExit: cleanup procedure; makes sure all launched tools are exited
#

proc atExit {} {
  ysend racy "after 1 exit"
  ysend cagey "after 1 exit"
  ysend classy "after 1 exit"
  ysend fancy "after 1 exit"
  ysend cosy "after 1 exit"
}

#
# getLocalHost: get hostname of computer
#

proc getLocalHost {} {
  global depfile

  if { $depfile(host) == "localhost" } {
    if [catch {exec uname -n} h] {
      if { ![catch {exec hostname} h] } {
        set depfile(hostarch) "$h ($depfile(arch))"
      } else {
        set depfile(hostarch) "localhost"
      }
    } else {
      set depfile(hostarch) "$h ($depfile(arch))"
    }
  } else {
    set depfile(hostarch) "$depfile(host) ($depfile(arch))"
  }
}

# ------------
# -- main code
# ------------

if {$argc == 0} {
  set dfile [findDep .]
} elseif {$argc == 1} {
  set dfile [findDep [lindex $argv 0]]
} elseif {$argc == 3} {
  set depfile(root) [lindex $argv 1]
  set depfile(arch) [lindex $argv 2]
  set dfile [findDep [lindex $argv 0]]
} else {
  puts stderr "usage: $myself \[\[host:\]depFile \| \[host:\]directory\]"
  exit
}
if { $dfile == "NOT_OK" } {
  if {$argc == 0} {
    SetDepfileBasics .
  } else {
    SetDepfileBasics [lindex $argv 0]
  }
  if { $depfile(host) == "localhost" } { cd $depfile(dir) }
} else {
  readDep -dumpall $dfile
}
getLocalHost
createWindow
