/* 
//  ----------------------------------
//  Darryl Brown
//  University of Oregon pC++/Sage++
//
//  breakExecDef.h  -  module for insecting a pC++ program
//       at breakpoints that must be compiled with the user
//       program.
//
//  ----------------------------------
*/



// see if this file has already been included;

#ifndef BRK_EXEC_DEF
#ifdef BRKPNT_ON

#include "TIbasics.h"
#include "brkSockets.h"
#include <string.h>



// declaring stdfunctions 
extern "C" {
void exit(int);
int close(int);
}

// these variables will be set to identify the location
// before call to generated barrier
int brk_funcId = -1;
char *brk_fileName = "";
int brk_lineNumber = -1;

// global structure for including pcxx_param,
//  allowing access to environment variables.
extern pcxx_PARAM_t pcxx_param;

extern void brk_CollectionAccessFunctionEntryInit();
extern void brk_elementTypeAccessFunctionEntryInit();

/* This points to the queue which names of collection variables are */
/*  placed just before they are created. */
extern brk_linkedList *brk_BPcollDeclQueue;

/* This points to the tables (list) of collection instances and a record */
/* of information about them. */
extern brk_linkedList *brk_BPcollInstanceTable;


#define BRK_LIST_ALL_INSTANCES 1
#define BRK_LIST_TYPES 2
#define BRK_SET_CONTINUE 3
#define BRK_EXIT 4
#define BRK_GET_INSTANCE 5
#define BRK_GET_FIELD 6

#define BRK_EXEC_DEF 0

#define USE_SOCKETS 1

//  this is the compiler switch for brkpnt...;

#define BP_DEBUG 0

//----------------------------------------------------------;
//----------------------------------------------------------;

// file to write output to;
FILE *brk_out_file;
FILE *brk_in_file;
int brk_SocketFD;


// continue var, for setting #of breakpoints to skip before stopping;
static int brk_BrkContinue = 0;

// to maintain collectoin instances
class brk_BPcollInstance : public brk_basePtr {

 public:
  virtual void print();
  virtual void print(int tabs);
  virtual void printToBuf(int tabs, char *buf);
  virtual void print(int tabs, FILE *fptr);
  char *varName;
  char *collName;
  char *typeElem;
  int numDims;
  int numElems;
  int sizeElem;
  int collectionId;
  Kernel *collectionPtr;
};

//----------------------------------------------------------;
//----------------------------------------------------------;

void brk_printItem(int num, char * str) {
  printf("\t%d\t%s\n", num, str);
}

//----------------------------------------------------------;
void brk_printMenu() {
  printf("\n\t\tEnter number preceding desired choice:\n\n");
  brk_printItem(BRK_LIST_ALL_INSTANCES, "list collections");
  brk_printItem(BRK_LIST_TYPES, "print listing of types");
  brk_printItem(BRK_SET_CONTINUE, 
	    "set the number of breakpoints to skip before stopping again.");
  brk_printItem(BRK_EXIT, 
	    "Exit breakpoint executive\n");
  printf("\t>");
}


//----------------------------------------------------------;
// global pointer to choice input;
static char brk_inBuf[MAXLINE];

// ---------------------------------------------------------;


#define CONNECT_ROLE "client"

int brk_open_sockets() {

  if (brk_strsame(CONNECT_ROLE, "server")) {
    // I am a server;
    brk_SocketFD = brk_server_connection(pcxx_param.SocketNum);
  } else {
    // I am the client;
    printf("opening client connection on socket num: %d on host: %s\n", 
	   pcxx_param.SocketNum, pcxx_param.HostAddr);
    brk_SocketFD = brk_client_connection(pcxx_param.HostAddr, 
					      pcxx_param.SocketNum);
  }
  return brk_SocketFD;
}

int brk_writeSocketBuf(char *buf) {
  int i = brk_sockWritestr(brk_SocketFD, buf);
#if BP_DEBUG  
  printf("writing :: %s, %d bytes", buf, i); 
  fflush(stdout);
#endif
  return i;
}

#define LIST_OUTPUT 0
#define BP_DEBUG_SOCK 0

#if BP_DEBUG_SOCK  
int numCount = 0;
#endif

int brk_writeSocket(char *buf) {
  int i;
  char tbuf[4];

#if BP_DEBUG_SOCK
  printf("writing socket string (%s)\n", buf);
  fflush(stdout);
#endif

  i = brk_writeSocketBuf(buf);
  if (i < strlen(buf)) {
    i += brk_writeSocketBuf((buf+i));
  }

#if BP_DEBUG_SOCK
  printf("done writing socket string\n");
  fflush(stdout);
#endif

  return i;
}


int brk_endMessage() {
  int i;

#if BP_DEBUG_SOCK
  printf("wrote %d numbers.\n", numCount);
  numCount = 0;
  printf("terminating message\n");
  fflush(stdout);
#endif

  i = brk_writeSocket("\n!!\n");

  return i;
}

int  brk_readSocket(char *buf) {
  int i;

#if BP_DEBUG_SOCK
  printf("reading socket string\n");
  fflush(stdout);
#endif
  i= brk_sockReadStr(brk_SocketFD, buf, MAXLINE);
#if BP_DEBUG_SOCK
  printf("done reading socket string %s\n", buf);
  fflush(stdout);
#endif
  return i;
}
	
void brk_cleanup() {
#if USE_SOCKETS  
  close(brk_SocketFD);
#endif
}

//----------------------------------------------------------;
int brk_getChoice() {
  char incoming[10];
  int choice;
#ifndef USE_SOCKETS
  int ch;
# endif

#if BP_DEBUG
  printf("in getChoice\n");
  fflush(stdout);
#endif

#if USE_SOCKETS      
  brk_readSocket(brk_inBuf);
  
  if(sscanf(brk_inBuf, "%d", &choice) == 1) {

#if BP_DEBUG
  printf("leaving getChoice with %d choice\n", choice);
  fflush(stdout);
#endif

    return choice;
  } else 
    // check to see if it is the begining of a collection,
    // specification;
  if (*brk_inBuf == '(') {

#if BP_DEBUG
    printf("leaving getChoice with %d choice\n", BRK_GET_FIELD);
    fflush(stdout);
#endif

    return BRK_GET_FIELD;
  } else {

    return -1;
  }
#else
  if ((ch = getc(brk_in_file)) == '(') {
    ungetc(ch, brk_in_file);
    fgets(brk_inBuf, 100, brk_in_file);
    // we want to get a collection..;

#if BP_DEBUG
    printf("leaving getChoice with %d choice\n", BRK_GET_FIELD);
    fflush(stdout);
#endif

    return BRK_GET_FIELD;
  } else {
    ungetc(ch, brk_in_file);
    fgets(brk_inBuf, 100, brk_in_file);
  }
  if(sscanf(brk_inBuf, "%d", &choice) == 1) {

#if BP_DEBUG
    printf("leaving getChoice with %d choice\n", choice);
    fflush(stdout);
#endif

    return choice;
  } else { 
    strcpy(brk_inBuf, "");

#if BP_DEBUG
    printf("leaving getChoice with error\n", choice);
    fflush(stdout);
#endif
    return -1;
  }
#endif
}


brk_aComposite *brk_collFromInstance(brk_BPcollInstance *coll) {
  brk_aComposite *comp;
  comp = brk_findInCompositeList(coll->collName, 
			     coll->typeElem, brk_collectionElemLIST);
  return comp;
}

//----------------------------------------------------------;
int brk_AccessFunctionsInitialized = 0;

//----------------------------------------------------------;
void brk_InitAccessFunctionTables () {


#if BP_DEBUG
  printf("initializing function tables\n");
  fflush(stdout);
#endif

#if USE_SOCKETS
  // init sockets, 
  if ( brk_open_sockets() == -1 ) {brk_cleanup(); exit(1);} /*BM*/

#if BP_DEBUG
  printf("sockets open\n");
  fflush(stdout);
#endif

#else
  // init outfile;
  brk_out_file = openOutFile();
  brk_in_file = openInFile();

#endif

  // read in type info;
  brk_readIn();


#if BP_DEBUG
  printf("TIF file read in\n");
  fflush(stdout);
#endif

  // initialize collection access functions;
  brk_CollectionAccessFunctionEntryInit();
  brk_elementTypeAccessFunctionEntryInit();
  brk_AccessFunctionsInitialized = 1;

#if BP_DEBUG
  printf("done initializing function tables\n");
  fflush(stdout);
#endif
}

//----------------------------------------------------------;
void brk_getCollElem() {

  brk_BPcollInstance * collInst;
  int collnum = 0, elemnum = 0, elemnum2 = 0;
  brk_aComposite *comp;
  char fname[40], buf[100];;
  int numscanned = 0;
  
  
#if BP_DEBUG
  printf("in getCollElem, brk_inBuf: %s\n", brk_inBuf);
  fflush(stdout);
#endif

  *fname = 0;
  // retrieve next line;
  
  if ((numscanned = sscanf(brk_inBuf, "( %d , %*d )->%*s",
			   &collnum)) != 1) {
    
    sprintf(buf, "invalid collection specification: %s\n", brk_inBuf);


#if USE_SOCKETS
    brk_writeSocket(buf);
#else
    fprintf(brk_out_file, "%s\n", buf);
#endif

    // clear brk_inBuf for next time;
    strcpy(brk_inBuf, "");

    return;
  } 
  

  // check for collnum out of range;
  if (collnum < 0  || collnum >= brk_BPcollInstanceTable->length) {
    sprintf(buf, "collection num out of range : %\n", collnum);

#if USE_SOCKETS
    brk_writeSocket(buf);
#else
    fprintf(brk_out_file, "%s\n", buf);
#endif

    // clear brk_inBuf for next time;
    strcpy(brk_inBuf, "");

    return;
  }
  // check for elemnum out of range;
  
  collInst = (brk_BPcollInstance *) brk_BPcollInstanceTable->getIth(collnum);
  comp = brk_collFromInstance(collInst);

  // check the number of dimensions of the distribution of the colleciton;
  if (collInst->numDims == 1) {
    if ((numscanned = sscanf(brk_inBuf, "( %*d , %d )->%s",
			     &elemnum, fname)) != 1) {
      if (elemnum < -1  || 
	  elemnum >= collInst->collectionPtr->GlobalInfo.dim1size) {
	sprintf(buf, "element number out of range: %d\n", elemnum);
	
#if USE_SOCKETS
	brk_writeSocket(buf);
#else
	fprintf(brk_out_file, "%s\n", buf);
#endif
	// clear brk_inBuf for next time;
	strcpy(brk_inBuf, "");
	return;
      } 
    }
  } else if (collInst->numDims == 2) {
    if ((numscanned = sscanf(brk_inBuf, "( %*d , %d , %d)->%s",
			     &elemnum, &elemnum2, fname)) != 1) {
      if ((elemnum < -1  || 
	   elemnum >= collInst->collectionPtr->GlobalInfo.dim1size) ||
	  (elemnum2 < -1  || 
	   elemnum2 >= collInst->collectionPtr->GlobalInfo.dim2size)) {
	sprintf(buf, "element indices out of range: [%d][%d]\n", 
		elemnum, elemnum2);
#if USE_SOCKETS
	brk_writeSocket(buf);
#else
	fprintf(brk_out_file, "%s\n", buf);
#endif
	// clear brk_inBuf for next time;
	strcpy(brk_inBuf, "");
	return;
      } 
    }
  } else {
    printf("cannot have more than 2D distributions for collections!\n");
    exit(1);
  }
    
  // HACKKKKKK - the (char *) cast shouldnt be necessary!!!!;
  // Note that fname may be null, in which case, the whole element,
  // will be retrieved.
    // there was no element field specified, return whole element;
  
#if BP_DEBUG
    printf("about to call access function with ");
    fflush(stdout);
    printf("elemnum %d and fname %s \n", elemnum, fname);
    fflush(stdout);
#endif 

  // this functions writes to the socket itself...;
  
  if (collInst->numDims == 1) {
    (*(comp->accessFunc))((void *) collInst->collectionPtr, elemnum, fname);
  } else if (collInst->numDims == 2) {
    (*(comp->accessFunc))((void *) collInst->collectionPtr, 
	  collInst->collectionPtr->GlobalIndex(elemnum, elemnum2),
			  fname);
  } else {
    printf("cannot have more than 2D distributions for collections!\n");
    exit(1);
  }
    
#if BP_DEBUG
  printf("leaving getCollElem\n");
  fflush(stdout);
#endif 

  // clear brk_inBuf for next time;
  strcpy(brk_inBuf, "");

}
  
//----------------------------------------------------------;

extern "C" void brk_BREAKPOINT_EXECUTIVE(void);

void brk_BREAKPOINT_EXECUTIVE(void) {
  
  int choice = 0, listnum = 0, i;
  char pbuf[1000], tbuf[1000];

  // only execute if you are the zeroth process 
  if (!(pcxx_NodeNumber() == 0)) {
    return;
  }

  if (brk_BREAKPOINT_TOGGLE_VAR == -1) {
    return;
  }

  if (brk_BrkContinue > 0) {
    brk_BrkContinue--;
    return;
  }


#if BP_DEBUG
  printf("in brk_BREAKPOINT_EXECUTIVE: stopped in %d\n", brk_funcId);
  printf("called at %s:%d\n", brk_fileName, brk_lineNumber);
  fflush(stdout);
#endif
    
  // make certain function tables have been initialized.;
  if (!brk_AccessFunctionsInitialized) {
    brk_InitAccessFunctionTables();
  }

  // let user examine collections.;
  do {

#if ! USE_SOCKETS
    brk_printMenu();
#endif
    
    choice = brk_getChoice();
    switch (choice) {
    case BRK_LIST_ALL_INSTANCES: {
#if USE_SOCKETS
      /* report funcId first*/
      sprintf(tbuf, "id=%d#fn=%s#ln=%d#", brk_funcId, brk_fileName,
                                          brk_lineNumber);
      brk_writeSocket(tbuf);

      /* print the list of collections */
      brk_BPcollInstanceTable->printToBuf(0, pbuf);
      brk_writeSocket(pbuf);
      brk_endMessage();
#else
      printf("There are %d collections available:\n",
	     brk_BPcollInstanceTable->length);
      brk_BPcollInstanceTable->printToBuf(0, pbuf);
      printf(pbuf);
      /* brk_BPcollInstanceTable->print(0, brk_out_file); */
#endif
      break;
    }
    case BRK_LIST_TYPES: {
#if USE_SOCKETS
      *pbuf = 0;
      /* print the list of collections */
      brk_writeSocket("collections:\n");
      brk_collectionLIST->printToBuf(1, tbuf);
      brk_writeSocket(tbuf);
      brk_writeSocket("classes:\n");
      brk_classLIST->printToBuf(1, tbuf);
      brk_writeSocket(tbuf);
      brk_writeSocket("structs:\n");
      brk_structLIST->printToBuf(1, tbuf);
      brk_writeSocket(tbuf);
      brk_endMessage();
#else
      /* print the list of collections */
      fprintf(brk_out_file, "type listing:\n\n");
      fprintf(brk_out_file, "collections:\n");
      brk_collectionLIST->print(1, brk_out_file);
      fprintf(brk_out_file, "classes:\n");
      brk_classLIST->print(1, brk_out_file);
      fprintf(brk_out_file, "structs:\n");
      brk_structLIST->print(1, brk_out_file);
#endif
      break;
    }
    case BRK_GET_FIELD: {
      /* lookup correct access function for each collection, and  */
      /* access the collections */
      
      brk_getCollElem();
      brk_endMessage();
      break;
    }
    case BRK_SET_CONTINUE: {
#if ! USE_SOCKETS      
      fprintf(brk_out_file, 
	      "enter the number of time to continue before breaking again: ");
#endif      
      if ((i = brk_getChoice()) > 0) {
	brk_BrkContinue = i;
#if ! USE_SOCKETS      
	printf("setting brk cont to %d\n", brk_BrkContinue);
#endif
      }
      brk_funcId = -1;
      brk_fileName = "";
      brk_lineNumber = -1;
      return;
    }
    case BRK_EXIT: {
      // stop executing program;
      brk_cleanup();
      exit(1);
    }
    default: {
      break;
    }
    }
    
    
  }  while (choice != -1);
#if ! USE_SOCKETS      
  printf("Returning to execution.\n");
#endif


  brk_funcId = -1;
  brk_fileName = "";
  brk_lineNumber = -1;

}

//----------------------------------------------------------;
//----------------------------------------------------------;
//----------------------------------------------------------;
//----------------------------------------------------------;

virtual void brk_BPcollInstance::print() {
  print(0, stdout);
}



virtual void brk_BPcollInstance::print(int tabs) {
  print(tabs, stdout);
}

virtual void brk_BPcollInstance::printToBuf(int tabs, char *buf) {
  char tbuf[1000];
  
#if BP_DEBUG
  printf("printing collInstance\n");
  fflush(stdout);
#endif

  *buf = 0;
  sprintf(tbuf, "\t");
  for (int i=0; i < tabs; i++) {
    strcat(buf, tbuf);
  }
  if (numDims == 1) {
    sprintf(tbuf, " %s < %s > [%d] %s\n", 
	    collName, typeElem, collectionPtr->GlobalInfo.dim1size, varName);
  } else if (numDims == 2) {
    // numDims must equal 2;
    sprintf(tbuf, " %s < %s > [%d][%d] %s\n", 
	    collName, typeElem, collectionPtr->GlobalInfo.dim1size, 
	    collectionPtr->GlobalInfo.dim2size, varName);
  } else {
    // fatal error, cannot handle more than 2 dimensions;
    printf("More than 2 dimensional distribution, no work\n");
    exit(1);
  }
      
  strcat(buf, tbuf);
}

virtual void brk_BPcollInstance::print(int tabs, FILE *fptr) {
  for (int i=0; i < tabs; i++) {
    fprintf(fptr, "\t");
  }
  fprintf(fptr, " %s < %s > [%d] %s\n", 
	  collName, typeElem, numElems, varName);
}

// ---------------------------------------------------------------------
// this function adds a colleciton name and type to a queue, so that when;
// it is created (immediatly after this), we know the variable name and type;
// of the collection.;

void brk_QUE_COLLECTION_INSTANCE (char *collName, char *elemType, 
                              char *varName) {
  brk_BPcollInstance *newColl;

  // only execute if you are the zeroth process 
  if (!(pcxx_NodeNumber() == 0)) {
    return;
  }

#if BP_DEBUG
  printf("queing collection instance \n");
  fflush(stdout);
#endif

  newColl = new brk_BPcollInstance;
  newColl->varName = brk_stringSave(varName);
  newColl->collName = brk_stringSave(collName);
  newColl->typeElem = brk_stringSave(elemType);
  brk_BPcollDeclQueue->push(newColl);
}


// ---------------------------------------------------------------------
// this function is called during the construction of a collection.  It
// retrieves some important information regarding the collection and
// stores it int the collection instance table.;
void brk_ADD_COLLECTION_ENTRY(int numElems, int sizeElem, 
			  Kernel *newC) {
  
  brk_BPcollInstance *collInst;

  // only execute if you are the zeroth process 
  if (!(pcxx_NodeNumber() == 0)) {
    return;
  }

#if BP_DEBUG
  printf("in adding collection entry\n");
  fflush(stdout);
#endif

  collInst = (brk_BPcollInstance *) brk_BPcollDeclQueue->pop();

  if (!collInst) {
    perror("created a collection that was not instrumented.\n");
    exit(1);
  }
  
  collInst->numElems = numElems;
  collInst->numDims = newC->GlobalInfo.number_dimension;
  collInst->sizeElem = sizeElem;
  collInst->collectionPtr = newC;

  // ;
  brk_BPcollInstanceTable->push(collInst);
}

// communications;
extern FILE *brk_openOutFile();
extern FILE *brk_openInFile();

// forward declaratin off this function.;
extern void brk_QUE_COLLECTION_INSTANCE (char *collName, char *elemType, 
                              char *varName);

// adds a collection to the collection instance table;
void brk_ADD_COLLECTION_ENTRY(int numElems, int sizeElem, Kernel *newC);

extern int brk_BREAKPOINT_TOGGLE_VAR;
#define brk_TOGGLE_BREAKPOINT (brk_BREAKPOINT_TOGGLE_VAR *= -1)

#else

//----------------------------------------------------------
// if breakpoint compiler switch is not on...;
// get rid of function calls;

#define brk_TOGGLE_BREAKPOINT
#define brk_QUE_COLLECTION_INSTANCE(a,b,c) 
#define brk_ADD_COLLECTION_ENTRY(a,b,c) 
#define brk_BREAKPOINT_EXECUTIVE()

#endif
#endif
