/*********************************************************************/
/*                  pC++/Sage++  Copyright (C) 1993                  */
/*  Indiana University  University of Oregon  University of Rennes   */
/*********************************************************************/


void  Kernel ::Fetch_Element(int index, int *buf, int startof, int size)
{
  int wh,offset;
  int data1,data2,data3,data4;
  int cs;
  int *pt;
  
  if (GlobalInfo.Table[index] != 0)
    pt = GlobalInfo.Table[index];
  else
    pt = (int *) GlobalInfo.Where[index];
  
  if (!pt)
    Message("Element not found",WARNING);
  
  /* This section of the code was changed by Neel on 2nd feb '93 */
  if (size != -1 || GlobalInfo.element_size <= pcxx_MsgBlockSize)
    cs = size;
  if (size == -1)
    cs = (GlobalInfo.element_size/(int)sizeof(int));
  else
    cs = size;
  for(offset=startof; offset< cs +startof; offset = offset +4)
    {
      /* offset of an int */
      buf[offset-startof] = pt[offset];
      buf[offset-startof+1] = pt[offset+1];
      buf[offset-startof+2] = pt[offset+2];
      buf[offset-startof+3] = pt[offset+3];
    }
}

int *Kernel::Get_Element(int index)
{
  int wh, offset;

  if (GlobalInfo.number_dimension == 2) { /* special case */
    return Get_Element(Get_2DLowIndex(index), Get_2DHighIndex(index));
  }
  TRACE_1D_ACS();
  if (Is_Local(index)) {
    if (!GlobalInfo.Table[GlobalIndex(index)])
      Message("Get_Element_1D: Error in loading local element.",FATAL);
    return GlobalInfo.Table[GlobalIndex(index)];
  } else { 
    return (int *) GlobalInfo.Where[GlobalIndex(index)];
  }
}

int *Kernel::Get_Element(int index1, int index2)
{
  int wh, offset;

  TRACE_2D_ACS();
  if (Is_Local(index1, index2)) {
    if (!GlobalInfo.Table[GlobalIndex(index1, index2)])
      Message("Get_Element_2D: Error in loading local element.",FATAL);
    return GlobalInfo.Table[GlobalIndex(index1, index2)];
  } else {
    return (int *) GlobalInfo.Where[GlobalIndex(index1, index2)];
  }
}

int *Kernel::Get_ElementPart(int index,int startof, int size)
{
  int wh, offset;

  if (GlobalInfo.number_dimension == 2) { /* special case */
    return Get_ElementPart(Get_2DLowIndex(index), Get_2DHighIndex(index),
			   startof, size);
  }
  TRACE_1D_ACS();
  if (Is_Local(index)) {
    if (!GlobalInfo.Table[GlobalIndex(index)])
      Message("Get_ElementPart_1D: Error in loading local element.",FATAL);
    return (GlobalInfo.Table[GlobalIndex(index)] + startof);
  }else {
    return ((int *) GlobalInfo.Where[GlobalIndex(index)]) + startof ;
  }
}

int *Kernel::Get_ElementPart(int index1, int index2, int startof, int size)
{
  int wh, offset;

  TRACE_2D_ACS();
  if (Is_Local(index1, index2)) {
    if (!GlobalInfo.Table[GlobalIndex(index1, index2)])
      Message("Get_ElementPart_2D: Error in loading local element.",FATAL);
    return (GlobalInfo.Table[GlobalIndex(index1, index2)] + startof);
  }else {
    return ((int *) GlobalInfo.Where[GlobalIndex(index1, index2)]) + startof ;
  }
}

int Kernel::Get_ElementPart_Int(int index,int startof)
{
  int wh, offset;

  TRACE_1D_ACS();
  if (Is_Local(index)) {
    if (!GlobalInfo.Table[GlobalIndex(index)])
      Message("Get_ElementPart_Int_1D: Error in loading local element.",FATAL);
    return (int)*((int*)(GlobalInfo.Table[GlobalIndex(index)] + startof));
  }else {
    return (int)*((int*)(GlobalInfo.Where[GlobalIndex(index)] + startof)) ;
  }
}

float Kernel::Get_ElementPart_Float(int index,int startof)
{
  int wh, offset;

  TRACE_1D_ACS();
  if (Is_Local(index)) {
    if (!GlobalInfo.Table[GlobalIndex(index)])
      Message("Get_ElementPart_Float_1D: Error in loading local element.",
	      FATAL);
    return (float)*((float*)(GlobalInfo.Table[GlobalIndex(index)] + startof));
  }else {
    return (float)*((float*)(GlobalInfo.Where[GlobalIndex(index)] + startof)) ;
  }
}

double Kernel::Get_ElementPart_Double(int index,int startof)
{
  int wh, offset;

  TRACE_1D_ACS();
  if (Is_Local(index)) {
    if (!GlobalInfo.Table[GlobalIndex(index)])
      Message("Get_ElementPart_Double_1D: Error in loading local element.",
	      FATAL);
    return (double)*((double*)(GlobalInfo.Table[GlobalIndex(index)]+startof));
  }else {
    return (double)*((double*)(GlobalInfo.Where[GlobalIndex(index)]+startof)) ;
  }
}

int Kernel::Get_ElementPart_Int(int index1, int index2, int startof)
{
  int wh, offset;
  
  TRACE_2D_ACS();
  if (Is_Local(index1, index2)) {
    if (!GlobalInfo.Table[GlobalIndex(index1, index2)])
      Message("Get_ElementPart_Int_2D: Error in loading local element.",FATAL);
    return (int)*((int*)(GlobalInfo.Table[GlobalIndex(index1,index2)]+startof));
  }else {
    return (int)*((int*)(GlobalInfo.Where[GlobalIndex(index1,index2)]+startof));
  }
}

float Kernel::Get_ElementPart_Float(int index1, int index2, int startof)
{
  int wh, offset;
  
  TRACE_2D_ACS();
  if (Is_Local(index1, index2)) {
    if (!GlobalInfo.Table[GlobalIndex(index1, index2)])
      Message("Get_ElementPart_Float_2D: Error in loading local element.",
	      FATAL);
    return (float)*((float*)(GlobalInfo.Table[GlobalIndex(index1,index2)]+startof));
  }else {
    return (float)*((float*)(GlobalInfo.Where[GlobalIndex(index1,index2)]+startof));
  }
}

double Kernel::Get_ElementPart_Double(int index1, int index2, int startof)
{
  int wh, offset;
  
  TRACE_2D_ACS();
  if (Is_Local(index1, index2)) {
    if (!GlobalInfo.Table[GlobalIndex(index1, index2)])
      Message("Get_ElementPart_Double_2D: Error in loading local element.",
	      FATAL);
    return (double)*((double*)(GlobalInfo.Table[GlobalIndex(index1,index2)]+startof));
  }else {
    return (double)*((double*)(GlobalInfo.Where[GlobalIndex(index1,index2)]+startof));
  }
}

int *Kernel::Get_CopyElem(int index)
{
  int wh, offset;
  int *NBuf;

  if (GlobalInfo.number_dimension == 2) { /* special case */
    return Get_CopyElem(Get_2DLowIndex(index), Get_2DHighIndex(index));
  }
  TRACE_1D_ACS();
  if (Is_Local(index)) {
    if (!GlobalInfo.Table[GlobalIndex(index)])
      Message("Get_CopyElem_1D: Error in loading local element.",FATAL);
    return GlobalInfo.Table[GlobalIndex(index)];
  }
  
  NBuf = new int[(GlobalInfo.element_size/(int)sizeof(int)) + 4];
  Fetch_Element(GlobalIndex(index),NBuf, 0, -1);
  return NBuf;
}

int *Kernel::Get_CopyElem(int index1, int index2)
{
  int wh, offset;
  int *NBuf;

  TRACE_2D_ACS();
  if (Is_Local(index1, index2)) {
    if (!GlobalInfo.Table[GlobalIndex(index1, index2)])
      Message("Get_CopyElem_2D: Error in loading local element",FATAL);
    return GlobalInfo.Table[GlobalIndex(index1, index2)];
  }
  
  NBuf = new int[(GlobalInfo.element_size/(int)sizeof(int)) + 4];
  Fetch_Element(GlobalIndex(index1,index2),NBuf, 0,-1);
  return NBuf;
}

int Kernel :: Install_Collection(Distribution *T, Align *A, int sizeelem)
{
  int i,j;
  int proc;
  int count;
  int *allocated;
  int **ElemAllocVector;

  /* initialize the number of collection */
  CollectionNumber = GetNumberOfCollection();
  IncNumberOfCollection();

  PCXX_EVENT(PCXX_EC_KERNEL, PCXX_INSTL_COLL, CollectionNumber);
  PCXX_EVENT(PCXX_EC_KERNEL, PCXX_COLL_INFO_1, GlobalInfo.number_dimension);
  PCXX_EVENT(PCXX_EC_KERNEL, PCXX_COLL_INFO_2, GlobalInfo.element_size);
  PCXX_EVENT(PCXX_EC_KERNEL, PCXX_COLL_INFO_3, GlobalInfo.nb_element);
  PCXX_EVENT(PCXX_EC_KERNEL, PCXX_COLL_INFO_4, GlobalInfo.dim1size);
  PCXX_EVENT(PCXX_EC_KERNEL, PCXX_COLL_INFO_5, GlobalInfo.dim2size);

  GlobalInfo.Table = (int **) malloc(A->nbelement*(int)sizeof(int *));
  if(GlobalInfo.Table == 0) 
    Message("Install_Collection: unable to allocate space.", FATAL);
  GlobalInfo.Where = (int **) malloc(A->nbelement*(int)sizeof(int *));
  if(GlobalInfo.Where == 0) 
    Message("Install_Collection : unable to allocate space", FATAL);
  SetTableOfCollection(CollectionNumber,(char *) &GlobalInfo);
  /* initialize the buffer */
  buffer = new int[(sizeelem/(int)sizeof(int))  + 4];
  if (A->nbelement == 0)
    Message("No Element in the collection",FATAL);

  /* -- determine number of local elements -- */
  count = 0;
  if (GlobalInfo.number_dimension == 1) {
    for (i=0; i < GlobalInfo.dim1size; i++) 
      if (AllocatedProc(i) == pcxx_NodeNumber()) count++;
  }
  else if (GlobalInfo.number_dimension == 2) {
    for (i=0; i < GlobalInfo.dim1size; i++)
      for (j=0; j < GlobalInfo.dim2size; j++)
	if (AllocatedProc(i,j) == pcxx_NodeNumber()) count++;
  }
  else if(GlobalInfo.number_dimension > 2) {
    Message("Install_Collection(): 3D implementation not available.",FATAL);
  }

  AweAllocElemSpace (pcxx_NodeNumber(), count * (sizeelem+4*(int)sizeof(int)));
  pcxx_Barrier();

  /* -- copy global elem space vector -- */
  ElemAllocVector = (int **) malloc(pcxx_TotalNodes() * (int)sizeof(int *));
  for (i=0; i<pcxx_TotalNodes(); i++)
    ElemAllocVector[i] = (int *) AweGetElemSpacePtr(i);
  
  /* -- fill global info table -- */
  if (GlobalInfo.number_dimension == 1) {
    for (i=0; i < GlobalInfo.dim1size; i++) {
      proc = AllocatedProc(i);
      allocated = ElemAllocVector[proc];
      ElemAllocVector[proc] += (sizeelem/(int)sizeof(int) + 4);
      if (proc == pcxx_NodeNumber()) {
	GlobalInfo.Table[GlobalIndex(i)] = allocated;
	GlobalInfo.Where[GlobalIndex(i)] = (int *) 0;
	PCXX_EVENT(PCXX_EC_KERNEL, PCXX_ELEM_INFO, GlobalIndex(i));
      } else {
	GlobalInfo.Table[GlobalIndex(i)] = (int *) 0;
	GlobalInfo.Where[GlobalIndex(i)] = allocated;
      }
    }
  }
  else if (GlobalInfo.number_dimension == 2) {
    for (i=0; i < GlobalInfo.dim1size; i++) {
      for (j=0; j < GlobalInfo.dim2size; j++) {
        proc = AllocatedProc(i,j);
	allocated = ElemAllocVector[proc];
	ElemAllocVector[proc] += (sizeelem/(int)sizeof(int) + 4);
	if (proc == pcxx_NodeNumber()) {
	  GlobalInfo.Table[GlobalIndex(i,j)] = allocated;
	  GlobalInfo.Where[GlobalIndex(i,j)] = (int *) 0;
	  PCXX_EVENT(PCXX_EC_KERNEL, PCXX_ELEM_INFO, GlobalIndex(i,j));
	} else {
	  GlobalInfo.Table[GlobalIndex(i,j)] = (int *) 0;
	  GlobalInfo.Where[GlobalIndex(i,j)] = allocated;
	}
      }
    }
  }
  free (ElemAllocVector);
  pcxx_Barrier();
  return 1;
}

void Kernel :: ~Kernel()
{
  int found;
  int i, j;
  SetTableOfCollection(CollectionNumber,(char *) 0);
  delete buffer;

  /* free local elem space buffer (= pointer to first LOCAL elem) */
  if (GlobalInfo.number_dimension == 1) {
    for (i=0; i < GlobalInfo.dim1size; i++) {
      if (GlobalInfo.Table[GlobalIndex(i)] != 0) {
        AweDealloc (GlobalInfo.Table[GlobalIndex(i)]);
        break;
      }
    }
  }
  else if (GlobalInfo.number_dimension == 2) {
    found = 0;
    for (i=0; (i < GlobalInfo.dim1size) && !found; i++) {
      for (j=0; j < GlobalInfo.dim2size; j++) {
        if (GlobalInfo.Table[GlobalIndex(i,j)] != 0) {
          AweDealloc (GlobalInfo.Table[GlobalIndex(i,j)]);
          found = 1;
          break;
        }
      }
    }
  }
  pcxx_Barrier();

  free(GlobalInfo.Where);
  free(GlobalInfo.Table);

}
