/*********************************************************************/
/*                  pC++/Sage++  Copyright (C) 1993                  */
/*  Indiana University  University of Oregon  University of Rennes   */
/*********************************************************************/

#define WHOLE_ELEMENT -1

/* This code still gets INT aligned data, " /sizeof(int) " */
/* "size" is size in INTs, **NOT** BYTES */
/* "startof" is size INTs, **NOT** BYTES */
/* pcxx_MsgBlockSize is currently in **INTS** */
void Kernel::Fetch_Element(int elemIndex, int *buf, int startof, int size)
{
  int owner, lastmsgsize, numofblocks, i, totalSize;  /* in INTs */

  /* Find processor that owns the data */
  owner = WhereIsData(elemIndex, CollectionNumber); 

  /* Find size of requested transfer */
  if (size == WHOLE_ELEMENT) 
    totalSize = (GlobalInfo.element_size/(int)sizeof(int));
  else
    totalSize = size;

  /* Is requested transfer too big for one message? */
  if (totalSize > pcxx_MsgBlockSize) {
    /* yes, blocking is required */

    lastmsgsize = totalSize % pcxx_MsgBlockSize;
    numofblocks = totalSize / pcxx_MsgBlockSize;

    /* Send "numofblock" messages of size pcxx_MsgBlockSize */
    for (i=0; i<numofblocks; i++) { 
      AskForDataBlock(owner, elemIndex, startof+i*pcxx_MsgBlockSize, 
		      CollectionNumber, pcxx_MsgBlockSize, 
		      buf+i*pcxx_MsgBlockSize); 
    }

    /* Send last remaining message (if any) */
    if (lastmsgsize > 0) {
      AskForDataBlock(owner, elemIndex, startof+numofblocks*pcxx_MsgBlockSize, 
		      CollectionNumber, lastmsgsize,
		      buf+numofblocks*pcxx_MsgBlockSize);
    }
  }
  else
    /* NO blocking is required, just fetch the whole buffer */
    AskForDataBlock(owner, elemIndex, startof, 
		    CollectionNumber, totalSize, buf);
}

int *Kernel::Get_Element(int index)
{
  if (GlobalInfo.number_dimension == 2) { /* special case */
    return Get_Element(Get_2DLowIndex(index), Get_2DHighIndex(index));
  }
  TRACE_1D_ACS();
  if(Is_Local(index)) {
    if(!GlobalInfo.Table[GlobalIndex(index)]) {
      printf("Processor %d : ", pcxx_NodeNumber());
      Message("Get_Element()_1D: Error in loading local element.",FATAL);
      printf("Element index is %d, AllocatedProc is %d\n", 
	     index, AllocatedProc(index));
    }
    PROFILE_LOCAL_ACS();
    return GlobalInfo.Table[GlobalIndex(index)];
  }
  PROFILE_REMOTE_ACS();
  Fetch_Element(GlobalIndex(index),buffer, 0,WHOLE_ELEMENT);
  return (int *) buffer;
}

int *Kernel::Get_Element(int index1, int index2)
{
  TRACE_2D_ACS();
  if (Is_Local(index1, index2)) {
    if (!GlobalInfo.Table[GlobalIndex(index1, index2)]) {
      printf("Processor %d : ", pcxx_NodeNumber());
      Message("Get_Element()_2D: Error in loading local element.",FATAL);
      printf("Element index is (%d, %d), AllocatedProc is %d\n", 
	     index1, index2, AllocatedProc(index1, index2));
    }
    PROFILE_LOCAL_ACS();
    return GlobalInfo.Table[GlobalIndex(index1, index2)];
  }
  PROFILE_REMOTE_ACS();
  Fetch_Element(GlobalIndex(index1, index2), buffer, 0, WHOLE_ELEMENT);
  return (int *) buffer;
}

int *Kernel::Get_ElementPart(int index,int startof, int size)
{
  if (GlobalInfo.number_dimension == 2) { /* special case */
    return Get_ElementPart(Get_2DLowIndex(index), Get_2DHighIndex(index),
			   startof, size);
  }
  TRACE_1D_ACS();
  if (Is_Local(index)) {
    if (!GlobalInfo.Table[GlobalIndex(index)]) {
      printf("Processor %d : ", pcxx_NodeNumber());
      Message("Get_ElementPart()_1D: Error in loading local element.",FATAL);
    }
    PROFILE_LOCAL_ACS();
    return (GlobalInfo.Table[GlobalIndex(index)] + startof);
  }
  PROFILE_REMOTE_ACS();
  Fetch_Element(GlobalIndex(index),buffer,startof, size);
  return (int *) buffer;
}

int *Kernel::Get_ElementPart(int index1, int index2,int startof, int size)
{
  TRACE_2D_ACS();
  if (Is_Local(index1, index2)) {
    if (!GlobalInfo.Table[GlobalIndex(index1, index2)]) {
      printf("Processor %d : ", pcxx_NodeNumber());
      Message("Get_ElementPart()_2D: Error in loading local element.",FATAL);
    }
    PROFILE_LOCAL_ACS();
    return (GlobalInfo.Table[GlobalIndex(index1, index2)] + startof);
  }
  PROFILE_REMOTE_ACS();
  Fetch_Element(GlobalIndex(index1, index2),buffer,startof, size);
  return (int *) buffer;
}

int Kernel::Get_ElementPart_Int(int index, int startof)
{
  TRACE_1D_ACS();
  if (Is_Local(index)) {
    if (!GlobalInfo.Table[GlobalIndex(index)]) {
      printf("Processor %d : ", pcxx_NodeNumber());
      Message("Get_ElementPart_Int()_1D: Error in loading local element.",
	      FATAL);
    }
    PROFILE_LOCAL_ACS();
    return (int)*((int*) (GlobalInfo.Table[GlobalIndex(index)] + startof));
  }
  PROFILE_REMOTE_ACS();
  Fetch_Element(GlobalIndex(index),buffer,startof,(int)sizeof(int));
  return (int)*((int*) buffer);
}

float Kernel::Get_ElementPart_Float(int index, int startof)
{
  TRACE_1D_ACS();
  if (Is_Local(index)) {
    if (!GlobalInfo.Table[GlobalIndex(index)]) {
      printf("Processor %d : ", pcxx_NodeNumber());
      Message("Get_ElementPart_Float()_1D: Error in loading local element.",
	      FATAL);
    }
    PROFILE_LOCAL_ACS();
    return (float)*((float*) (GlobalInfo.Table[GlobalIndex(index)] + startof));
  }
  PROFILE_REMOTE_ACS();
  Fetch_Element(GlobalIndex(index),buffer,startof,(int)sizeof(float));
  return (float)*((float*) buffer);
}

double Kernel::Get_ElementPart_Double(int index, int startof)
{
  TRACE_1D_ACS();
  if (Is_Local(index)) {
    if (!GlobalInfo.Table[GlobalIndex(index)]) {
      printf("Processor %d : ", pcxx_NodeNumber());
      Message("Get_ElementPart_Double()_1D: Error in loading local element.",
	      FATAL);
    }
    PROFILE_LOCAL_ACS();
    return (double)*((double*)(GlobalInfo.Table[GlobalIndex(index)]+startof));
  }
  PROFILE_REMOTE_ACS();
  Fetch_Element(GlobalIndex(index),buffer,startof,(int)sizeof(double));
  return (double)*((double*) buffer);
}

int Kernel::Get_ElementPart_Int(int index1, int index2, int startof)
{
  TRACE_2D_ACS();
  if (Is_Local(index1, index2)) {
    if (!GlobalInfo.Table[GlobalIndex(index1, index2)]) {
      printf("Processor %d : ", pcxx_NodeNumber());
      Message("Get_ElementPart_Int()_2D: Error in loading local element.",
	      FATAL);
    }
    PROFILE_LOCAL_ACS();
    return (int)*((int*)(GlobalInfo.Table[GlobalIndex(index1, index2)] + 
			  startof));
  }
  PROFILE_REMOTE_ACS();
  Fetch_Element(GlobalIndex(index1, index2),buffer,startof,(int)sizeof(int));
  return (int)*((int *)buffer);
}

float Kernel::Get_ElementPart_Float(int index1, int index2, int startof)
{
  TRACE_2D_ACS();
  if (Is_Local(index1, index2)) {
    if (!GlobalInfo.Table[GlobalIndex(index1, index2)]) {
      printf("Processor %d : ", pcxx_NodeNumber());
      Message("Get_ElementPart_Float()_2D: Error in loading local element.",
	      FATAL);
    }
    PROFILE_LOCAL_ACS();
    return (float)*((float*)(GlobalInfo.Table[GlobalIndex(index1, index2)] + 
			  startof));
  }
  PROFILE_REMOTE_ACS();
  Fetch_Element(GlobalIndex(index1, index2),buffer,startof,(int)sizeof(float));
  return  (float)*((float*)buffer);
}

double Kernel::Get_ElementPart_Double(int index1, int index2, int startof)
{
  TRACE_2D_ACS();
  if (Is_Local(index1, index2)) {
    if (!GlobalInfo.Table[GlobalIndex(index1, index2)]) {
      printf("Processor %d : ", pcxx_NodeNumber());
      Message("Get_ElementPart_Double()_2D: Error in loading local element.",
	      FATAL);
    }
    PROFILE_LOCAL_ACS();
    return (double)*((double*)(GlobalInfo.Table[GlobalIndex(index1, index2)] + 
			  startof));
  }
  PROFILE_REMOTE_ACS();
  Fetch_Element(GlobalIndex(index1, index2),
              buffer,startof,(int)sizeof(double));
  return (double)*((double*)buffer);
}

int *Kernel::Get_CopyElem(int index)
{
  int *NBuf;

  if (GlobalInfo.number_dimension == 2) { /* special case */
    return Get_CopyElem(Get_2DLowIndex(index), Get_2DHighIndex(index));
  }
  TRACE_1D_ACS();
  if (Is_Local(index)) {
    if (!GlobalInfo.Table[GlobalIndex(index)]) {
      printf("Processor %d : ", pcxx_NodeNumber());
      Message("Get_CopyElem()_1D: Error in loading local element.",FATAL);
    }
    PROFILE_LOCAL_ACS();
    return GlobalInfo.Table[GlobalIndex(index)];
  }
  
  PROFILE_REMOTE_ACS();
  NBuf = (int *) new double[GlobalInfo.element_size];
  Fetch_Element(GlobalIndex(index), NBuf, 0, WHOLE_ELEMENT);
  return NBuf;
}

int *Kernel::Get_CopyElem(int index1, int index2)
{
  int *NBuf;

  TRACE_2D_ACS();
  if (Is_Local(index1, index2)) {
    if (!GlobalInfo.Table[GlobalIndex(index1, index2)]) {
      printf("Processor %d : ", pcxx_NodeNumber());
      Message("Get_CopyElem()_2D: Error in loading local element.",FATAL);
    }
    PROFILE_LOCAL_ACS();
    return GlobalInfo.Table[GlobalIndex(index1, index2)];
  }

  PROFILE_REMOTE_ACS();
  NBuf = (int *) new double[GlobalInfo.element_size];
  Fetch_Element(GlobalIndex(index1,index2),NBuf, 0,WHOLE_ELEMENT);
  return NBuf;
}

int Kernel :: Install_Collection(Distribution *, Align *A, int sizeelem)
{
  int i, j;
  CollectionNumber = GetNumberOfCollection();
  IncNumberOfCollection();

  PCXX_EVENT(PCXX_EC_KERNEL, PCXX_INSTL_COLL, CollectionNumber);
  PCXX_EVENT(PCXX_EC_KERNEL, PCXX_COLL_INFO_1, GlobalInfo.number_dimension);
  PCXX_EVENT(PCXX_EC_KERNEL, PCXX_COLL_INFO_2, GlobalInfo.element_size);
  PCXX_EVENT(PCXX_EC_KERNEL, PCXX_COLL_INFO_3, GlobalInfo.nb_element);
  PCXX_EVENT(PCXX_EC_KERNEL, PCXX_COLL_INFO_4, GlobalInfo.dim1size);
  PCXX_EVENT(PCXX_EC_KERNEL, PCXX_COLL_INFO_5, GlobalInfo.dim2size);

  PROFILE_COLL_INIT(A->nbelement, GlobalInfo.number_dimension, sizeelem);

  brk_TOGGLE_BREAKPOINT;
  brk_ADD_COLLECTION_ENTRY(A->nbelement, sizeelem, this);

  ElementLink = new int* [A->nbelement];
  GlobalInfo.Table = new int * [A->nbelement];
  GlobalInfo.Where = new int[A->nbelement];
  //pass address of collection to lower level runtime system, KernelCom
  SetTableOfCollection(CollectionNumber,(char *) &GlobalInfo);
  buffer = (int *) new double[sizeelem/(int)sizeof(double)+4];
  if (A->nbelement == 0)
    Message("Install_Collection(): No element in collection.",FATAL);
  
  if (GlobalInfo.number_dimension == 1) {
    for(i = 0; i < GlobalInfo.dim1size; i++) {
      if(AllocatedProc(i) == pcxx_NodeNumber()) {
	GlobalInfo.Table[GlobalIndex(i)] = (int*) 1;
	GlobalInfo.Where[GlobalIndex(i)] = pcxx_NodeNumber(); 
	PCXX_EVENT(PCXX_EC_KERNEL, PCXX_ELEM_INFO, GlobalIndex(i));
      } else {
	GlobalInfo.Table[GlobalIndex(i)] = (int *) 0;
	GlobalInfo.Where[GlobalIndex(i)] = AllocatedProc(i);
      }
    }
  } 
  
  if (GlobalInfo.number_dimension == 2) {
    for (i=0; i < GlobalInfo.dim1size; i++) {
      for (j=0; j < GlobalInfo.dim2size; j++) {
	if (AllocatedProc(i, j) == pcxx_NodeNumber()) {
	  GlobalInfo.Table[GlobalIndex(i,j)] = (int*) 1; 
	  GlobalInfo.Where[GlobalIndex(i,j)] = pcxx_NodeNumber();
	  PCXX_EVENT(PCXX_EC_KERNEL, PCXX_ELEM_INFO, GlobalIndex(i,j));
	} else {
	  GlobalInfo.Table[GlobalIndex(i,j)] = (int *) 0;
	  GlobalInfo.Where[GlobalIndex(i,j)] = AllocatedProc(i,j);
	}
      }
    }
  }
 
  if(GlobalInfo.number_dimension > 2) {
    Message("Install_Collection(): 3D implementation not available.",FATAL);
  }
  
  return 1; 
}

void Kernel::~Kernel()
{
  delete [] buffer;
  delete [] GlobalInfo.Table;
  delete [] GlobalInfo.Where;
  SetTableOfCollection(CollectionNumber, (char *) 0);
}
