/*********************************************************************/
/*                  pC++/Sage++  Copyright (C) 1993                  */
/*  Indiana University  University of Oregon  University of Rennes   */
/*********************************************************************/


void  Kernel ::Fetch_Element(int index, int *buf, int startof, int size)
{
  int cs;
  int *pt;
  
  pt = (int*)GlobalInfo.Table[index] + startof;
  
  if (!pt)
    Message("Element not found",WARNING);

  cs = size;
  if (size == -1)
    cs = GlobalInfo.element_size/sizeof(int)+4;
  for(int i = 0; i < cs; i++)
    buf[i] = GlobalInfo.Table[index][startof+i];
}

int *Kernel::Get_Element(int index)
{
  if (GlobalInfo.number_dimension == 2) { /* special case */
    return Get_Element(Get_2DLowIndex(index), Get_2DHighIndex(index));
  }
  TRACE_1D_ACS();
  if (Is_Local(index)) {
    if (!GlobalInfo.Table[GlobalIndex(index)])
      Message("Get_Element_1D: Error in loading local element.",FATAL);
    PROFILE_LOCAL_ACS();
    return GlobalInfo.Table[GlobalIndex(index)];
  } else { 
    PROFILE_REMOTE_ACS();
//    return (int *) GlobalInfo.Where[GlobalIndex(index)];
    return (int *) GlobalInfo.Table[GlobalIndex(index)];
  }
}

int *Kernel::Get_Element(int index1, int index2)
{
  TRACE_2D_ACS();
  if (Is_Local(index1, index2)) {
    if (!GlobalInfo.Table[GlobalIndex(index1, index2)])
      Message("Get_Element_2D: Error in loading local element.",FATAL);
    PROFILE_LOCAL_ACS();
    return GlobalInfo.Table[GlobalIndex(index1, index2)];
  } else {
    PROFILE_REMOTE_ACS();
//    return (int *) GlobalInfo.Where[GlobalIndex(index1, index2)];
    return GlobalInfo.Table[GlobalIndex(index1, index2)];
  }
}

int *Kernel::Get_ElementPart(int index,int startof, int size)
{
  if (GlobalInfo.number_dimension == 2) { /* special case */
    return Get_ElementPart(Get_2DLowIndex(index), Get_2DHighIndex(index),
			   startof, size);
  }
  TRACE_1D_ACS();
  if (Is_Local(index)) {
    if (!GlobalInfo.Table[GlobalIndex(index)])
      Message("Get_ElementPart_1D: Error in loading local element.",FATAL);
    PROFILE_LOCAL_ACS();
    return (GlobalInfo.Table[GlobalIndex(index)] + startof);
  }else {
    PROFILE_REMOTE_ACS();
//    return ((int *) GlobalInfo.Where[GlobalIndex(index)]) + startof ;
    return (int *) (GlobalInfo.Table[GlobalIndex(index)] + startof);
  }
}

int *Kernel::Get_ElementPart(int index1, int index2, int startof, int )
{
  TRACE_2D_ACS();
  if (Is_Local(index1, index2)) {
    if (!GlobalInfo.Table[GlobalIndex(index1, index2)])
      Message("Get_ElementPart_2D: Error in loading local element.",FATAL);
    PROFILE_LOCAL_ACS();
    return (GlobalInfo.Table[GlobalIndex(index1, index2)] + startof);
  }else {
    PROFILE_REMOTE_ACS();
//    return ((int *) GlobalInfo.Where[GlobalIndex(index1, index2)]) + startof ;
    return (GlobalInfo.Table[GlobalIndex(index1, index2)] + startof);
  }
}

int Kernel::Get_ElementPart_Int(int index,int startof)
{
  TRACE_1D_ACS();
  if (Is_Local(index)) {
    if (!GlobalInfo.Table[GlobalIndex(index)])
      Message("Get_ElementPart_Int_1D: Error in loading local element.",FATAL);
    PROFILE_LOCAL_ACS();
    return (int)*((int*)(GlobalInfo.Table[GlobalIndex(index)] + startof));
  }else {
    PROFILE_REMOTE_ACS();
//    return (int)*((int*)(GlobalInfo.Where[GlobalIndex(index)] + startof)) ;
    return (int)*((int*)(GlobalInfo.Table[GlobalIndex(index)] + startof));
  }
}

float Kernel::Get_ElementPart_Float(int index,int startof)
{
  TRACE_1D_ACS();
  if (Is_Local(index)) {
    if (!GlobalInfo.Table[GlobalIndex(index)])
      Message("Get_ElementPart_Float_1D: Error in loading local element.",
	      FATAL);
    PROFILE_LOCAL_ACS();
    return (float)*((float*)(GlobalInfo.Table[GlobalIndex(index)] + startof));
  }else {
    PROFILE_REMOTE_ACS();
//    return (float)*((float*)(GlobalInfo.Where[GlobalIndex(index)] + startof)) ;
    return (float)*((float*)(GlobalInfo.Table[GlobalIndex(index)] + startof));
  }
}

double Kernel::Get_ElementPart_Double(int index,int startof)
{
  TRACE_1D_ACS();
  if (Is_Local(index)) {
    if (!GlobalInfo.Table[GlobalIndex(index)])
      Message("Get_ElementPart_Double_1D: Error in loading local element.",
	      FATAL);
    PROFILE_LOCAL_ACS();
    return (double)*((double*)(GlobalInfo.Table[GlobalIndex(index)]+startof));
  }else {
    PROFILE_REMOTE_ACS();
//    return (double)*((double*)(GlobalInfo.Where[GlobalIndex(index)]+startof)) ;
    return (double)*((double*)(GlobalInfo.Table[GlobalIndex(index)]+startof));
  }
}

int Kernel::Get_ElementPart_Int(int index1, int index2, int startof)
{
  TRACE_2D_ACS();
  if (Is_Local(index1, index2)) {
    if (!GlobalInfo.Table[GlobalIndex(index1, index2)])
      Message("Get_ElementPart_Int_2D: Error in loading local element.",FATAL);
    PROFILE_LOCAL_ACS();
    return (int)*((int*)(GlobalInfo.Table[GlobalIndex(index1,index2)]+startof));
  }else {
    PROFILE_REMOTE_ACS();
//    return (int)*((int*)(GlobalInfo.Where[GlobalIndex(index1,index2)]+startof));
    return (int)*((int*)(GlobalInfo.Table[GlobalIndex(index1,index2)]+startof));
  }
}

float Kernel::Get_ElementPart_Float(int index1, int index2, int startof)
{
  TRACE_2D_ACS();
  if (Is_Local(index1, index2)) {
    if (!GlobalInfo.Table[GlobalIndex(index1, index2)])
      Message("Get_ElementPart_Float_2D: Error in loading local element.",
	      FATAL);
    PROFILE_LOCAL_ACS();
    return (float)*((float*)(GlobalInfo.Table[GlobalIndex(index1,index2)]+startof));
  }else {
    PROFILE_REMOTE_ACS();
//    return (float)*((float*)(GlobalInfo.Where[GlobalIndex(index1,index2)]+startof));
    return (float)*((float*)(GlobalInfo.Table[GlobalIndex(index1,index2)]+startof));
  }
}

double Kernel::Get_ElementPart_Double(int index1, int index2, int startof)
{
  TRACE_2D_ACS();
  if (Is_Local(index1, index2)) {
    if (!GlobalInfo.Table[GlobalIndex(index1, index2)])
      Message("Get_ElementPart_Double_2D: Error in loading local element.",
	      FATAL);
    PROFILE_LOCAL_ACS();
    return (double)*((double*)(GlobalInfo.Table[GlobalIndex(index1,index2)]+startof));
  }else {
    PROFILE_REMOTE_ACS();
//    return (double)*((double*)(GlobalInfo.Where[GlobalIndex(index1,index2)]+startof));
    return (double)*((double*)(GlobalInfo.Table[GlobalIndex(index1,index2)]+startof));
  }
}

int *Kernel::Get_CopyElem(int index)
{
//  int *NBuf;

  if (GlobalInfo.number_dimension == 2) { /* special case */
    return Get_CopyElem(Get_2DLowIndex(index), Get_2DHighIndex(index));
  }
  TRACE_1D_ACS();
  if (Is_Local(index)) {
    if (!GlobalInfo.Table[GlobalIndex(index)])
      Message("Get_CopyElem_1D: Error in loading local element.",FATAL);
    PROFILE_LOCAL_ACS();
    return GlobalInfo.Table[GlobalIndex(index)];
  }
  
  PROFILE_REMOTE_ACS();
  return GlobalInfo.Table[GlobalIndex(index)];
//  NBuf = new int[(GlobalInfo.element_size/(int)sizeof(int)) + 4];
//  Fetch_Element(GlobalIndex(index),NBuf, 0, -1);
//  return NBuf;
}

int *Kernel::Get_CopyElem(int index1, int index2)
{
//  int *NBuf;

  TRACE_2D_ACS();
  if (Is_Local(index1, index2)) {
    if (!GlobalInfo.Table[GlobalIndex(index1, index2)])
      Message("Get_CopyElem_2D: Error in loading local element",FATAL);
    PROFILE_LOCAL_ACS();
    return GlobalInfo.Table[GlobalIndex(index1, index2)];
  }
  
  PROFILE_REMOTE_ACS();
  return GlobalInfo.Table[GlobalIndex(index1, index2)];
//  NBuf = new int[(GlobalInfo.element_size/(int)sizeof(int)) + 4];
//  Fetch_Element(GlobalIndex(index1,index2),NBuf, 0,-1);
//  return NBuf;
}

int Kernel :: Install_Collection(Distribution *, Align *A, int sizeelem)
{
  int i, j;
  int flag;
  int *p;

  /* initialize the number of collection */
  CollectionNumber = GetNumberOfCollection();
  IncNumberOfCollection();

  PCXX_EVENT(PCXX_EC_KERNEL, PCXX_INSTL_COLL, CollectionNumber);
  PCXX_EVENT(PCXX_EC_KERNEL, PCXX_COLL_INFO_1, GlobalInfo.number_dimension);
  PCXX_EVENT(PCXX_EC_KERNEL, PCXX_COLL_INFO_2, GlobalInfo.element_size);
  PCXX_EVENT(PCXX_EC_KERNEL, PCXX_COLL_INFO_3, GlobalInfo.nb_element);
  PCXX_EVENT(PCXX_EC_KERNEL, PCXX_COLL_INFO_4, GlobalInfo.dim1size);
  PCXX_EVENT(PCXX_EC_KERNEL, PCXX_COLL_INFO_5, GlobalInfo.dim2size);

  PROFILE_COLL_INIT(A->nbelement, GlobalInfo.number_dimension, sizeelem);

  brk_TOGGLE_BREAKPOINT;
  brk_ADD_COLLECTION_ENTRY(A->nbelement, sizeelem, this);

  if (A->nbelement == 0)
    Message("No Element in the collection",FATAL);

  flag = 1;
  if(pcxx_NodeNumber() == 0) {
    p = (int*) SMRTSmalloc(A->nbelement*(int)sizeof(int *));
    flag = 0;
  }
  pcxx_BroadcastBytes(flag, sizeof(int*), &p); 
  ElementLink = (int**) p;
  GlobalInfo.Table = new int*[A->nbelement];
  GlobalInfo.Where = new int[A->nbelement];
  SetTableOfCollection(CollectionNumber,(char *) &GlobalInfo);
  /* allocate the buffer */
  buffer = new int[(sizeelem/(int)sizeof(int))  + 4];

  /* -- fill global info table -- */
  if (GlobalInfo.number_dimension == 1) {
    for (i=0; i < GlobalInfo.dim1size; i++) {
      if (AllocatedProc(i) == pcxx_NodeNumber()) {
        GlobalInfo.Table[GlobalIndex(i)] = (int*) 1;
        GlobalInfo.Where[GlobalIndex(i)] = pcxx_NodeNumber();
        PCXX_EVENT(PCXX_EC_KERNEL, PCXX_ELEM_INFO, GlobalIndex(i));
      } else {
	GlobalInfo.Table[GlobalIndex(i)] = (int*) 0;
        GlobalInfo.Where[GlobalIndex(i)] = AllocatedProc(i);
      }
    }
  }
  
  if (GlobalInfo.number_dimension == 2) {
    for (i=0; i < GlobalInfo.dim1size; i++) {
      for (j=0; j < GlobalInfo.dim2size; j++) {
	if (AllocatedProc(i,j) == pcxx_NodeNumber()) {
          GlobalInfo.Table[GlobalIndex(i,j)] = (int*) 1;
	  GlobalInfo.Where[GlobalIndex(i,j)] = pcxx_NodeNumber();
          PCXX_EVENT(PCXX_EC_KERNEL, PCXX_ELEM_INFO, GlobalIndex(i,j));
        } else {
	  GlobalInfo.Table[GlobalIndex(i,j)] = (int*) 0;
	  GlobalInfo.Where[GlobalIndex(i,j)] = AllocatedProc(i,j);
	}
      }
    }
  }
  
  return 1; 
}

void Kernel :: ~Kernel()
{
  SetTableOfCollection(CollectionNumber,(char *) 0);
  delete [] buffer;
  delete [] GlobalInfo.Table;
  delete [] GlobalInfo.Where;
}
