/*********************************************************************/
/*                  pC++/Sage++  Copyright (C) 1993                  */
/*  Indiana University  University of Oregon  University of Rennes   */
/*********************************************************************/

/***************************************************************************
 *                pC++ Routine for managing access to data                 *
 *                Paragon NX Version                                       *
 *                                                                         *
 *                                                                         *
 *                Written by D. Gannon                                     *
 *                Rewritten by P. Beckman                                  *
 *                Based on cm5 version by F. Bodin                         *
 *                Last overhaul by S. Yang (2/11/94)                   *
 ***************************************************************************/

#include "pcxx_events_def.h"
#include "pcxx_timers.h"
#include "profiler_defs.h"

#include "passive.h"

extern pcxx_PARAM_t pcxx_param;

/**************************************************************************/
/* MACHINE DEPENDANT CODE */
/**************************************************************************/


/**************************************************************************/
/* Machine Dependant Message Prototypes:
/* Paragon NX */
/**************************************************************************/

/* This version uses a BLOCKING recv, so for Polling, a probe must
   ASSURE a message has arrived before crecv is issused */

/*#include <nx.h> */
extern crecv(long tag, char * buff, long len);
extern csend(long tag, long * buff, long len, long node, long ptype);

/*******************************************************************/
void pcxx_Abort() { 
  exit(1); 
}

/*******************************************************************/
/* Receive a buffer that has already arrived and is waiting */
/* This procedure NEEDS the type and size of the message to be received */
void MyRecv(int type, int size)
{
  /* EXCEPTION TO THE RULE... */
  /* If type == EndOfBlockMsg || BroadcastMsg unpack DIRECTLY into 
     buffer pointed by BlockDataBuff */
  if ((type == EndOfBlockMsg) || (type == BroadcastMsg))
    crecv((long)type, (char*)BlockDataBuff, (long)size);
  else
    crecv((long)type, (char*)ReceiveBuffer, (long)size);
}

/*******************************************************************/
/* Send out a buffer to a destination */
void MySend(int type, char * buffer, int len, int ToWhom)
{
  csend((long) type, (long *) buffer, (long) len, (long) ToWhom, 0);
}

/*******************************************************************/
void pcxx_MainBegin(int *argc, char *argv[])
{
  pcxx_InitParam (argc, argv);
  pcxx_InitClocks ();

  MyNodeNumber = mynode();
  MyTotalNodes = numnodes();

  pcxx_EvInit (pcxx_param.tracefile);
  PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_NEWPROC_ENTER, 0);

  /* INITIALIZE BARRIER VARIABLES */
  left_child = (MyTotalNodes > 2*MyNodeNumber+1)? 
                                   2*MyNodeNumber+1: 0;
  right_child = (MyTotalNodes > 2*MyNodeNumber+2)? 
                                   2*MyNodeNumber+2: 0;
  if (left_child && right_child) num_children = 2;
  else if(left_child || right_child) num_children = 1;
  else num_children = 0;
  parent_id = (MyNodeNumber-1)/2;

  barrierFlag = 0;

  ResetCollectionTraversalLevel ();
  PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_NEWPROC_EXIT, MyNodeNumber);
  PCXX_EVENT(PCXX_EC_BASIC, PCXX_PROC_MAIN_CALL, 0);
}

/*******************************************************************/
void pcxx_MainEnd()
{
  PCXX_EVENT(PCXX_EC_BASIC, PCXX_PROC_MAIN_RETURN, 0);
  pcxx_EvClose();
}


/*********************************************************************
  POLL FUNCTION
  *********************************************************************/
void pcxx_Poll()
{
  /* for full NX (with a reasonable hrecv()) no polling is needed. */

  int size, type, from;
  
  PROFILE_ENTER(PROFILE_Poll);

  /* Has a message come in? */
  /*printf("%d Polling()\n",MyNodeNumber);*/

  while(iprobe(-1) > 0) {
    type = (int)infotype(); 
    size = (int)infocount(); 
    from = (int)infonode();
    
    /*printf("Processor %d received a message of type %d from %d\n",
              MyNodeNumber, type, from);*/

    MyRecv(type,size); 
    
    switch (type) {
      case BroadcastMsg:
        broadcast_flag = 1;
	break; 
      case ReduceMsg:
        reduction_flag++;
        PROFILE_EXIT(PROFILE_Poll);
	return;
      case HereIsMsg:
	here_is_data_handler(type, from);
	break;
      case WhereIsMsg:
	where_is_data_handler(type, from);
	break;
      case ReadBlockMsg:
	read_block_handler(type, from);
	break;
      case EndOfBlockMsg:
	end_of_block_handler(type, from);
	break;
      case ReadFArrayMsg:
	read_farray_handler(type, from);
	break;
      case ReadFArrayElemMsg:
	read_farray_elem_handler(type, from);
	break;
      case BarrierFanInMsg:
	BarrierFanIn(from);
	break;
      case BarrierFanOutMsg:
	barrierFlag = DONE;
	break;
      case RemoteExecuteMsg:
	remote_execute_handler(type, from);
	break;
      case EndOfRemoteExecuteMsg:
	end_of_remote_handler(type, from);
	break;
      default:
	fprintf(stderr,"Flushing message type %d on node %d\n",type, 
		pcxx_NodeNumber());
	break;
      };
  }
  PROFILE_EXIT(PROFILE_Poll);
  return;
}

/*********************************************************************
  BROADCAST FUNCTION
  *********************************************************************/
/*-----------------------------------------------------------
  pcxx_BrodcastBytes sends messages to all other nodes from 
  the node with flag = 0.  the message is in the buffer and 
  the length is known.  This function blocks until all 
  processors call this function.
  ------------------------------------------------------------*/
void pcxx_NativeBroadcastBytes(int flag, int length, void * buffer)
{
  int i;
  char *c, *p;

  BlockDataBuff = (char*)buffer;

  /* The following barrier is very important.  First, it prevents a receiver
     from receiving more than one broadcast during poll which may
     overwrite the receiving buffer.  Second, it prevents
     a receiver from receiving a broadcast when it is not ready to receive
     the broadcast, i.e., when it doest not have a buffer to save the
     broadcast.  Also, do not move the above assignment statement to anywhere
     else.            --- S.Y. (2/11/94)
   */
  pcxx_Barrier();

  if (flag == 0) {
    MySend(BroadcastMsg, buffer, length, -1);
  } else {
    while (broadcast_flag == 0) pcxx_Poll();
  }
  broadcast_flag = 0;
}


/*********************************************************************
  REDUCTION FUNCTIONS
  *********************************************************************/
/*-------------------------------------------------------------
  pcxx_NativeReduceDoubleAdd adds together corresponding buffer elements and
  broadcasts the resulting buffer to all nodes.  The original
  buffer content is overwritten.
 --------------------------------------------------------------*/
void pcxx_NativeReduceDoubleAdd(void* buffer,int elem_count)
{
  if(elem_count*sizeof(double) > MAXBUFSIZE) 
    Message("Reduction requires a too large buffer.", FATAL);
  gdsum((double*)buffer,elem_count,(double*)ReceiveBuffer);  
}

void pcxx_NativeReduceFloatAdd(void* buffer,int elem_count)
{
  if(elem_count*sizeof(float) > MAXBUFSIZE) 
    Message("Reduction requires a too large buffer.", FATAL);
  gssum((float*)buffer,elem_count,(float*)ReceiveBuffer);
}

void pcxx_NativeReduceIntAdd(void* buffer,int elem_count)
{
  if(elem_count*sizeof(int) > MAXBUFSIZE) 
    Message("Reduction requires a too large buffer.", FATAL);
  gisum((int*)buffer,elem_count,(int*)ReceiveBuffer);
}

void pcxx_NativeReduceDoubleMax(void* buffer,int elem_count)
{
  if(elem_count*sizeof(double) > MAXBUFSIZE) 
    Message("Reduction requires a too large buffer.", FATAL);
  gdhigh((double*)buffer,elem_count,(double*)ReceiveBuffer);  
}

void pcxx_NativeReduceFloatMax(void* buffer,int elem_count)
{
  if(elem_count*sizeof(float) > MAXBUFSIZE) 
    Message("Reduction requires a too large buffer.", FATAL);
  gshigh((float*)buffer,elem_count,(float*)ReceiveBuffer);  
}

void pcxx_NativeReduceIntMax(void* buffer,int elem_count)
{
  if(elem_count*sizeof(int) > MAXBUFSIZE) 
    Message("Reduction requires a too large buffer.", FATAL);
  gihigh((int*)buffer,elem_count,(int*)ReceiveBuffer);  
}

void pcxx_NativeReduceDoubleMin(void* buffer,int elem_count)
{
  if(elem_count*sizeof(double) > MAXBUFSIZE) 
    Message("Reduction requires a too large buffer.", FATAL);
  gdlow((double*)buffer,elem_count,(double*)ReceiveBuffer);  
}

void pcxx_NativeReduceFloatMin(void* buffer,int elem_count)
{
  if(elem_count*sizeof(float) > MAXBUFSIZE) 
    Message("Reduction requires a too large buffer.", FATAL);
  gslow((float*)buffer,elem_count,(float*)ReceiveBuffer);  
}

void pcxx_NativeReduceIntMin(void* buffer,int elem_count)
{
  if(elem_count*sizeof(int) > MAXBUFSIZE) 
    Message("Reduction requires a too large buffer.", FATAL);
  gilow((int*)buffer,elem_count,(int*)ReceiveBuffer);  
}
