/***************************************************************************
 * Program:         ENVOY EUI Communication Library 
 *
 * Module:          mpceui.h
 *
 * Purpose:         EUI c interface routines
 *
 * Classification:  IBM Internal Use Only
 *
 * Author:          Peter de Jong
 * modified by RKK on 2-8-1993 to make it compatible with C++ 
 ***************************************************************************/
 
#ifndef MPCEUI
#define MPCEUI

/* #include <Viper.h> */
/* #include <stddef.h> */

int mpc_init ();

int mpc_end ();

int mpc_environ (int *, int *);
 
int mpc_stopall (int);
 
int mpc_task_query (int [], int, int);
 
int mpc_task_set (int [], int);
 
int mpc_brecv (char *, int, int *, int *, int*);
 
int mpc_bsend (char *, int, int, int);
 
int mpc_bsendrecv (char *, int, int, int, char *, int, int *, int *);
 
int mpc_bvrecv (char *, int, int, int, int *, int *, int *);
 
int mpc_bvsend (char *, int, int, int, int, int);
 
int mpc_recv (char *, int , int *, int *, int *);
 
int mpc_send (char *, int, int, int, int *);
 
int mpc_status (int);
 
int mpc_vrecv (char *, int, int, int, int *, int *, int *);
 
int mpc_vsend (char *, int, int, int, int, int, int *);
 
int mpc_wait (int *, int *);
 
int mpc_bcast (char *, int, int, int);
 
#ifndef __cplusplus
int mpc_combine (char *, char *, int, int (*)(), int);
#else
int mpc_combine(char *, char *, int, ...);
#endif
 
int mpc_concat (char *, char *, int, int);
 
int mpc_gather (char *, char *, int, int, int);
 
int mpc_index (char *, char *, int, int);
 
#ifndef __cplusplus
int mpc_prefix (char *, char *, int, int (*)(), int);
#else
int mpc_prefix(char *, char *, int, ...);
#endif
 
#ifndef __cplusplus
int mpc_reduce (char *, char *, int, int, int (*)(), int);
#else
int mpc_reduce(char *, char *, int, int, ...);
#endif
 
int mpc_scatter (char *, char *, int, int, int);
 
int mpc_shift (char *, char *, int, int, int, int);
 
int mpc_sync (int);
 
int mpc_getlabel (int *, int);
 
int mpc_getmembers (int [], int);
 
int mpc_getrank (int *, int, int);
 
int mpc_getsize (int *, int);
 
int mpc_gettaskid  (int, int *, int);
 
int mpc_group (int, int [], int, int *);
 
int mpc_partition (int, int, int, int *);
 
int mpc_pack (char *, char *, int, int, int);
 
int mpc_unpack (char *, char *, int, int, int);
 

/* mp_rf.h  */ 

/*
 *	ADD
 */


void i_vadd(int in1[], int in2[], int out[], int *len);
/*
 * 	Integer vector addition.
 */

void s_vadd(float in1[], float in2[], float out[], int *len);
/*
 *      Single precision floating point vector addition.
 */


void d_vadd(double in1[], double in2[], double out[], int *len);
/*
 *      Double precision floating point vector addition.
 */

/*
 *	MUL
 */


void i_vmul( int in1 [], int in2[], int out[], int *len);
/*
 *      Integer vector multiplication.
 */

void s_vmul(float in1[], float in2[] , float out[], int *len);
/*
 *      Single precision floating point vector multiplication.
 */


void d_vmul(double in1[], double in2[], double out[], int *len);
/*
 *      Double precision floating point vector multiplication.
 */

/*
 *	MAX
 */


void i_vmax(int in1[], int in2[], int out[], int *len);
/*
 *      Integer vector  maximun
 */


void s_vmax(float in1[], float in2[], float out[], int *len);
/*
 *      Single precision floating point vector maximun
 */


void d_vmax(double in1[], double in2[], double out[], int *len);
/*
 *      Double precision floating point vector maximun.
 */


void i_vmin(int in1[], int in2[], int out[], int *len);
/*
 *      Integer vector  minimun
 */


void s_vmin(float in1[], float in2[], float out[], int *len);
/*
 *      Single precision floating point vector minimun.
 */

void d_vmin(double in1[], double in2[], double out[], int *len);
/*
 *      Double precision floating point vector minimun.
 */


/*
 *	Bitwise AND/OR/XOR
 */	


void b_vand(char in1[], char in2[], char out[], int *len);
/*
 *     	Bitwise Vector AND 
 */

void b_vor(char in1[], char in2[], char out[], int *len);
/*
 *	Bitwise Vector OR
 */



void b_vxor(char in1[], char in2[], char out[], int *len);
/*
 *      Bitwise Vector XOR
 */

/*
 *	Logical AND/OR
 */



void l_vand(char in1[], char in2[], char out[], int *len);
/*
 *       Logical Vector AND
 */

void l_vor(char in1[], char in2[], char out[],int *len);
/*
 *      Logical Vector OR
 */


/*  group.h */
typedef struct Group {
    int pgid;
    int gid;
    int gsize;
    int *glist;
    int glabel;
    int gkey;
} GROUP;

extern int _ccl_max_ngroup;
extern GROUP *_ccl_p_group;

/********* remove lines when constants are available elsewhere *******/

#define MAX_NGROUP 128

const int ALLGRP = 0;
const int ALL_LABEL = 0;



#define DONTCARE -1
#define VR_SUCCEED 0
#define VR_IDEXIST 1
#define EUI_SUCCEED -2

/**********************************************************************/ 
 
#endif
