/*********************************************************************/
/*                  pC++/Sage++  Copyright (C) 1993                  */
/*  Indiana University  University of Oregon  University of Rennes   */
/*********************************************************************/

/***************************************************************************
 *                                                                         *
 *                                                                         *
 *                HPC++ Routine for managing access to data                *
 *                Sequential Set Of routine                                *
 *                                                                         *
 *                F. Bodin, Indiana University                             *
 *                                                                         *
 ***************************************************************************/
#include <stdio.h>
 
#include "compatible.h"
#ifdef SYS5
#include <string.h>
#else
#include <strings.h>
#endif
 
 
#ifndef NeXT
#  include <malloc.h>
#endif
 
#include <stdlib.h>

#include <stdarg.h>
#include <fcntl.h>

#include "sage.h"
#include "GlobalStruct.h"
#include "pcxx_events_def.h"
#include "pcxx_timers.h"
#include "profiler_defs.h"
#include "brkExecInc.h" 

extern pcxx_PARAM_t pcxx_param;

/* buffer that represente the global data for testing purpose */
#define MaxCollection 50

static int NumberOfCollection = 0;
static PTR_INFO TableOfCollection[MaxCollection];
static char *CurrentCollectionTraversal[MAXPARAL];
static int TraversalLevel = -1;
static int WarningCount =0;
void Message();

/**************************************************************************
 ********************** General Routine ***********************************
 **************************************************************************/
#include "common.h" /* Common support routines for EVERY KernelCom.c (phb) */

/*************************************************************/

/*******************************************************************/
void pcxx_BroadcastBytes(int t, int s, void *b){}

int GetNumberOfCollection()
{
return NumberOfCollection;
}

/*******************************************************************/
void IncNumberOfCollection()
{
  if (NumberOfCollection >= MaxCollection)
    Message("Too many collections.",FATAL);
  NumberOfCollection++;
}

/*******************************************************************/
void SetTableOfCollection(num, addr)
int num;
char *addr;
{
  TableOfCollection[num] = (PTR_INFO)  addr;
}

/*******************************************************************/
void SetNextCollectionTraversalLevel(p)
char *p;
{
  TraversalLevel++;
  if (TraversalLevel >= MAXPARAL)
    Message("ResetLocal(): Too much parallelism.",FATAL);
  CurrentCollectionTraversal[TraversalLevel] = p;
}

/*******************************************************************/
char *GetCollectionTraversalLevel ()
{
  return CurrentCollectionTraversal[TraversalLevel];
}

/*******************************************************************/
void DecCollectionTraversalLevel ()
{
  TraversalLevel--;
}

/*******************************************************************/
void ResetCollectionTraversalLevel ()
{
  TraversalLevel = -1;
}

/*******************************************************************/
void pcxx_Poll()
{
  return;
}

/*******************************************************************/
int pcxx_TotalNodes()
{
  return 1;
}

/*******************************************************************/
int pcxx_NodeNumber()
{
  return 0;
}

/*******************************************************************/
void pcxx_Barrier()
{
  PROFILE_ENTER(PROFILE_Barrier);
  brk_BREAKPOINT_EXECUTIVE();
  PROFILE_EXIT(PROFILE_Barrier);
}


/*******************************************************************/
void Message(st, flag)
char *st;
int flag;
{
  if (flag == WARNING)
    {
      WarningCount++;
      fprintf(stderr, "Warning: %s\n",st);
      return;
    }
  if (flag == FATAL)
    {
      fprintf(stderr, "Fatal error: %s\n",st);
      return;
    }
}

/**************************************************************************
 ********************** Kernel Request ************************************
 **************************************************************************/
 /* basically all these functions activates something on other proc */

int IamtheManager(id)
int id;
{
  return 1;
}

/***************************************************************************
 *                                                                         *
 *   Routine for getting the processor id of a collection element          *
 *   -id: is the index of the collection element                           *
 *   -gdata: is the global data structure of the collection                *
 *   - return the process id that knows where the data is                  *
 *                                                                         *
 ***************************************************************************/


int WhereIsData(id, info)
int id;
int info;
{
  static int manager = 0;
  return 0;
}

/***************************************************************************
 *                                                                         *
 *      Ask the processor that has the data                              *
 *                                                                         *
 ***************************************************************************/

int AskForData(where, index, offset, info)
int where;
int index;
int offset;
int info;
{
  Message("AskForData called in sequential process",FATAL);
  return 0;
}


int AskForDataBlock(where, index, info, buffer)
int where;
int index;
int info;
int *buffer;
{
  Message("AskForDataBlock called in sequential process",FATAL);
  return 0;
}

void pcxx_MainBegin(int *argc, char *argv[])
{
  pcxx_InitParam (argc, argv);
  pcxx_InitClocks ();

  pcxx_EvInit (pcxx_param.tracefile);
  PCXX_EVENT (PCXX_EC_BASIC, PCXX_SETUP_ENTER, 1);
  PCXX_EVENT (PCXX_EC_BASIC, PCXX_SETUP_EXIT, 1);

  PCXX_EVENT (PCXX_EC_BASIC, PCXX_FORK_ENTER, 0);

  /* This used to be "New_Main()" */
  /**********************************************/
  ResetCollectionTraversalLevel();
  PCXX_EVENT (PCXX_EC_BASIC, PCXX_PROC_MAIN_CALL, 0);
}

void pcxx_MainEnd()
{
  PCXX_EVENT (PCXX_EC_BASIC, PCXX_PROC_MAIN_RETURN, 0);
  /**********************************************/

  PCXX_EVENT (PCXX_EC_BASIC, PCXX_FORK_EXIT, 0);
  pcxx_EvClose ();
  exit (0);
}

void AskForRemoteExecute(int CollectionNumber,int element_index,
                         int function_number,char *arg_buffer,int size,
                         long tag_location)
{
  Message("AskForRemoteExecute called in single processor mode.",FATAL);
}

int AskForFArray(int where, unsigned index, unsigned info, int offset,
                       char* buffer, int size)
{
  Message("AskForFArray called in single processor mode.",FATAL);
  return 0;
}

void CopyFArray(char *dest,char *source,int rank,int size)
{
  memcpy(dest,source,size);
/*  
  int i;
  for(i = 0; i < size; i++) dest[i] = source[i];
*/
}

void CopyFArrayElem(char* source, char *buffer,
                  int elem_size, int rank,
                  int i_len, int i_inc, int i_off, 
                  int j_len, int j_inc, int j_off,
                  int k_len, int k_inc, int k_off, 
                  int i, int j, int k)
{
  int m;
  char *pt;
  pt = source + (i + j*i_len + k*i_len*j_len)*elem_size;
  for(m = 0; m < elem_size; m++) buffer[m] = pt[m];
}

int AskForFArrayElem(int where,int index,int Coll_Id,int offset,char *buffer,
                   int elem_size, int rank,
                   int i_len, int i_inc, int i_off, 
                   int j_len, int j_inc, int j_off,
                   int k_len, int k_inc, int k_off, 
                   int i, int j, int k)
{
  Message("AskForFArray called in single processor mode.",FATAL);
  return 0;
}

void pcxx_ReduceIntAdd(void* buffer,int elem_count) {}

void pcxx_ReduceFloatAdd(void* buffer,int elem_count) {}

void pcxx_ReduceDoubleAdd(void* buffer,int elem_count) {}

void pcxx_ReduceDoubleMax(void* buffer,int elem_count) {}

void pcxx_ReduceFloatMax(void* buffer,int elem_count) {}

void pcxx_ReduceIntMax(void* buffer,int elem_count) {}

void pcxx_ReduceDoubleMin(void* buffer,int elem_count) {}

void pcxx_ReduceFloatMin(void* buffer,int elem_count) {}

void pcxx_ReduceIntMin(void* buffer,int elem_count) {}
